Type: Package
Package: thisutils
Title: Collection of Utility Functions for Data Analysis and Computing
Version: 0.4.7
Date: 2026-05-24
Authors@R: 
    c(
    person("Meng", "Xu", email = "mengxu98@qq.com", role = c("aut", "cre"), comment = c(ORCID = "0000-0002-8300-1054")),
    person(given = "Haoliang", family = "Zhu", email = "hurry060215@gmail.com", role = c("aut"), comment = c(ORCID = "0009-0004-9766-5179"))
    )
Maintainer: Meng Xu <mengxu98@qq.com>
Description: Provides utility functions for data analysis and computing. Includes functions for logging, parallel processing, and other computational tasks to streamline workflows.
License: MIT + file LICENSE
URL: https://mengxu98.github.io/thisutils/
BugReports: https://github.com/mengxu98/thisutils/issues
Depends: R (>= 4.1.0)
Imports: cli, doParallel, foreach, Matrix, pak, parallel, Rcpp, rlang,
        stats, utils
Suggests: testthat (>= 3.0.0)
LinkingTo: Rcpp
Config/Needs/website: mengxu98/thistemplate
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
Language: en-US
NeedsCompilation: yes
Packaged: 2026-05-24 13:33:26 UTC; mx
Author: Meng Xu [aut, cre] (ORCID: <https://orcid.org/0000-0002-8300-1054>),
  Haoliang Zhu [aut] (ORCID: <https://orcid.org/0009-0004-9766-5179>)
Repository: CRAN
Date/Publication: 2026-05-24 14:00:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-24 14:07:49 UTC; unix
Archs: thisutils.so.dSYM

Type: Package
Package: thisutils
Title: Reliable Utilities for Reusable Research Workflows
Version: 0.5.0
Date: 2026-08-13
Authors@R: 
    c(
    person(given = "Meng", family = "Xu", email = "mengxu98@qq.com", role = c("aut", "cre"), comment = c(ORCID = "0000-0002-8300-1054")),
    person(given = "Haoliang", family = "Zhu", email = "hurry060215@gmail.com", role = c("aut"), comment = c(ORCID = "0009-0004-9766-5179"))
    )
Maintainer: Meng Xu <mengxu98@qq.com>
Description: Provides reusable building blocks for research packages across
    matrix representation, numerical computation, neighborhood evaluation,
    controlled execution, and runtime interoperability. Core tools preserve
    declared sparse semantics, expose dense-memory and output boundaries,
    standardize neighborhood and classification results, and align serial and
    parallel result and error behavior. Additional helpers support structured
    messages, optional dependency checks, and common statistical workflows.
License: MIT + file LICENSE
URL: https://mengxu98.github.io/thisutils/
BugReports: https://github.com/mengxu98/thisutils/issues
Depends: R (>= 4.1.0)
Imports: callr (>= 3.5.0), cli, methods, Matrix, pak, parallel, ps,
        Rcpp, rlang, stats, utils
Suggests: BiocNeighbors, knitr, remotes, rmarkdown, testthat (>= 3.1.7)
LinkingTo: cli, Rcpp
Config/Needs/website: mengxu98/thistemplate
Config/testthat/edition: 3
VignetteBuilder: knitr
Encoding: UTF-8
RoxygenNote: 7.3.3
Language: en-US
NeedsCompilation: yes
Packaged: 2026-08-13 15:23:52 UTC; mx
Author: Meng Xu [aut, cre] (ORCID: <https://orcid.org/0000-0002-8300-1054>),
  Haoliang Zhu [aut] (ORCID: <https://orcid.org/0009-0004-9766-5179>)
Repository: CRAN
Date/Publication: 2026-08-21 10:40:02 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-08-21 13:00:35 UTC; unix
Archs: thisutils.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(as.character,figlet_text)
S3method(print,figlet_font)
S3method(print,figlet_text)
S3method(print,thisutils_logo)
export("%ss%")
export(add_pkg_file)
export(as_matrix)
export(capitalize)
export(check_ci_env)
export(check_pkg_status)
export(check_r)
export(check_sparsity)
export(classification_metrics_compute)
export(collapse_sparse_rows)
export(compute_lisi)
export(download)
export(fast_row_vars)
export(figlet)
export(figlet_font)
export(filter_nonzero_variance_features)
export(get_namespace_fun)
export(get_verbose)
export(invoke_fun)
export(is_apple_silicon)
export(is_linux)
export(is_osx)
export(is_outlier)
export(is_windows)
export(log_message)
export(matrix_process)
export(matrix_to_table)
export(max_depth)
export(maximump)
export(meanp)
export(minimump)
export(normalization)
export(parallelize_fun)
export(parse_inline_expressions)
export(pearson_correlation)
export(r_square)
export(remove_r)
export(remove_space)
export(run_biocneighbors_knn)
export(run_dense_topk_by_column)
export(run_sparse_topk_by_column)
export(simulate_sparse_matrix)
export(sparse_cor)
export(split_indices)
export(sump)
export(table_to_matrix)
export(thisutils_logo)
export(try_get)
export(unnest_fun)
export(votep)
export(wilkinsonp)
importFrom(Rcpp,sourceCpp)
useDynLib(thisutils, .registration = TRUE)

# **thisutils** <img src="man/figures/logo.svg" align="right" width="120"/>

<!-- badges: start -->

[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/thisutils)](https://CRAN.R-project.org/package=thisutils) [![Conda Version](https://img.shields.io/conda/vn/conda-forge/r-thisutils?color=green)](https://anaconda.org/conda-forge/r-thisutils) [![version](https://img.shields.io/github/r-package/v/mengxu98/thisutils?label=version&color=blue)](https://github.com/mengxu98/thisutils/blob/main/DESCRIPTION) [![R-CMD-check](https://github.com/mengxu98/thisutils/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/mengxu98/thisutils/actions/workflows/R-CMD-check.yaml) [![test-coverage](https://github.com/mengxu98/thisutils/actions/workflows/test-coverage.yaml/badge.svg)](https://github.com/mengxu98/thisutils/actions/workflows/test-coverage.yaml) [![pkgdown](https://github.com/mengxu98/thisutils/actions/workflows/pkgdown.yaml/badge.svg)](https://mengxu98.github.io/thisutils/reference/index.html) [![RStudio CRAN mirror downloads](https://cranlogs.r-pkg.org/badges/grand-total/thisutils)](https://CRAN.R-project.org/package=thisutils)

<!-- badges: end -->

## **Introduction**

[thisutils](https://mengxu98.github.io/thisutils/) is an R package for providing utility functions for data analysis and computing. Includes functions for logging, parallel processing, and other computational tasks to streamline workflows.

## **Installation**

Install CRAN version:

``` r
install.packages("thisutils")
# or
if (!require("pak", quietly = TRUE)) {
  install.packages("pak")
}
pak::pak("thisutils")
```

Install development version from [GitHub](https://github.com/mengxu98/thisutils) use [pak](https://github.com/r-lib/pak):

``` r
if (!require("pak", quietly = TRUE)) {
  install.packages("pak")
}
pak::pak("mengxu98/thisutils")
```

## **Usage**

Please reference [here](https://mengxu98.github.io/thisutils/reference/index.html).

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(thisutils)
library(Matrix)

## ----sparse-semantics---------------------------------------------------------
x <- Matrix(
  c(-3, 0, 2,
    -1, 4, 0),
  nrow = 3,
  sparse = TRUE,
  dimnames = list(paste0("r", 1:3), paste0("c", 1:2))
)

run_sparse_topk(x, k = 2, by = "col")

## ----graph-semantics----------------------------------------------------------
run_sparse_topk_stored(x, k = 2, by = "col")

## ----sparse-table-------------------------------------------------------------
compact <- matrix_to_table(x, keep_zero = FALSE)
expanded <- matrix_to_table(x)
c(compact_rows = nrow(compact), expanded_rows = nrow(expanded))

## ----stable-variance----------------------------------------------------------
offset <- matrix(1e12 + c(0, 1, 2, 3, 0, 1, 2, 3), nrow = 2, byrow = TRUE)
fast_row_vars(offset)
apply(offset, 1, stats::var)

## ----bounded-correlation------------------------------------------------------
set.seed(1)
a <- rsparsematrix(100, 20, density = 0.1)
b <- rsparsematrix(100, 8, density = 0.1)

cors <- sparse_cor(
  a,
  b,
  threshold = 0.15,
  block_size = 4,
  max_dense_bytes = 1024^2,
  max_output_entries = 200
)
cors

## ----controlled-execution-----------------------------------------------------
datasets <- list(first = a, second = b)
correlate <- function(mat) {
  thisutils::sparse_cor(
    mat,
    threshold = 0.15,
    block_size = 4,
    max_dense_bytes = 1024^2,
    max_output_entries = 400
  )
}

serial <- parallelize_fun(
  datasets,
  correlate,
  cores = 1,
  seed = 2026,
  verbose = FALSE
)
psock <- parallelize_fun(
  datasets,
  correlate,
  cores = 2,
  backend = "psock",
  seed = 2026,
  verbose = TRUE,
  progress = FALSE,
  timestamp_format = ""
)

vapply(
  Map(function(lhs, rhs) isTRUE(all.equal(lhs, rhs)), serial, psock),
  identity,
  logical(1)
)

## ----reproducible-execution---------------------------------------------------
draw <- function(cores) {
  parallelize_fun(
    1:4,
    function(i) stats::rnorm(2),
    cores = cores,
    backend = "psock",
    seed = 42,
    verbose = FALSE
  )
}

identical(draw(1), draw(2))

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
has_neighbors <- requireNamespace("BiocNeighbors", quietly = TRUE)

## ----representation-----------------------------------------------------------
library(Matrix)
library(thisutils)

x <- Matrix(
  c(-3, 0, 2, -1, 4, 0),
  nrow = 3,
  sparse = TRUE,
  dimnames = list(paste0("r", 1:3), paste0("c", 1:2))
)

compact <- matrix_to_table(x, keep_zero = FALSE)
roundtrip <- table_to_matrix(compact, return_sparse = TRUE)
collapsed <- collapse_sparse_rows(x, c("g1", "g1", "g2"))

c(
  sparsity = check_sparsity(x),
  stored_coordinates = nrow(compact),
  collapsed_rows = nrow(collapsed),
  roundtrip_equal = isTRUE(all.equal(as.matrix(roundtrip), as.matrix(x)))
)

## ----topk---------------------------------------------------------------------
run_sparse_topk(x, k = 2, by = "col")$value
run_sparse_topk_stored(x, k = 2, by = "col")$value

## ----embedding-data-----------------------------------------------------------
set.seed(20260810)
cell_type <- rep(c("T", "B", "Mono"), each = 40)
batch <- rep(rep(c("A", "B"), each = 20), 3)
metadata <- data.frame(batch = batch, cell_type = cell_type)
centers <- rbind(
  T = c(-3, 0, 0),
  B = c(3, 0, 0),
  Mono = c(0, 3, 0)
)
embedding <- centers[cell_type, , drop = FALSE] +
  matrix(stats::rnorm(120 * 3, sd = 0.7), ncol = 3)

## ----evaluation, eval=has_neighbors-------------------------------------------
mixing <- compute_lisi(
  embedding,
  metadata,
  c("batch", "cell_type"),
  perplexity = 10,
  n_threads = 1,
  max_dense_bytes = 16 * 1024^2
)

neighbors <- run_biocneighbors_knn(
  embedding,
  k = 7,
  exclude_self = TRUE,
  n_threads = 1
)
predicted <- apply(neighbors$idx, 1L, function(index) {
  votes <- sort(table(metadata$cell_type[index]), decreasing = TRUE)
  names(votes)[[1L]]
})
metrics <- classification_metrics_compute(predicted, metadata$cell_type)

c(
  median_batch_lisi = median(mixing$batch),
  median_cell_type_lisi = median(mixing$cell_type),
  accuracy = metrics$accuracy,
  macro_f1 = metrics$macro_f1
)

## ----execution----------------------------------------------------------------
draw <- function(cores, verbose = FALSE) {
  parallelize_fun(
    setNames(1:4, letters[1:4]),
    function(i) stats::rnorm(2),
    cores = cores,
    backend = "psock",
    seed = 42,
    verbose = verbose,
    progress = FALSE,
    timestamp_format = ""
  )
}

serial <- draw(1)
psock <- draw(2, verbose = TRUE)
identical(serial, psock)

## ----dependencies-------------------------------------------------------------
status <- check_r(
  c("Matrix", "BiocNeighbors"),
  install = FALSE,
  verbose = FALSE
)
backend <- if (isTRUE(status[["BiocNeighbors"]])) {
  get_namespace_fun("BiocNeighbors", "findKNN")
} else {
  NULL
}

c(unlist(status), backend_is_function = is.function(backend))

#' @title Print formatted message
#'
#' @description
#' Integrate the message printing function with the \href{https://cli.r-lib.org}{cli} package,
#' and the [base::message] function.
#' The message could be suppressed by [base::suppressMessages].
#'
#' @md
#' @param ... The message to print.
#' @param expr An optional expression to evaluate while capturing its standard output,
#' messages, and warnings, then re-printing them with `log_message()` formatting.
#' The evaluated result is returned invisibly unless it is visible by default.
#' @param verbose Whether to print the message.
#' Default is `TRUE`.
#' @param message_type Type of message.
#' Could be choose one of `"info"`, `"success"`, `"warning"`, `"error"`, `"running"`, and `"ask"`.
#' When `"ask"` is used, the function will prompt the user for a Yes/No/Cancel response using [utils::askYesNo],
#' and returns `TRUE` for Yes, `FALSE` for No, and `NA` for Cancel.
#' Default is `"info"`.
#' @param cli_model Whether to use the `cli` package to print the message.
#' Default is `TRUE`.
#' @param level The level of the message, which affects the indentation.
#' Level `1` has no indentation, higher levels add more indentation.
#' Default is `1`.
#' @param symbol The symbol used for indentation.
#' When specified, it ignores the level parameter and uses the symbol directly.
#' Default is `"  "` (two spaces).
#' @param text_color Color for the message text.
#' Supports R color names (e.g., `"orange"`),
#' hexadecimal colors (e.g., `"#000000"`),
#' basic colors: `"red"`, `"green"`, `"blue"`, `"yellow"`, `"magenta"`,
#' `"cyan"`, `"white"`, `"black"`, `"grey"`, `"silver"`, `"none"`,
#' and bright colors: `"br_red"`, `"br_green"`, `"br_blue"`,
#' `"br_yellow"`, `"br_magenta"`, `"br_cyan"`, `"br_white"`, `"br_black"`.
#' Default is `NULL`.
#' @param back_color Background color for the message text.
#' Details see parameter `text_color`.
#' Default is `NULL`.
#' @param text_style Text styles to apply.
#' Can be one or more of:
#' `"bold"`, `"italic"`, `"underline"`, `"strikethrough"`, `"dim"`, `"inverse"`.
#' Multiple styles can be combined (e.g., `c("bold", "underline")`).
#' Default is `NULL`.
#' @param timestamp Whether to show the current time in the message.
#' Default is `TRUE`.
#' @param timestamp_format Format string for timestamp display.
#' Default is `"%Y-%m-%d %H:%M:%S"`.
#' @param multiline_indent Whether to apply consistent formatting (timestamp and indentation) to each line in multiline messages.
#' When `TRUE`, each line gets the full formatting; when `FALSE`, only the first line gets the timestamp.
#' Default is `FALSE`.
#' @param timestamp_style Whether to apply the same text styling to the timestamp as the message text.
#' When `TRUE`, timestamp formatting matches the message;
#' when `FALSE`, timestamp keeps its default appearance.
#' Default is `FALSE`.
#' @param plain_text Whether to print only the text content.
#' When `TRUE`, level, symbol, timestamp, and message type formatting are suppressed,
#' but color and multiline settings still apply.
#' @param .envir The environment to evaluate calls in.
#' Default is [parent.frame].
#' @param .frame The frame to use for error reporting.
#' Default is `.envir`.
#'
#' @return
#' Formated message, a logical value (`TRUE`/`FALSE`/`NA`) if `message_type = "ask"`,
#' or the evaluated result of `expr` if `expr` is supplied.
#'
#' @references
#' <https://cli.r-lib.org/articles/index.html>
#'
#' @export
#' @examples
#' # basic usage
#' log_message("Hello, ", "world!")
#'
#' log_message("hello, world!")
#'
#' log_message("Hello, world!", timestamp = FALSE)
#'
#' log_message(
#'   "Hello, ", "world!",
#'   message_type = "success"
#' )
#'
#' log_message(
#'   "Hello, world!",
#'   message_type = "warning"
#' )
#'
#' log_message(
#'   "Processing data...",
#'   message_type = "running"
#' )
#'
#' log_message(
#'   "Hello, ", "world!",
#'   cli_model = FALSE
#' )
#'
#'
#' # suppress messages
#' suppressMessages(log_message("Hello, world!"))
#' log_message("Hello, world!", verbose = FALSE)
#' options(log_message.verbose = FALSE)
#' log_message("Hello, world!")
#'
#' # for global verbose option
#' options(log_message.verbose = TRUE)
#' log_message("Hello, world!", verbose = FALSE)
#' options(log_message.verbose = NULL)
#'
#'
#' # cli inline markup
#' log_message("{.arg abc} is a argument")
#'
#' ## 'message' can not deal with cli inline markup
#' message("hello, {.code world}!")
#'
#' log_message("{.val list('abc')} is a {.cls {class(list('abc'))}}")
#'
#' log_message("{.code lm(y ~ x)} is a code example")
#'
#' log_message("{.dt List}list('abc')")
#'
#' log_message("address: {.email example@example.com}")
#'
#' log_message("{.emph R} is a programming language")
#'
#' log_message("{.envvar R_HOME}")
#'
#' log_message("{.file log_message.R} is a file")
#'
#' log_message("{.fn lm} is a function")
#'
#' log_message("{.fun lm} is a function")
#'
#' log_message("{.help lm} to get help")
#'
#' log_message("... see {.help [{.fun lm}](stats::lm)} to learn more")
#'
#' log_message(
#'   "See the {.href [cli homepage](https://cli.r-lib.org)} for details"
#' )
#'
#' log_message("press {.kbd ENTER}")
#'
#' log_message("press {.key ENTER}")
#'
#' log_message("URL: {.url https://cli.r-lib.org}")
#'
#' log_message("Some {.field field}")
#'
#' log_message("{.path /usr/bin/R} is a path")
#'
#' log_message("{.pkg cli} is a package")
#'
#' log_message("{.val object} is a variable")
#'
#' log_message("{.run Rscript log_message.R} is a runnable file")
#'
#' log_message("{.str object} is a string")
#'
#' log_message("{.strong abc} is a strong string")
#'
#' log_message("{.topic stats::lm} is a topic")
#'
#' log_message("{.vignette cli} is a vignette")
#'
#'
#' # set indentation
#' log_message("Hello, world!", level = 2)
#'
#' log_message("Hello, world!", symbol = "->")
#'
#' log_message(
#'   "Hello, world!",
#'   symbol = "#####",
#'   level = 3
#' )
#'
#' # color formatting
#' log_message(
#'   "This is a red message",
#'   text_color = "#ff9900"
#' )
#'
#' log_message(
#'   "This is a message with background",
#'   back_color = "#EE4000"
#' )
#'
#' log_message(
#'   "This is a message with both text and background",
#'   text_color = "white",
#'   back_color = "cyan"
#' )
#'
#' log_message(
#'   "This is a message with background",
#'   back_color = "#EE4000",
#'   cli_model = FALSE
#' )
#'
#' log_message(
#'   "This is a message with both text and background",
#'   text_color = "red",
#'   back_color = "cyan",
#'   cli_model = FALSE
#' )
#'
#' log_message(
#'   "Hex color with {.arg cli_model = FALSE}",
#'   text_color = "#FF5733",
#'   cli_model = FALSE
#' )
#'
#' log_message(
#'   "Bright red text",
#'   text_color = "br_red"
#' )
#'
#' log_message(
#'   "Bright background",
#'   back_color = "br_yellow"
#' )
#'
#' log_message(
#'   "Combined grey and style",
#'   text_color = "grey",
#'   text_style = "bold"
#' )
#'
#' # text style formatting
#' log_message(
#'   "Bold message",
#'   text_style = "bold"
#' )
#'
#' log_message(
#'   "Italic message",
#'   text_style = "italic"
#' )
#'
#' log_message(
#'   "Underlined message",
#'   text_style = "underline"
#' )
#'
#' log_message(
#'   "Combined styles",
#'   text_style = c("bold", "underline")
#' )
#'
#' log_message(
#'   "Color and style",
#'   text_color = "blue",
#'   text_style = c("bold", "italic")
#' )
#'
#' log_message(
#'   "Hex color and style",
#'   text_color = "#FF5733",
#'   text_style = c("bold", "underline")
#' )
#'
#'
#' # multiline message
#' log_message(
#'   "Line 1\nLine 2\nLine 3",
#'   multiline_indent = TRUE,
#'   text_style = "italic"
#' )
#'
#' log_message(
#'   "Multi-line\ncolored\nmessage",
#'   text_color = "blue",
#'   text_style = "italic"
#' )
#'
#' log_message(
#'   "Multi-line\ncolored\nmessage",
#'   text_color = "blue",
#'   timestamp = FALSE
#' )
#'
#' # timestamp styling
#' log_message(
#'   "Multi-line message\nwith timestamp styling",
#'   text_color = "red",
#'   text_style = "bold",
#'   timestamp_style = TRUE
#' )
#'
#' log_message(
#'   "Multi-line message\nwithout timestamp styling",
#'   text_color = "#669999",
#'   text_style = c("bold", "italic")
#' )
#'
#'
#' # combine cli package and log_message
#' log_message(
#'   cli::col_green(
#'     "I am a green line ",
#'     cli::col_blue(
#'       cli::style_underline(
#'         cli::style_bold("with a blue substring")
#'       )
#'     ),
#'     " that becomes green again!"
#'   )
#' )
#'
#' # cli variables
#' fun <- function(x = 1) {
#'   log_message("{.val x}")
#'   log_message("{.val {x}}")
#'   log_message("{.val {x + 1}}")
#' }
#' fun()
#'
#'
#' # print objects directly
#' df <- data.frame(
#'   x = 1:3,
#'   y = letters[1:3],
#'   z = c(" a", "b  ", "c")
#' )
#' log_message("Content:\n", df)
#'
#' # interactive prompt
#' if (interactive()) {
#'   log_message(
#'     "Do you want to continue?",
#'     message_type = "ask"
#'   )
#' }
#'
#' # capture output from another expression
#' fun <- function() {
#'   cat("This is standard output\n")
#'   message("This is a message")
#'   return(1 + 1)
#' }
#' fun()
#'
#' log_message(
#'   expr = fun(),
#'   message_type = "running"
#' )
log_message <- function(
  ...,
  expr = NULL,
  verbose = NULL,
  message_type = c(
    "info",
    "success",
    "warning",
    "error",
    "running",
    "ask"
  ),
  cli_model = TRUE,
  level = 1,
  symbol = "  ",
  text_color = NULL,
  back_color = NULL,
  text_style = NULL,
  multiline_indent = FALSE,
  timestamp = TRUE,
  timestamp_format = paste0(
    "[",
    format(Sys.time(), "%Y-%m-%d %H:%M:%S"),
    "] "
  ),
  timestamp_style = FALSE,
  plain_text = FALSE,
  .envir = parent.frame(),
  .frame = .envir
) {
  expr_supplied <- !missing(expr)
  expr_quo <- if (expr_supplied) substitute(expr) else NULL
  verbose <- get_verbose(verbose)
  message_type <- match.arg(message_type)
  msg <- build_message(...)
  emit_message <- function(msg_text, msg_type = message_type, multiline = multiline_indent) {
    output_message(
      msg = msg_text,
      message_type = msg_type,
      cli_model = cli_model,
      text_color = text_color,
      back_color = back_color,
      text_style = text_style,
      timestamp = timestamp,
      timestamp_format = timestamp_format,
      level = level,
      symbol = symbol,
      multiline_indent = multiline,
      timestamp_style = timestamp_style,
      plain_text = plain_text,
      .envir = .envir
    )
  }

  if (is.null(.envir) || !is.environment(.envir)) {
    .envir <- parent.frame()
  }
  caller_call <- get_caller_call(.frame)

  if (expr_supplied && message_type %in% c("error", "ask")) {
    cli::cli_abort(
      paste0(
        "{.arg message_type} cannot be {.val ",
        message_type,
        "} when {.arg expr} is supplied"
      ),
      call = caller_call,
      .envir = .envir
    )
  }

  if (!expr_supplied && message_type == "error") {
    cli::cli_abort(msg, call = caller_call, .envir = .envir)
  }

  validate_params(
    level = level,
    symbol = symbol,
    text_color = text_color,
    back_color = back_color,
    text_style = text_style,
    timestamp_style = timestamp_style,
    .frame = .frame
  )

  if (!expr_supplied) {
    if (!verbose) {
      return(invisible(NULL))
    }

    if (message_type == "ask") {
      emit_message(msg)
      return(utils::askYesNo(msg = ""))
    }

    emit_message(msg)

    return(invisible(NULL))
  }

  if (verbose && nzchar(msg)) {
    emit_message(msg)
  }

  captured <- capture_logged_expression(
    expr = expr_quo,
    .envir = .envir
  )

  if (verbose) {
    emit_logged_events(
      events = captured$events,
      default_message_type = message_type,
      cli_model = cli_model,
      text_color = text_color,
      back_color = back_color,
      text_style = text_style,
      timestamp = timestamp,
      timestamp_format = timestamp_format,
      level = level,
      symbol = symbol,
      multiline_indent = TRUE,
      timestamp_style = timestamp_style,
      plain_text = plain_text,
      .envir = .envir
    )
  }

  if (!is.null(captured$error)) {
    stop(captured$error)
  }

  if (isTRUE(captured$visible)) {
    return(captured$value)
  }

  invisible(captured$value)
}

#' @title Get the verbose option
#'
#' @description
#' Get the verbose option from the global options or the local argument.
#'
#' @param verbose The verbose option.
#' Default is `NULL`, which means to get the verbose option from the global options.
#'
#' @return The verbose option.
#' @export
#'
#' @examples
#' get_verbose()
#' get_verbose(verbose = FALSE)
#' get_verbose(verbose = TRUE)
#'
#' options(log_message.verbose = FALSE)
#' get_verbose()
#' get_verbose(verbose = TRUE)
#'
#' options(log_message.verbose = TRUE)
#' get_verbose()
#'
#' options(log_message.verbose = NULL)
get_verbose <- function(verbose = NULL) {
  verbose_global <- getOption("log_message.verbose", NULL)

  if (!is.null(verbose)) {
    if (!is.logical(verbose) || length(verbose) != 1) {
      cli::cli_alert_warning(
        "{.arg verbose} is not a logical value, set to {.pkg TRUE}"
      )
      return(TRUE)
    }

    return(verbose)
  }

  if (is.null(verbose_global)) {
    return(TRUE)
  }

  if (!is.logical(verbose_global) || length(verbose_global) != 1) {
    cli::cli_alert_warning(
      "{.arg log_message.verbose} is not a logical value, set to {.pkg NULL}"
    )
    cli::cli_alert_warning(
      "or using {.code options(log_message.verbose = TRUE/FALSE)}"
    )
    options(log_message.verbose = NULL)
    return(TRUE)
  }

  verbose_global
}

build_message <- function(...) {
  args <- list(...)

  if (length(args) == 0) {
    return("")
  }

  processed_args <- lapply(
    args,
    function(arg) {
      if (is.character(arg) && length(arg) == 1) {
        return(arg)
      } else if (is.character(arg)) {
        return(paste(arg, collapse = ""))
      } else {
        tryCatch(
          {
            output <- utils::capture.output(print(arg))
            if (length(output) > 0) {
              return(paste(output, collapse = "\n"))
            } else {
              return(as.character(arg))
            }
          },
          error = function(e) {
            return(as.character(arg))
          }
        )
      }
    }
  )

  msg <- paste0(processed_args, collapse = "")

  if (length(msg) != 1) {
    msg <- paste(msg, collapse = "")
  }

  capitalize(msg)
}

capture_logged_expression <- function(
  expr,
  .envir = parent.frame()
) {
  events <- list()
  output_buffer <- character()
  output_index <- 0L
  pending_output <- character(0)
  result <- NULL
  visible <- FALSE
  captured_error <- NULL

  append_event <- function(type, text) {
    events[[length(events) + 1]] <<- list(
      type = type,
      text = text,
      message_type = NULL
    )
  }

  append_cli_event <- function(text, message_type) {
    events[[length(events) + 1]] <<- list(
      type = "message",
      text = text,
      message_type = message_type
    )
  }

  append_output_text <- function(text) {
    if (!nzchar(text)) {
      return(invisible(NULL))
    }

    combined_text <- paste0(
      paste0(pending_output, collapse = ""),
      text
    )
    lines <- strsplit(combined_text, "\n", fixed = TRUE)[[1]]
    ends_with_newline <- substr(combined_text, nchar(combined_text), nchar(combined_text)) == "\n"

    if (ends_with_newline) {
      complete_lines <- lines
      if (length(complete_lines) > 0 && identical(utils::tail(complete_lines, 1), "")) {
        complete_lines <- utils::head(complete_lines, -1)
      }
      pending_output <<- character(0)
    } else {
      complete_lines <- utils::head(lines, -1)
      pending_output <<- utils::tail(lines, 1)
    }

    for (line in complete_lines) {
      append_event("output", line)
    }

    invisible(NULL)
  }

  flush_pending_output <- function() {
    if (length(pending_output) == 0) {
      return(invisible(NULL))
    }

    append_event("output", paste0(pending_output, collapse = ""))
    pending_output <<- character(0)

    invisible(NULL)
  }

  flush_output_events <- function() {
    if (length(output_buffer) <= output_index) {
      return(invisible(NULL))
    }

    new_lines <- output_buffer[seq.int(
      output_index + 1L,
      length(output_buffer)
    )]
    output_index <<- length(output_buffer)

    for (line in new_lines) {
      append_output_text(paste0(line, "\n"))
    }

    invisible(NULL)
  }

  rewrite_cat_calls <- function(x, helper_symbol) {
    if (is.call(x)) {
      pieces <- lapply(as.list(x), rewrite_cat_calls, helper_symbol = helper_symbol)

      if (is.symbol(pieces[[1]]) && identical(as.character(pieces[[1]]), "cat")) {
        pieces[[1]] <- helper_symbol
      }

      return(as.call(pieces))
    }

    if (is.pairlist(x)) {
      return(as.pairlist(lapply(x, rewrite_cat_calls, helper_symbol = helper_symbol)))
    }

    x
  }

  helper_name <- ".__thisutils_log_message_cat__"
  helper_symbol <- as.symbol(helper_name)
  helper_exists <- exists(helper_name, envir = .envir, inherits = FALSE)
  if (helper_exists) {
    helper_old <- get(helper_name, envir = .envir, inherits = FALSE)
  }
  cat_exists <- exists("cat", envir = .envir, inherits = FALSE)
  if (cat_exists) {
    cat_old <- get("cat", envir = .envir, inherits = FALSE)
  }

  intercepted_cat <- function(
    ...,
    file = "",
    sep = " ",
    fill = FALSE,
    labels = NULL,
    append = FALSE
  ) {
    use_base_cat <- !identical(file, "") || !identical(fill, FALSE) ||
      !is.null(labels) || !identical(append, FALSE)

    if (use_base_cat) {
      return(base::cat(
        ...,
        file = file,
        sep = sep,
        fill = fill,
        labels = labels,
        append = append
      ))
    }

    pieces <- unlist(
      lapply(list(...), as.character),
      use.names = FALSE
    )
    append_output_text(paste0(paste0(pieces, collapse = sep), collapse = ""))

    invisible(NULL)
  }

  assign(helper_name, intercepted_cat, envir = .envir)
  assign("cat", intercepted_cat, envir = .envir)
  on.exit(
    {
      if (helper_exists) {
        assign(helper_name, helper_old, envir = .envir)
      } else if (exists(helper_name, envir = .envir, inherits = FALSE)) {
        rm(list = helper_name, envir = .envir)
      }

      if (cat_exists) {
        assign("cat", cat_old, envir = .envir)
      } else if (exists("cat", envir = .envir, inherits = FALSE)) {
        rm(list = "cat", envir = .envir)
      }
    },
    add = TRUE
  )

  output_con <- textConnection("output_buffer", open = "w", local = TRUE)
  sink(output_con, type = "output")

  output_sink_active <- TRUE
  output_con_open <- TRUE

  release_output_capture <- function() {
    if (output_sink_active) {
      sink(type = "output")
      output_sink_active <<- FALSE
    }

    if (output_con_open) {
      close(output_con)
      output_con_open <<- FALSE
    }

    flush_output_events()
    flush_pending_output()

    invisible(NULL)
  }

  on.exit(
    release_output_capture(),
    add = TRUE
  )

  value <- tryCatch(
    withVisible(
      withCallingHandlers(
        eval(
          rewrite_cat_calls(expr, helper_symbol = helper_symbol),
          envir = .envir
        ),
        message = function(m) {
          flush_output_events()
          flush_pending_output()

          message_text <- sub("[\r\n]+$", "", conditionMessage(m))
          cli_event <- detect_cli_event(
            text = message_text,
            classes = class(m)
          )

          if (is.null(cli_event)) {
            append_event("message", message_text)
          } else {
            append_cli_event(
              text = cli_event$text,
              message_type = cli_event$message_type
            )
          }
          invokeRestart("muffleMessage")
        },
        warning = function(w) {
          flush_output_events()
          flush_pending_output()
          append_event(
            "warning",
            sub("[\r\n]+$", "", conditionMessage(w))
          )
          invokeRestart("muffleWarning")
        }
      )
    ),
    error = function(e) {
      flush_output_events()
      flush_pending_output()
      captured_error <<- e
      list(value = NULL, visible = FALSE)
    }
  )

  release_output_capture()

  result <- value$value
  visible <- value$visible

  list(
    value = result,
    visible = visible,
    events = events,
    error = captured_error
  )
}

emit_logged_events <- function(
  events,
  default_message_type,
  cli_model,
  text_color,
  back_color,
  text_style,
  timestamp,
  timestamp_format,
  level,
  symbol,
  multiline_indent,
  timestamp_style,
  plain_text,
  .envir = parent.frame()
) {
  if (length(events) == 0) {
    return(invisible(NULL))
  }

  for (event in events) {
    event_message_type <- if (!is.null(event$message_type)) {
      event$message_type
    } else if (identical(event$type, "warning")) {
      "warning"
    } else {
      default_message_type
    }

    output_message(
      msg = event$text,
      message_type = event_message_type,
      cli_model = cli_model,
      text_color = text_color,
      back_color = back_color,
      text_style = text_style,
      timestamp = timestamp,
      timestamp_format = timestamp_format,
      level = level,
      symbol = symbol,
      multiline_indent = multiline_indent,
      timestamp_style = timestamp_style,
      plain_text = plain_text,
      .envir = .envir
    )
  }

  invisible(NULL)
}

detect_cli_event <- function(text, classes = character()) {
  if (!length(classes) || !any(grepl("^cli", classes))) {
    return(NULL)
  }

  info_symbol <- intToUtf8(0x2139)
  success_symbol <- intToUtf8(0x2714)
  warning_symbol <- intToUtf8(0x26A0)
  danger_symbol <- intToUtf8(0x2716)

  text_trimmed <- trimws(text, which = "left")
  text_lower <- tolower(text_trimmed)

  strip_literal_prefix <- function(input, prefix) {
    trimws(substr(input, nchar(prefix) + 1L, nchar(input)), which = "left")
  }

  for (prefix in c(paste0(info_symbol, " "), "i ", "I ")) {
    if (startsWith(text_trimmed, prefix)) {
      return(list(
        message_type = "info",
        text = strip_literal_prefix(text_trimmed, prefix)
      ))
    }
  }
  if (startsWith(text_lower, "info: ")) {
    return(list(
      message_type = "info",
      text = sub("^info:\\s*", "", text_trimmed, ignore.case = TRUE)
    ))
  }

  for (prefix in c(paste0(success_symbol, " "), "v ", "V ")) {
    if (startsWith(text_trimmed, prefix)) {
      return(list(
        message_type = "success",
        text = strip_literal_prefix(text_trimmed, prefix)
      ))
    }
  }
  if (startsWith(text_lower, "success: ")) {
    return(list(
      message_type = "success",
      text = sub("^success:\\s*", "", text_trimmed, ignore.case = TRUE)
    ))
  }

  for (prefix in c("! ", paste0(warning_symbol, " "))) {
    if (startsWith(text_trimmed, prefix)) {
      return(list(
        message_type = "warning",
        text = strip_literal_prefix(text_trimmed, prefix)
      ))
    }
  }
  if (startsWith(text_lower, "warning: ")) {
    return(list(
      message_type = "warning",
      text = sub("^warning:\\s*", "", text_trimmed, ignore.case = TRUE)
    ))
  }

  for (prefix in c(paste0(danger_symbol, " "), "x ", "X ")) {
    if (startsWith(text_trimmed, prefix)) {
      return(list(
        message_type = "warning",
        text = strip_literal_prefix(text_trimmed, prefix)
      ))
    }
  }
  if (startsWith(text_lower, "danger: ") || startsWith(text_lower, "error: ")) {
    return(list(
      message_type = "warning",
      text = sub("^(danger|error):\\s*", "", text_trimmed, ignore.case = TRUE)
    ))
  }

  NULL
}

validate_params <- function(
  level,
  symbol,
  text_color,
  back_color,
  text_style,
  timestamp_style,
  .frame
) {
  caller_call <- get_caller_call(.frame)

  if (
    !is.numeric(level) ||
      length(level) != 1 ||
      level < 1 ||
      level != round(level)
  ) {
    cli::cli_abort(
      "{.arg level} must be a positive integer",
      call = caller_call,
      .envir = parent.frame()
    )
  }

  if (!is.character(symbol) || length(symbol) != 1) {
    cli::cli_abort(
      "{.arg symbol} must be a single character string",
      call = caller_call,
      .envir = parent.frame()
    )
  }

  validate_color_param <- function(color_value, param_name, caller_call) {
    if (!is.null(color_value) && !check_color(color_value)) {
      error_msg <- paste0(
        "{.arg {param_name}} must be a valid color name, ",
        "hexadecimal color code (e.g., '#000000'), or R color name"
      )
      cli::cli_abort(
        error_msg,
        call = caller_call,
        .envir = parent.frame(2)
      )
    }
  }

  validate_color_param(text_color, "text_color", caller_call)
  validate_color_param(back_color, "back_color", caller_call)

  if (
    !is.null(text_color) && !is.null(back_color) && text_color == back_color
  ) {
    cli::cli_abort(
      "{.arg text_color} and {.arg back_color} cannot be the same color",
      call = caller_call,
      .envir = parent.frame()
    )
  }

  if (!is.null(text_style)) {
    valid_styles <- c(
      "bold",
      "italic",
      "underline",
      "strikethrough",
      "dim",
      "inverse"
    )
    if (!all(text_style %in% valid_styles)) {
      cli::cli_abort(
        "{.arg text_style} must be one or more of: {.val {valid_styles}}",
        call = caller_call,
        .envir = parent.frame()
      )
    }
  }

  if (!is.logical(timestamp_style) || length(timestamp_style) != 1) {
    cli::cli_abort(
      "{.arg timestamp_style} must be a single logical value",
      call = caller_call,
      .envir = parent.frame()
    )
  }
}

get_indent_part <- function(symbol, level) {
  if (symbol != "  ") {
    paste0(paste(rep(symbol, level), collapse = ""), " ")
  } else if (level > 1) {
    paste(rep("  ", level - 1), collapse = "")
  } else {
    ""
  }
}

format_line_with_style <- function(
  line,
  prefix,
  text_color,
  back_color,
  text_style,
  timestamp_style,
  .envir
) {
  if (is.null(text_color) && is.null(back_color) && is.null(text_style)) {
    return(paste0(prefix, line))
  }

  if (timestamp_style) {
    full_line <- paste0(prefix, line)
    style_formatting(
      msg = full_line,
      text_color = text_color,
      back_color = back_color,
      text_style = text_style,
      cli_model = TRUE,
      .envir = .envir
    )
  } else {
    styled_line <- style_formatting(
      msg = line,
      text_color = text_color,
      back_color = back_color,
      text_style = text_style,
      cli_model = TRUE,
      .envir = .envir
    )
    paste0(prefix, styled_line)
  }
}

output_cli_message <- function(
  message,
  message_type,
  .envir = parent.frame()
) {
  switch(
    EXPR = message_type,
    "info" = cli::cli_alert_info(message, .envir = .envir),
    "success" = cli::cli_alert_success(message, .envir = .envir),
    "warning" = cli::cli_alert_warning(message, .envir = .envir),
    "running" = cli::cli_text(
      paste0(
        cli::make_ansi_style("orange")(cli::symbol$circle_dotted),
        " ",
        message
      ),
      .envir = .envir
    ),
    "ask" = {
      magenta_style <- cli::make_ansi_style("magenta")
      cli::cli_text(
        paste0(magenta_style("?"), " ", message),
        .envir = .envir
      )
    }
  )
}

output_message <- function(
  msg,
  message_type,
  cli_model,
  text_color,
  back_color,
  text_style,
  timestamp,
  timestamp_format,
  level,
  symbol,
  multiline_indent,
  timestamp_style,
  plain_text,
  .envir = parent.frame()
) {
  if (plain_text) {
    plain_msg <- if (length(msg) == 1) {
      msg
    } else {
      paste(msg, collapse = "")
    }

    if (length(plain_msg) == 1 && grepl("\n", plain_msg)) {
      lines <- strsplit(plain_msg, "\n", fixed = TRUE)[[1]]
      for (line in lines) {
        plain_text_output(
          text = line,
          cli_model = cli_model,
          text_color = text_color,
          back_color = back_color,
          text_style = text_style,
          .envir = .envir
        )
      }
    } else {
      plain_text_output(
        text = plain_msg,
        cli_model = cli_model,
        text_color = text_color,
        back_color = back_color,
        text_style = text_style,
        .envir = .envir
      )
    }

    return(invisible(NULL))
  }

  if (cli_model && length(msg) == 1 && grepl("\n", msg)) {
    lines <- strsplit(msg, "\n", fixed = TRUE)[[1]]

    for (i in seq_along(lines)) {
      line <- lines[i]

      if (i == 1 || multiline_indent) {
        timestamp_part <- if (timestamp) {
          timestamp_format
        } else {
          ""
        }
        indent_part <- get_indent_part(symbol, level)
        prefix <- paste0(timestamp_part, indent_part)
      } else {
        indent_part <- get_indent_part(symbol, level)
        alignment_spaces <- if (timestamp) {
          timestamp_width <- nchar(
            timestamp_format
          )
          paste(rep(" ", timestamp_width), collapse = "")
        } else {
          ""
        }
        prefix <- paste0(alignment_spaces, indent_part)
      }

      formatted_line <- format_line_with_style(
        line = line,
        prefix = prefix,
        text_color = text_color,
        back_color = back_color,
        text_style = text_style,
        timestamp_style = timestamp_style,
        .envir = .envir
      )
      output_cli_message(
        message = formatted_line,
        message_type = message_type,
        .envir = .envir
      )
    }
    return(invisible(NULL))
  }

  if (cli_model) {
    timestamp_part <- if (timestamp) {
      timestamp_format
    } else {
      ""
    }
    indent_part <- get_indent_part(symbol, level)

    if (symbol != "  ") {
      final_msg <- paste0(
        timestamp_part,
        paste(rep(symbol, level), collapse = ""),
        " ",
        msg
      )
    } else {
      final_msg <- paste0(
        timestamp_part,
        indent_part,
        msg
      )
    }

    if (!is.null(text_color) || !is.null(back_color) || !is.null(text_style)) {
      if (timestamp_style) {
        final_msg <- style_formatting(
          msg = final_msg,
          text_color = text_color,
          back_color = back_color,
          text_style = text_style,
          cli_model = cli_model,
          .envir = .envir
        )
      } else {
        styled_msg <- style_formatting(
          msg = msg,
          text_color = text_color,
          back_color = back_color,
          text_style = text_style,
          cli_model = cli_model,
          .envir = .envir
        )
        final_msg <- paste0(
          timestamp_part,
          if (symbol != "  ") {
            paste(rep(symbol, level), collapse = "")
          } else {
            indent_part
          },
          if (symbol != "  ") {
            " "
          } else {
            ""
          },
          styled_msg
        )
      }
    }

    output_cli_message(
      message = final_msg,
      message_type = message_type,
      .envir = .envir
    )
  } else {
    formatted_msg <- tryCatch(
      cli::format_inline(msg, .envir = .envir),
      error = function(e) {
        msg
      }
    )

    if (!is.null(text_color) || !is.null(back_color) || !is.null(text_style)) {
      formatted_msg <- style_formatting(
        msg = formatted_msg,
        text_color = text_color,
        back_color = back_color,
        text_style = text_style,
        cli_model = cli_model,
        .envir = .envir
      )
    }

    prefix <- switch(
      EXPR = message_type,
      "info" = "",
      "success" = "SUCCESS: ",
      "warning" = "WARNING: ",
      "running" = "RUNNING: ",
      "ask" = "? "
    )
    message(paste0(prefix, formatted_msg))
  }
}

style_formatting <- function(
  msg,
  text_color,
  back_color,
  text_style,
  cli_model,
  .envir = parent.frame()
) {
  if (is.null(text_color) && is.null(back_color) && is.null(text_style)) {
    return(msg)
  }

  if (!is.null(text_color)) {
    text_fun <- make_color_style(text_color)
    msg <- text_fun(msg)
  }

  if (!is.null(back_color)) {
    back_fun <- make_color_style(back_color, bg = TRUE)
    msg <- back_fun(msg)
  }

  if (!is.null(text_style)) {
    for (style in text_style) {
      style_fun <- switch(
        EXPR = style,
        "bold" = cli::style_bold,
        "italic" = cli::style_italic,
        "underline" = cli::style_underline,
        "strikethrough" = cli::style_strikethrough,
        "dim" = cli::style_dim,
        "inverse" = cli::style_inverse,
        function(x) x
      )

      msg <- style_fun(msg)
    }
  }

  msg
}

plain_text_output <- function(
  text,
  cli_model,
  text_color,
  back_color,
  text_style,
  .envir = parent.frame()
) {
  if (cli_model) {
    if (!is.null(text_color) || !is.null(back_color) || !is.null(text_style)) {
      text <- style_formatting(
        msg = text,
        text_color = text_color,
        back_color = back_color,
        text_style = text_style,
        cli_model = cli_model,
        .envir = .envir
      )
    }
    cli::cli_verbatim(text, .envir = .envir)
  } else {
    formatted_text <- tryCatch(
      cli::format_inline(text, .envir = .envir),
      error = function(e) {
        text
      }
    )

    if (!is.null(text_color) || !is.null(back_color) || !is.null(text_style)) {
      formatted_text <- style_formatting(
        msg = formatted_text,
        text_color = text_color,
        back_color = back_color,
        text_style = text_style,
        cli_model = cli_model,
        .envir = .envir
      )
    }

    message(formatted_text)
  }
}

get_caller_call <- function(
  frame = parent.frame(),
  max_depth = 10
) {
  if (is.null(frame)) {
    return(NULL)
  }

  tryCatch(
    {
      internal_functions <- c(
        "log_message",
        "validate_params",
        "output_message",
        "style_formatting",
        "check_color",
        "make_color_style",
        "extract_function_name"
      )

      calls <- sys.calls()

      if (length(calls) <= 1) {
        return(NULL)
      }

      current_frame <- sys.nframe()

      for (depth in seq_len(min(max_depth, length(calls)))) {
        target_frame <- current_frame - depth

        if (target_frame < 1 || target_frame > length(calls)) {
          next
        }

        call <- calls[[target_frame]]

        if (!is.call(call) || is.null(call[[1]])) {
          next
        }

        fun_name <- extract_function_name(call)

        if (!is.null(fun_name) && !fun_name %in% internal_functions) {
          if (!identical(call[[1]], quote(get_caller_call))) {
            return(call)
          }
        }
      }

      return(NULL)
    },
    error = function(e) {
      NULL
    }
  )
}

extract_function_name <- function(call) {
  tryCatch(
    {
      fun_part <- call[[1]]

      if (is.symbol(fun_part)) {
        return(as.character(fun_part))
      } else if (is.call(fun_part) && length(fun_part) >= 3) {
        if (as.character(fun_part[[1]]) %in% c("::", ":::")) {
          return(as.character(fun_part[[3]]))
        }
      } else if (is.character(fun_part)) {
        return(fun_part[1])
      }

      return(NULL)
    },
    error = function(e) {
      NULL
    }
  )
}

make_color_style <- function(color, bg = FALSE) {
  if (!bg) {
    text_color_map <- list(
      "black" = cli::col_black,
      "red" = cli::col_red,
      "green" = cli::col_green,
      "yellow" = cli::col_yellow,
      "blue" = cli::col_blue,
      "magenta" = cli::col_magenta,
      "cyan" = cli::col_cyan,
      "white" = cli::col_white,
      "grey" = cli::col_grey,
      "silver" = cli::col_silver,
      "none" = cli::col_none,
      "br_black" = cli::col_br_black,
      "br_red" = cli::col_br_red,
      "br_green" = cli::col_br_green,
      "br_yellow" = cli::col_br_yellow,
      "br_blue" = cli::col_br_blue,
      "br_magenta" = cli::col_br_magenta,
      "br_cyan" = cli::col_br_cyan,
      "br_white" = cli::col_br_white
    )

    if (color %in% names(text_color_map)) {
      return(text_color_map[[color]])
    }
  } else {
    bg_color_map <- list(
      "black" = cli::bg_black,
      "red" = cli::bg_red,
      "green" = cli::bg_green,
      "yellow" = cli::bg_yellow,
      "blue" = cli::bg_blue,
      "magenta" = cli::bg_magenta,
      "cyan" = cli::bg_cyan,
      "white" = cli::bg_white,
      "none" = cli::bg_none,
      "br_black" = cli::bg_br_black,
      "br_red" = cli::bg_br_red,
      "br_green" = cli::bg_br_green,
      "br_yellow" = cli::bg_br_yellow,
      "br_blue" = cli::bg_br_blue,
      "br_magenta" = cli::bg_br_magenta,
      "br_cyan" = cli::bg_br_cyan,
      "br_white" = cli::bg_br_white
    )

    if (color %in% names(bg_color_map)) {
      return(bg_color_map[[color]])
    }
  }

  return(cli::make_ansi_style(color, bg = bg))
}

check_color <- function(color) {
  if (!is.character(color) || length(color) != 1) {
    return(FALSE)
  }

  all_cli_colors <- c(
    "black",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "white",
    "grey",
    "silver",
    "none",
    "br_black",
    "br_red",
    "br_green",
    "br_yellow",
    "br_blue",
    "br_magenta",
    "br_cyan",
    "br_white"
  )

  if (color %in% all_cli_colors) {
    return(TRUE)
  }

  tryCatch(
    {
      invisible(cli::make_ansi_style(color))
      return(TRUE)
    },
    error = function(e) {
      FALSE
    }
  )
}

#' @title Parse inline expressions
#'
#' @description
#' Parse `{}` inline expressions and evaluate them in the current environment,
#' while preserving outer formatting markers like `\{.val ...\}`.
#'
#' @param text A character string containing inline expressions to parse.
#' @param env Environment in which to evaluate expressions.
#' Defaults to the calling environment.
#'
#' @return
#' A character string with expressions evaluated but formatting preserved.
#'
#' @export
#'
#' @examples
#' i <- 1
#' parse_inline_expressions(
#'   "{.val {i}}"
#' )
#'
#' x <- 5
#' y <- 10
#' parse_inline_expressions(
#'   "{.pkg {x + y}}"
#' )
#'
#' name <- "testing"
#' name <- parse_inline_expressions(
#'   "{.pkg {name}}"
#' )
#' name
#'
#' log_message(name)
parse_inline_expressions <- function(
  text,
  env = parent.frame()
) {
  vapply(
    text,
    replace_expressions,
    character(1),
    env = env,
    USE.NAMES = FALSE
  )
}

replace_expressions <- function(text, env) {
  pattern <- "\\{([^{}]+)\\}"

  max_iterations <- 10
  iteration <- 0

  while (grepl(pattern, text) && iteration < max_iterations) {
    iteration <- iteration + 1

    matches <- gregexpr(pattern, text, perl = TRUE)[[1]]
    if (matches[1] == -1) {
      break
    }

    match_starts <- as.numeric(matches)
    match_lengths <- attr(matches, "match.length")

    order_idx <- order(match_starts, decreasing = TRUE)

    for (i in order_idx) {
      start_pos <- match_starts[i]
      match_length <- match_lengths[i]

      match <- substr(text, start_pos, start_pos + match_length - 1)
      replacement <- process_match(match, env)

      text <- paste0(
        substr(text, 1, start_pos - 1),
        replacement,
        substr(text, start_pos + match_length, nchar(text))
      )
    }
  }

  text
}

process_match <- function(match, env) {
  inner_content <- substr(match, 2, nchar(match) - 1)

  if (grepl("^\\.[a-zA-Z_]+\\s+", inner_content)) {
    process_format_expression(inner_content, env)
  } else {
    evaluate_expression(inner_content, env)
  }
}

process_format_expression <- function(content, env) {
  parts <- strsplit(content, "\\s+", 2)[[1]]

  if (length(parts) != 2) {
    return(paste0("{", content, "}"))
  }

  format_tag <- parts[1]
  format_content <- parts[2]

  if (grepl("\\{", format_content)) {
    return(paste0("{", content, "}"))
  }

  evaluated <- evaluate_expression(format_content, env)
  paste0("{", format_tag, " ", evaluated, "}")
}

evaluate_expression <- function(expr, env) {
  tryCatch(
    {
      result <- eval(parse(text = expr), envir = env)
      as.character(result)
    },
    error = function(e) {
      expr
    }
  )
}

test_that("log_message info produces message", {
  expect_message(log_message("test info"))
})

test_that("log_message success produces message", {
  expect_message(log_message("test success", message_type = "success"))
})

test_that("log_message warning produces message", {
  expect_message(log_message("test warning", message_type = "warning"))
})

test_that("log_message error throws error", {
  expect_error(log_message("test error", message_type = "error"))
})

test_that("log_message respects verbose = FALSE", {
  expect_silent(log_message("silent", verbose = FALSE))
})

test_that("log_message respects global verbose option", {
  old <- options(log_message.verbose = FALSE)
  on.exit(options(old))
  expect_silent(log_message("silent"))
})

test_that("log_message error still throws even when verbose = FALSE", {
  expect_error(
    log_message("err", message_type = "error", verbose = FALSE)
  )
})

test_that("log_message with cli_model = FALSE", {
  expect_message(
    log_message("test", cli_model = FALSE)
  )
})

test_that("log_message with timestamp = FALSE", {
  expect_message(log_message("test", timestamp = FALSE))
})

test_that("log_message with text_color", {
  expect_message(log_message("red text", text_color = "red"))
})

test_that("log_message with text_style", {
  expect_message(log_message("bold text", text_style = "bold"))
})

test_that("log_message with level", {
  expect_message(log_message("indented", level = 2))
})

test_that("log_message with multiline", {
  expect_message(log_message("line1\nline2", multiline_indent = TRUE))
})

test_that("log_message captures expression output and returns value", {
  expect_message(
    result <- log_message(
      expr = {
        cat("triMean is used\n")
        42
      },
      cli_model = FALSE,
      timestamp = FALSE
    ),
    "triMean is used"
  )

  expect_equal(result, 42)
})

test_that("log_message captures expression messages and warnings", {
  expect_no_warning(
    expect_message(
      result <- log_message(
        expr = {
          message("Create a CellChat object")
          warning("Potential issue")
          "ok"
        },
        cli_model = FALSE,
        timestamp = FALSE
      ),
      "Create a CellChat object"
    )
  )

  expect_equal(result, "ok")
})

test_that("log_message captures cat output without trailing newline", {
  expect_no_warning(
    expect_message(
      result <- log_message(
        expr = {
          cat("This is standard output")
          message("This is a message")
          warning("This is a warning")
          2
        },
        cli_model = FALSE,
        timestamp = FALSE,
        message_type = "running"
      ),
      "This is standard output"
    )
  )

  expect_equal(result, 2)
})

test_that("log_message preserves output order for cat without trailing newline", {
  captured <- capture.output(
    result <- log_message(
      expr = {
        cat("This is standard output")
        message("This is a message")
        warning("This is a warning")
        2
      },
      cli_model = FALSE,
      timestamp = FALSE,
      message_type = "running"
    ),
    type = "message"
  )

  expect_equal(
    captured,
    c(
      "RUNNING: This is standard output",
      "RUNNING: This is a message",
      "WARNING: This is a warning"
    )
  )
  expect_equal(result, 2)
})

test_that("log_message evaluates expr silently when verbose is FALSE", {
  expect_silent(
    result <- log_message(
      expr = {
        cat("hidden output\n")
        message("hidden message")
        7
      },
      verbose = FALSE,
      cli_model = FALSE,
      timestamp = FALSE
    )
  )

  expect_equal(result, 7)
})

test_that("log_message remaps cli alerts captured from expr", {
  captured <- capture.output(
    log_message(
      expr = {
        cli::cli_alert_info("hello")
        cli::cli_alert_warning("warn")
        cli::cli_alert_success("ok")
        cli::cli_text("plain text")
      },
      cli_model = FALSE,
      timestamp = FALSE,
      message_type = "running"
    ),
    type = "message"
  )

  expect_equal(
    captured,
    c(
      "hello",
      "WARNING: warn",
      "SUCCESS: ok",
      "RUNNING: plain text"
    )
  )
})

test_that("log_message avoids double cli prefixes for expr output", {
  captured <- capture.output(
    log_message(
      expr = {
        cli::cli_alert_info("hello")
        cli::cli_alert_warning("warn")
      },
      cli_model = TRUE,
      timestamp = FALSE
    ),
    type = "message"
  )

  expect_length(captured, 2)
  expect_false(grepl("ℹ .*ℹ ", captured[1]))
  expect_false(grepl("! .*!", captured[2]))
  expect_match(captured[1], "hello$")
  expect_match(captured[2], "warn$")
})

test_that("log_message disallows ask and error message types with expr", {
  expect_error(
    log_message(expr = 1 + 1, message_type = "ask")
  )
  expect_error(
    log_message(expr = 1 + 1, message_type = "error")
  )
})

test_that("log_message errors on invalid level", {
  expect_error(log_message("test", level = -1))
  expect_error(log_message("test", level = 1.5))
})

test_that("log_message errors on same text_color and back_color", {
  expect_error(
    log_message("test", text_color = "red", back_color = "red")
  )
})

test_that("log_message errors on invalid text_style", {
  expect_error(log_message("test", text_style = "nonexistent"))
})

test_that("get_verbose defaults to TRUE", {
  old <- options(log_message.verbose = NULL)
  on.exit(options(old))
  expect_true(get_verbose())
})

test_that("get_verbose respects explicit FALSE", {
  expect_false(get_verbose(verbose = FALSE))
})

test_that("get_verbose respects global option", {
  old <- options(log_message.verbose = FALSE)
  on.exit(options(old))
  expect_false(get_verbose())
})

test_that("get_verbose gives explicit argument precedence over global option", {
  old <- options(log_message.verbose = FALSE)
  on.exit(options(old))
  expect_true(get_verbose(verbose = TRUE))
})

test_that("log_message explicit verbose overrides global option", {
  old <- options(log_message.verbose = FALSE)
  on.exit(options(old))

  expect_message(
    log_message("visible", verbose = TRUE, cli_model = FALSE, timestamp = FALSE),
    "Visible"
  )
})

test_that("log_message keeps visible expr results visible", {
  captured <- withVisible(
    log_message(
      expr = 1 + 1,
      cli_model = FALSE,
      timestamp = FALSE
    )
  )

  expect_true(captured$visible)
  expect_equal(captured$value, 2)
})

test_that("parse_inline_expressions evaluates expressions", {
  i <- 1
  result <- parse_inline_expressions("{.val {i}}")
  expect_true(grepl("1", result))
})

test_that("parse_inline_expressions handles arithmetic", {
  x <- 5
  y <- 10
  result <- parse_inline_expressions("{.pkg {x + y}}")
  expect_true(grepl("15", result))
})

#' @title Add a package file and print package information
#'
#' @md
#' @inheritParams log_message
#' @param use_figlet Whether to use figlet for ASCII art generation.
#' Default is `TRUE`. Details see [figlet].
#' @param figlet_font Character string, figlet font to use.
#' Default is `"Slant"`.
#' @param colors Character vector, colors to use for the logo elements.
#'
#' @return
#' Creates a file named `R/<pkg_name>-package.R`.
#'
#' @export
add_pkg_file <- function(
    use_figlet = TRUE,
    figlet_font = "Slant",
    colors = c(
      "red", "yellow", "green", "magenta", "cyan",
      "yellow", "green", "white", "magenta", "cyan"
    ),
    verbose = TRUE) {
  desc_file <- "DESCRIPTION"
  pkgdown_file <- "_pkgdown.yml"

  desc_info <- read_description(desc_file, verbose)

  pkg_name <- desc_info$Package
  title <- desc_info$Title
  pkg_description <- desc_info$Description
  author_name <- desc_info$Author
  author_email <- desc_info$Email
  github_url <- desc_info$GitHub_URL

  if (is.null(pkg_name)) {
    log_message(
      "{.cls package_name} not found in {.file {desc_file}}.
      Please add 'Package: {.cls package_name}'",
      message_type = "error"
    )
  }
  if (is.null(title)) {
    log_message(
      "{.cls package_title} not found in {.file {desc_file}}.
      Please add 'Title: {.cls package_title}'",
      message_type = "error"
    )
  }
  if (is.null(author_name)) {
    log_message(
      "{.cls author_name} not found in {.file {desc_file}}.
      Please add 'Authors@R:' or 'Maintainer:' field",
      message_type = "error"
    )
  }
  if (is.null(author_email)) {
    log_message(
      "{.cls author_email} not found in {.file {desc_file}}.
      Please add email to 'Authors@R:' or 'Maintainer:' field",
      message_type = "error"
    )
  }

  log_message(
    "Creating logo for: {.pkg {pkg_name}}",
    verbose = verbose
  )

  ascii_lines <- NULL
  if (use_figlet) {
    tryCatch(
      {
        ascii_art <- figlet(
          pkg_name,
          font = figlet_font,
          width = 60,
          justify = "centre",
          absolute = TRUE
        )
        ascii_lines <- as.character(ascii_art)
        log_message(
          "Generated ASCII art logo successfully",
          message_type = "success",
          verbose = verbose
        )
        if (verbose) {
          print(ascii_art)
        }
      },
      error = function(e) {
        log_message(
          "ASCII art logo generation failed, using simple logo",
          message_type = "warning",
          verbose = verbose
        )
      }
    )
  }

  if (is.null(ascii_lines)) {
    ascii_lines <- c(
      paste0(
        "  ", paste0(rep("*", nchar(pkg_name) + 4))
      ),
      paste0(
        "  * ", pkg_name, " *"
      ),
      paste0(
        "  ", paste0(rep("*", nchar(pkg_name) + 4))
      )
    )
  }

  file_content <- generate_content(
    pkg_name = pkg_name,
    title = title,
    pkg_description = pkg_description,
    author_name = author_name,
    author_email = author_email,
    github_url = github_url,
    ascii_lines = ascii_lines,
    colors = colors,
    src_exist = dir.exists("src")
  )

  output_file <- file.path(
    paste0("R/", pkg_name, "-package.R")
  )
  writeLines(file_content, output_file)
  log_message(
    "Written to {.file {output_file}}",
    message_type = "success",
    verbose = verbose
  )
  check_dependencies(desc_file, verbose)
  check_pkgdown(pkgdown_file, pkg_name, verbose)

  invisible(file_content)
}

generate_content <- function(
    pkg_name,
    title,
    pkg_description,
    author_name,
    author_email,
    github_url,
    ascii_lines,
    colors,
    src_exist = FALSE) {
  ascii_with_numbers <- add_ascii_numbers(
    ascii_lines,
    length(colors)
  )

  use_dynlib_line <- if (src_exist) {
    c(
      paste0("#' @useDynLib ", pkg_name),
      "#'"
    )
  } else {
    NULL
  }

  content <- c(
    "# -*- coding: utf-8 -*-",
    "",
    paste0("#' @title ", title),
    "#'",
    use_dynlib_line,
    "#' @description",
    paste0("#' ", pkg_description),
    "#'",
    paste0(
      "#' @author ", author_name, " (Maintainer), \\email{", author_email, "}"
    ),
    "#'",
    paste0("#' @source \\url{", github_url, "}"),
    "#'",
    "#' @md",
    "#' @docType package",
    paste0("#' @name ", pkg_name, "-package"),
    "\"_PACKAGE\"",
    "",
    paste0("#' @title The logo of ", pkg_name),
    "#'",
    "#' @description",
    paste0(
      "#' The ", pkg_name, " logo, using ASCII or Unicode characters"
    ),
    "#' Use [cli::ansi_strip] to get rid of the colors.",
    "#'",
    "#' @md",
    "#' @param unicode Unicode symbols on UTF-8 platforms.",
    "#' Default is [cli::is_utf8_output].",
    "#'",
    "#' @return",
    paste0(
      "#' A character vector with class `", tolower(pkg_name), "_logo`."
    ),
    "#'",
    "#' @references",
    "#' \\url{https://github.com/tidyverse/tidyverse/blob/main/R/logo.R}",
    "#'",
    "#' @export",
    "#' @examples",
    paste0("#' ", tolower(pkg_name), "_logo()"),
    paste0(
      tolower(pkg_name),
      "_logo <- function(unicode = cli::is_utf8_output()) {"
    ),
    "  logo <- c(",
    paste0(
      "    \"", ascii_with_numbers, "\""
    ),
    "  )",
    "",
    generate_hexa(length(colors), colors),
    "",
    "  col_hexa <- mapply(",
    "    function(x, y) cli::make_ansi_style(y)(x),",
    "    hexa, cols,",
    "    SIMPLIFY = FALSE",
    "  )",
    "",
    paste0(
      "  for (i in 0:", length(colors) - 1, ") {"
    ),
    "    pat <- paste0(\"\\\\b\", i, \"\\\\b\")",
    "    logo <- sub(pat, col_hexa[[i + 1]], logo)",
    "  }",
    "",
    "  structure(",
    "    cli::col_blue(logo),",
    paste0(
      "    class = \"", tolower(pkg_name), "_logo\""
    ),
    "  )",
    "}",
    "",
    "#' @title Print logo",
    "#'",
    "#' @param x Input information.",
    "#' @param ... Other parameters.",
    "#'",
    "#' @return Print the ASCII logo",
    "#'",
    paste0(
      "#' @method print ", tolower(pkg_name), "_logo"
    ),
    "#'",
    "#' @export",
    paste0(
      "print.", tolower(pkg_name), "_logo <- function(x, ...) {"
    ),
    "  cat(x, ..., sep = \"\\n\")",
    "  invisible(x)",
    "}",
    "",
    ".onAttach <- function(libname, pkgname) {",
    "  verbose <- thisutils::get_verbose()",
    "  if (isTRUE(verbose)) {",
    "    version <- utils::packageVersion(pkgname)",
    "    date <- utils::packageDate(pkgname)",
    "    url <- utils::packageDescription(",
    "      pkgname, fields = \"URL\"",
    "    )",
    "",
    "    msg <- paste0(",
    "      cli::col_grey(strrep(\"-\", 60)),",
    "      \"\\n\",",
    "      cli::col_blue(\"Version: \", version, \" (\", date, \" update)\"),",
    "      \"\\n\",",
    "      cli::col_blue(\"Website: \", cli::style_italic(url)),",
    "      \"\\n\\n\",",
    "      cli::col_grey(\"This message can be suppressed by:\"),",
    "      \"\\n\",",
    paste0(
      "      cli::col_grey(\"  suppressPackageStartupMessages(library(",
      pkg_name, "))\"),"
    ),
    "      \"\\n\",",
    paste0(
      "      cli::col_grey(\"  or options(log_message.verbose = FALSE)\"),"
    ),
    "      \"\\n\",",
    "      cli::col_grey(strrep(\"-\", 60))",
    "    )",
    "",
    paste0(
      "    packageStartupMessage(",
      tolower(pkg_name), "_logo())"
    ),
    "    packageStartupMessage(msg)",
    "  }",
    "}"
  )

  content
}

add_ascii_numbers <- function(
    ascii_lines,
    num_colors) {
  if (length(ascii_lines) == 0) {
    return("")
  }

  top_numbers <- "          0          1        2             3     4"
  bottom_numbers <- "      5               6      7        8          9"

  all_lines <- c(top_numbers, ascii_lines, bottom_numbers)
  result <- paste(all_lines, collapse = "\n")

  result
}

generate_hexa <- function(
    num_colors,
    colors) {
  symbols <- rep(c("*", ".", "o"), length.out = num_colors)

  code <- c(
    paste0(
      "  hexa <- c(", paste0("\"", symbols, "\"", collapse = ", "), ")"
    ),
    "  if (unicode) {",
    "    hexa <- c(\"*\" = \"\\u2b22\", \"o\" = \"\\u2b21\", \".\" = \".\")[hexa]",
    "  }",
    "",
    "  cols <- c(",
    paste0(
      "    ",
      paste0(
        "\"", colors[1:(num_colors / 2)], "\"",
        collapse = ", "
      ), ","
    ),
    paste0(
      "    ",
      paste0(
        "\"", colors[(num_colors / 2 + 1):num_colors], "\"",
        collapse = ", "
      )
    ),
    "  )"
  )

  code
}

read_description <- function(
    desc_file,
    verbose) {
  if (!file.exists(desc_file)) {
    log_message(
      "DESCRIPTION file not found",
      message_type = "error"
    )
  }

  log_message(
    "Reading package information from {.file {desc_file}}",
    verbose = verbose
  )

  desc_lines <- readLines(desc_file, warn = FALSE)
  desc_content <- paste(desc_lines, collapse = "\n")

  package_match <- regexpr(
    "Package:\\s*([^\\n]+)",
    desc_content,
    perl = TRUE
  )
  package_name <- if (package_match != -1) {
    trimws(
      sub(
        "Package:\\s*", "",
        regmatches(desc_content, package_match)
      )
    )
  } else {
    NULL
  }

  title_match <- regexpr(
    "Title:\\s*([^\\n]+)",
    desc_content,
    perl = TRUE
  )
  title <- if (title_match != -1) {
    trimws(
      sub(
        "Title:\\s*", "", regmatches(desc_content, title_match)
      )
    )
  } else {
    NULL
  }

  desc_match <- regexpr(
    "Description:\\s*([^\\n]+)",
    desc_content,
    perl = TRUE
  )
  description <- if (desc_match != -1) {
    trimws(
      sub(
        "Description:\\s*", "",
        regmatches(desc_content, desc_match)
      )
    )
  } else {
    title
  }

  authors_pattern <- "Authors@R:\\s*\\n?\\s*person\\([^)]*name\\s*=\\s*[\"']([^\"']+)[\"'][^)]*email\\s*=\\s*[\"']([^\"']+)[\"']"
  authors_match <- regexpr(
    authors_pattern,
    desc_content,
    perl = TRUE
  )

  if (authors_match != -1) {
    authors_text <- regmatches(
      desc_content,
      authors_match
    )

    name_pattern <- "name\\s*=\\s*[\"']([^\"']+)[\"']"
    name_match <- regexpr(
      name_pattern,
      authors_text,
      perl = TRUE
    )
    author_name <- if (name_match != -1) {
      sub(".*name\\s*=\\s*[\"']([^\"']+)[\"'].*", "\\1", authors_text)
    } else {
      "Author"
    }

    email_match <- regexpr(
      "email\\s*=\\s*[\"']([^\"']+)[\"']",
      authors_text,
      perl = TRUE
    )
    author_email <- if (email_match != -1) {
      sub(".*email\\s*=\\s*[\"']([^\"']+)[\"'].*", "\\1", authors_text)
    } else {
      "author@example.com"
    }
  } else {
    maintainer_match <- regexpr(
      "Maintainer:\\s*([^<]+)<([^>]+)>",
      desc_content,
      perl = TRUE
    )

    if (maintainer_match != -1) {
      maintainer_text <- regmatches(
        desc_content, maintainer_match
      )
      author_name <- trimws(
        sub(
          "Maintainer:\\s*([^<]+)<.*", "\\1",
          maintainer_text
        )
      )
      author_email <- sub(
        ".*<([^>]+)>.*", "\\1",
        maintainer_text
      )
    } else {
      author_name <- "Author"
      author_email <- "author@example.com"
    }
  }

  url_match <- regexpr(
    "URL:\\s*([^\\n]+)",
    desc_content,
    perl = TRUE
  )
  bug_reports_match <- regexpr(
    "BugReports:\\s*([^\\n]+)",
    desc_content,
    perl = TRUE
  )

  github_url <- NULL
  if (url_match != -1) {
    url <- trimws(
      sub("URL:\\s*", "", regmatches(desc_content, url_match))
    )
    github_url <- url
  } else if (bug_reports_match != -1) {
    bug_url <- trimws(
      sub("BugReports:\\s*", "", regmatches(desc_content, bug_reports_match))
    )
    github_url <- sub("/issues$", "", bug_url)
  }

  if (is.null(github_url) && !is.null(package_name)) {
    github_url <- paste0("https://github.com/username/", package_name)
  }

  log_message(
    "Package: {.pkg {package_name}}\n",
    "Title: {.pkg {title}}\n",
    "Description: {.pkg {description}}\n",
    "Author: {.pkg {author_name}}, {.email {author_email}}\n",
    "GitHub URL: {.url {github_url}}\n",
    message_type = "success",
    verbose = verbose
  )

  list(
    Package = package_name,
    Title = title,
    Description = description,
    Author = author_name,
    Email = author_email,
    GitHub_URL = github_url
  )
}

check_dependencies <- function(desc_file, verbose) {
  if (!file.exists(desc_file)) {
    log_message(
      "{.file {desc_file}} not found",
      message_type = "error"
    )
    return()
  }

  desc_content <- readLines(desc_file, warn = FALSE)

  imports_bounds <- find_section(desc_content, "Imports")
  if (is.null(imports_bounds)) {
    log_message(
      "No {.cls Imports} section found in {.file {desc_file}}",
      message_type = "warning",
      verbose = verbose
    )
    return()
  }

  imports_packages <- parse_packages(
    desc_content,
    imports_bounds$start,
    imports_bounds$end
  )
  imports_names <- sapply(imports_packages, `[[`, "name")
  has_cli <- "cli" %in% imports_names

  modified <- FALSE

  if (!has_cli) {
    log_message(
      "Adding {.pkg cli} dependency to {.file {desc_file}}",
      verbose = verbose
    )
    imports_packages <- c(
      imports_packages,
      list(list(name = "cli", full = "cli"))
    )
    modified <- TRUE
  }

  imports_lines <- format_packages(imports_packages)
  new_imports <- c("Imports:", imports_lines)

  suggests_bounds <- find_section(desc_content, "Suggests")
  suggests_packages <- NULL
  if (!is.null(suggests_bounds)) {
    suggests_packages <- parse_packages(
      desc_content,
      suggests_bounds$start,
      suggests_bounds$end
    )
  }

  if (imports_bounds$end < length(desc_content)) {
    after_imports <- desc_content[(imports_bounds$end + 1):length(desc_content)]
  } else {
    after_imports <- character(0)
  }

  if (!is.null(suggests_bounds) && suggests_bounds$start > imports_bounds$end) {
    suggests_rel_start <- suggests_bounds$start - imports_bounds$end
    suggests_rel_end <- suggests_bounds$end - imports_bounds$end

    if (suggests_rel_start <= length(after_imports)) {
      if (suggests_rel_end < length(after_imports)) {
        after_imports <- c(
          if (suggests_rel_start > 1) after_imports[1:(suggests_rel_start - 1)] else character(0),
          after_imports[(suggests_rel_end + 1):length(after_imports)]
        )
      } else {
        after_imports <- if (suggests_rel_start > 1) {
          after_imports[1:(suggests_rel_start - 1)]
        } else {
          character(0)
        }
      }
    }
  }

  new_content <- desc_content[1:(imports_bounds$start - 1)]
  new_content <- c(new_content, new_imports)

  if (!is.null(suggests_packages) && length(suggests_packages) > 0) {
    suggests_lines <- format_packages(suggests_packages)
    new_suggests <- c("Suggests:", suggests_lines)
    new_content <- c(new_content, new_suggests)
  }

  new_content <- c(new_content, after_imports)
  desc_content <- new_content

  writeLines(desc_content, desc_file)

  if (modified) {
    log_message(
      "Successfully added {.pkg cli} and sorted dependencies in {.file {desc_file}}",
      message_type = "success",
      verbose = verbose
    )
  } else {
    log_message(
      "{.pkg cli} already present, dependencies sorted in {.file {desc_file}}",
      verbose = verbose
    )
  }
}

check_pkgdown <- function(pkgdown_file, pkg_name, verbose) {
  if (!file.exists(pkgdown_file)) {
    log_message(
      "{.file {pkgdown_file}} not found",
      message_type = "warning",
      verbose = verbose
    )
    return()
  }

  pkgdown_content <- readLines(pkgdown_file, warn = FALSE)

  reference_start <- which(grepl("^reference:", pkgdown_content))
  if (length(reference_start) == 0) {
    log_message(
      "No {.cls reference} section found in {.file {pkgdown_file}}",
      message_type = "warning",
      verbose = verbose
    )
    return()
  }
  reference_start <- reference_start[1]

  reference_end <- length(pkgdown_content)
  if (reference_start < length(pkgdown_content)) {
    for (i in (reference_start + 1):length(pkgdown_content)) {
      if (grepl("^[A-Za-z]", pkgdown_content[i]) && !grepl("^\\s", pkgdown_content[i])) {
        reference_end <- i - 1
        break
      }
    }
  }

  target_entries <- c(
    paste0("  - ", pkg_name, "-package"),
    paste0("  - ", tolower(pkg_name), "_logo"),
    paste0("  - print.", tolower(pkg_name), "_logo")
  )

  ref_lines <- if (reference_start + 1 <= reference_end) {
    pkgdown_content[(reference_start + 1):reference_end]
  } else {
    character(0)
  }
  subtitle_in_ref <- which(grepl('^- subtitle: "Package overview"$', ref_lines))

  modified <- FALSE

  if (length(subtitle_in_ref) == 0) {
    insert_block <- c(
      "",
      "- subtitle: \"Package overview\"",
      "- contents:",
      target_entries
    )

    pkgdown_content <- c(
      pkgdown_content[1:reference_start],
      insert_block,
      pkgdown_content[(reference_start + 1):length(pkgdown_content)]
    )
    modified <- TRUE
  } else {
    subtitle_global <- reference_start + subtitle_in_ref[1]

    next_subtitle_in_ref <- which(
      grepl("^- subtitle:", ref_lines) & (seq_along(ref_lines) > subtitle_in_ref[1])
    )
    block_end <- if (length(next_subtitle_in_ref) > 0) {
      reference_start + next_subtitle_in_ref[1] - 1
    } else {
      reference_end
    }

    if (subtitle_global + 1 <= block_end) {
      block_lines <- pkgdown_content[(subtitle_global + 1):block_end]
    } else {
      block_lines <- character(0)
    }

    contents_rel <- which(grepl("^- contents:$", block_lines))

    if (length(contents_rel) == 0) {
      pkgdown_content <- append(
        pkgdown_content,
        values = c("- contents:", target_entries),
        after = subtitle_global
      )
      modified <- TRUE
    } else {
      contents_global <- subtitle_global + contents_rel[1]

      item_start <- contents_global + 1
      item_end <- contents_global
      while (item_end + 1 <= block_end && grepl("^  - ", pkgdown_content[item_end + 1])) {
        item_end <- item_end + 1
      }

      existing_items <- if (item_end >= item_start) {
        pkgdown_content[item_start:item_end]
      } else {
        character(0)
      }

      other_items <- existing_items[!(existing_items %in% target_entries)]
      new_items <- c(target_entries, other_items)

      if (!identical(existing_items, new_items)) {
        if (item_end >= item_start) {
          pkgdown_content <- c(
            pkgdown_content[1:(item_start - 1)],
            new_items,
            pkgdown_content[(item_end + 1):length(pkgdown_content)]
          )
        } else {
          pkgdown_content <- append(
            pkgdown_content,
            values = new_items,
            after = contents_global
          )
        }
        modified <- TRUE
      }
    }
  }

  if (modified) {
    log_message(
      "Updating {.file {pkgdown_file}} with package overview section",
      message_type = "running",
      verbose = verbose
    )
    writeLines(pkgdown_content, pkgdown_file)
    log_message(
      "Successfully updated {.file {pkgdown_file}}",
      message_type = "success",
      verbose = verbose
    )
  } else {
    log_message(
      "Package overview section already present in {.file {pkgdown_file}}",
      verbose = verbose
    )
  }
}

find_section <- function(content, section_name) {
  section_start <- which(grepl(paste0("^", section_name, ":"), content))
  if (length(section_start) == 0) {
    return(NULL)
  }
  section_start <- section_start[1]

  section_end <- section_start
  if (section_start < length(content)) {
    for (i in (section_start + 1):length(content)) {
      if (grepl("^[A-Za-z]", content[i]) && !grepl("^\\s", content[i])) {
        section_end <- i - 1
        break
      }
      if (i == length(content)) {
        section_end <- length(content)
      }
    }
  }

  list(start = section_start, end = section_end)
}

parse_packages <- function(content, start, end) {
  section_lines <- content[start:end]
  section_text <- paste(section_lines, collapse = "\n")
  section_text <- sub("^[A-Za-z]+:\\s*", "", section_text)

  packages_raw <- strsplit(section_text, ",")[[1]]
  packages_raw <- trimws(packages_raw)
  packages_raw <- packages_raw[packages_raw != ""]

  packages <- lapply(packages_raw, function(pkg) {
    pkg_name <- sub("\\s*\\(.*\\)\\s*$", "", pkg)
    pkg_name <- trimws(pkg_name)
    list(name = pkg_name, full = pkg)
  })

  packages
}

format_packages <- function(packages) {
  packages <- packages[order(tolower(sapply(packages, `[[`, "name")))]

  n <- length(packages)
  lines <- character(n)
  for (i in seq_len(n)) {
    if (i < n) {
      lines[i] <- paste0("    ", packages[[i]]$full, ",")
    } else {
      lines[i] <- paste0("    ", packages[[i]]$full)
    }
  }

  lines
}

test_that("non-interactive attach does not emit startup messages", {
  testthat::local_mocked_bindings(
    interactive = function() FALSE,
    .package = "base"
  )

  expect_silent(thisutils:::.onAttach(NULL, "thisutils"))
})

test_that("interactive attach honors verbose = FALSE", {
  old <- options(log_message.verbose = FALSE)
  on.exit(options(old), add = TRUE)

  testthat::local_mocked_bindings(
    interactive = function() TRUE,
    .package = "base"
  )

  expect_silent(thisutils:::.onAttach(NULL, "thisutils"))
})

#' @title Generate a simulated sparse matrix
#'
#' @description
#' This function generates a sparse matrix with a specified number of rows and columns,
#' a given sparsity level, and a distribution function for the non-zero values.
#'
#' @md
#' @param nrow Number of rows in the matrix.
#' @param ncol Number of columns in the matrix.
#' @param sparsity Proportion of zero elements (sparsity level).
#' Default is `0.95`, meaning 95% of elements are zero (5% are non-zero).
#' @param distribution_fun Function to generate non-zero values.
#' @param decimal Controls the number of decimal places in the generated values.
#' If set to `0`, values will be integers.
#' When `decimal` > 0, random decimal parts are uniformly distributed across the full range.
#' Default is `0`.
#' @param seed Random seed for reproducibility.
#'
#' @return
#' A sparse matrix of class "dgCMatrix".
#'
#' @export
#'
#' @examples
#' simulate_sparse_matrix(1000, 500) |>
#'   check_sparsity()
#'
#' simulate_sparse_matrix(10, 10, decimal = 1)
#' simulate_sparse_matrix(10, 10, decimal = 5)
simulate_sparse_matrix <- function(
    nrow,
    ncol,
    sparsity = 0.95,
    distribution_fun = function(n) stats::rpois(n, lambda = 0.5) + 1,
    decimal = 0,
    seed = 1) {
  set.seed(seed)

  nnz <- round(nrow * ncol * (1 - sparsity))

  total_positions <- nrow * ncol
  if (nnz > total_positions) {
    nnz <- total_positions
  }

  positions <- sample(total_positions, nnz, replace = FALSE)

  i <- ((positions - 1) %% nrow) + 1
  j <- ((positions - 1) %/% nrow) + 1

  x <- distribution_fun(nnz)

  if (decimal > 0) {
    decimal_part <- stats::runif(nnz, min = 0, max = 1)
    x <- x + decimal_part
  }

  x <- round(x, decimal)

  Matrix::sparseMatrix(
    i = i,
    j = j,
    x = x,
    dims = c(nrow, ncol),
    dimnames = list(
      paste0("row_", 1:nrow),
      paste0("col_", 1:ncol)
    )
  )
}

#' @title Check sparsity of matrix
#'
#' @param x A matrix.
#'
#' @return
#' Sparsity of matrix.
#'
#' @export
check_sparsity <- function(x) {
  total_counts <- prod(dim(x))
  
  if (inherits(x, "sparseMatrix")) {
    non_zero_count <- Matrix::nnzero(x)
  } else {
    non_zero_count <- sum(x != 0)
  }

  sparsity_ratio <- non_zero_count / total_counts

  1 - sparsity_ratio
}

.rmse <- function(true, pred) {
  sqrt(mean((true - pred)^2))
}

.sse <- function(y_true, y_pred) {
  sum((y_true - y_pred)**2)
}

.rse <- function(y_true, y_pred) {
  .sse(y_true, y_pred) / .sse(y_true, mean(y_true))
}

#' @title Coefficient of determination (\eqn{R^2})
#'
#' @param y_true A numeric vector with ground truth values.
#' @param y_pred A numeric vector with predicted values.
#'
#' @return
#' The \eqn{R^2} value.
#'
#' @export
#'
#' @examples
#' y <- rnorm(100)
#' y_pred <- y + rnorm(100, sd = 0.5)
#' r_square(y, y_pred)
r_square <- function(y_true, y_pred) {
  1 - .rse(y_true, y_pred)
}

#' @title Normalize numeric vector
#'
#' @md
#' @param x Input numeric vector.
#' @param method Method used for normalization.
#' @param na_rm Whether to remove `NA` values,
#' and if setting `TRUE`, using `0` instead.
#' Default is `TRUE`.
#' @param ... Parameters for other methods.
#'
#' @return
#' Normalized numeric vector.
#'
#' @export
#'
#' @examples
#' x <- c(runif(2), NA, -runif(2))
#' x
#' normalization(x, method = "max_min")
#' normalization(x, method = "maximum")
#' normalization(x, method = "sum")
#' normalization(x, method = "softmax")
#' normalization(x, method = "z_score")
#' normalization(x, method = "mad")
#' normalization(x, method = "unit_vector")
#' normalization(x, method = "unit_vector", na_rm = FALSE)
normalization <- function(
    x,
    method = "max_min",
    na_rm = TRUE,
    ...) {
  method <- match.arg(
    method,
    c(
      "max_min",
      "maximum",
      "sum",
      "softmax",
      "z_score",
      "mad",
      "unit_vector",
      "robust_scale"
    )
  )
  na_index <- which(is.na(x))
  x_calc <- x
  x_calc[na_index] <- NA
  safe_fill <- function(value = 0) {
    out <- rep_len(value, length(x_calc))
    out[na_index] <- NA_real_
    out
  }
  x <- switch(
    EXPR = method,
    "max_min" = {
      min_x <- min(x_calc, na.rm = TRUE)
      max_x <- max(x_calc, na.rm = TRUE)
      range_x <- max_x - min_x
      if (!is.finite(range_x) || range_x == 0) {
        safe_fill(0)
      } else {
        (x_calc - min_x) / range_x
      }
    },
    "maximum" = {
      max_abs_x <- max(abs(x_calc), na.rm = TRUE)
      if (!is.finite(max_abs_x) || max_abs_x == 0) {
        safe_fill(0)
      } else {
        x_calc / max_abs_x
      }
    },
    "sum" = {
      abs_sum <- sum(abs(x_calc), na.rm = TRUE)
      if (!is.finite(abs_sum) || abs_sum == 0) {
        safe_fill(0)
      } else {
        x_calc / abs_sum
      }
    },
    "softmax" = {
      x_mean <- mean(x_calc, na.rm = TRUE)
      x_sd <- stats::sd(x_calc, na.rm = TRUE)
      if (!is.finite(x_sd) || x_sd == 0) {
        valid_count <- sum(!is.na(x_calc))
        if (valid_count == 0) {
          safe_fill(0)
        } else {
          out <- safe_fill(0)
          out[!is.na(x_calc)] <- 1 / valid_count
          out
        }
      } else {
        temp <- (x_calc - x_mean) / x_sd
        exp_temp <- exp(temp)
        exp_temp / sum(exp_temp, na.rm = TRUE)
      }
    },
    "z_score" = {
      x_sd <- stats::sd(x_calc, na.rm = TRUE)
      if (!is.finite(x_sd) || x_sd == 0) {
        safe_fill(0)
      } else {
        (x_calc - mean(x_calc, na.rm = TRUE)) / x_sd
      }
    },
    "mad" = {
      x_mad <- stats::mad(x_calc, na.rm = TRUE)
      if (!is.finite(x_mad) || x_mad == 0) {
        safe_fill(0)
      } else {
        (x_calc - stats::median(x_calc, na.rm = TRUE)) / x_mad
      }
    },
    "unit_vector" = {
      norm_x <- sqrt(sum(x_calc^2, na.rm = TRUE))
      if (!is.finite(norm_x) || norm_x == 0) {
        safe_fill(0)
      } else {
        x_calc / norm_x
      }
    },
    "robust_scale" = {
      q1 <- stats::quantile(x_calc, 0.25, na.rm = TRUE)
      q3 <- stats::quantile(x_calc, 0.75, na.rm = TRUE)
      iqr <- q3 - q1
      med <- stats::median(x_calc, na.rm = TRUE)
      if (!is.finite(iqr) || iqr == 0) {
        safe_fill(0)
      } else {
        (x_calc - med) / iqr
      }
    }
  )

  if (na_rm) {
    x[na_index] <- 0
  }

  x
}

fc_matrix <- function(matrix) {
  row_means <- if (inherits(matrix, "sparseMatrix")) {
    Matrix::rowMeans(matrix)
  } else {
    rowMeans(matrix)
  }
  matrix / row_means
}

zscore_matrix <- function(matrix, ...) {
  Matrix::t(scale(Matrix::t(matrix), ...))
}

log2fc_matrix <- function(matrix) {
  row_means <- if (inherits(matrix, "sparseMatrix")) {
    Matrix::rowMeans(matrix)
  } else {
    rowMeans(matrix)
  }
  log2(matrix / row_means)
}

#' @title Process matrix
#'
#' @param matrix A matrix.
#' @param method Method to use for processing the matrix.
#' @param ... Other arguments passed to the method.
#'
#' @return A processed matrix.
#'
#' @export
#'
#' @examples
#' m <- simulate_sparse_matrix(10, 10)
#' matrix_process(m, method = "raw")
#' matrix_process(m, method = "zscore")
#' matrix_process(m, method = "fc")
#' matrix_process(m, method = "log2fc")
#' matrix_process(m, method = "log1p")
#' m <- as_matrix(m)
#' matrix_process(m, method = function(x) x / rowMeans(x))
matrix_process <- function(
    matrix,
    method = c("raw", "zscore", "fc", "log2fc", "log1p"),
    ...) {
  if (is.function(method)) {
    matrix_processed <- method(matrix, ...)
  } else {
    method <- match.arg(method)
    matrix_processed <- switch(
      EXPR = method,
      "raw" = {
        matrix
      },
      "fc" = {
        fc_matrix(matrix, ...)
      },
      "zscore" = {
        zscore_matrix(matrix, ...)
      },
      "log2fc" = {
        log2fc_matrix(matrix)
      },
      "log1p" = {
        log1p(matrix)
      }
    )
  }

  if (!identical(dim(matrix_processed), dim(matrix))) {
    log_message(
      "The dimensions of the matrix are changed after processing",
      message_type = "error"
    )
  }
  return(matrix_processed)
}

test_that("simulate_sparse_matrix creates correct dimensions", {
  m <- simulate_sparse_matrix(100, 50)
  expect_equal(nrow(m), 100)
  expect_equal(ncol(m), 50)
  expect_true(inherits(m, "sparseMatrix"))
})

test_that("simulate_sparse_matrix respects sparsity", {
  m <- simulate_sparse_matrix(100, 100, sparsity = 0.9)
  sp <- check_sparsity(m)
  expect_true(sp > 0.85 && sp < 0.95)
})

test_that("simulate_sparse_matrix with decimal", {
  m <- simulate_sparse_matrix(10, 10, decimal = 2)
  nonzero <- m@x
  expect_true(all(nonzero == round(nonzero, 2)))
})

test_that("check_sparsity works for dense matrix", {
  m <- matrix(0, 10, 10)
  m[1, 1] <- 1
  expect_equal(check_sparsity(m), 0.99)
})

test_that("check_sparsity works for sparse matrix", {
  m <- simulate_sparse_matrix(100, 100, sparsity = 0.95)
  sp <- check_sparsity(m)
  expect_true(sp > 0.9)
})

test_that("r_square computes correctly", {
  y <- 1:10
  y_pred <- y
  expect_equal(r_square(y, y_pred), 1)
})

test_that("r_square handles imperfect predictions", {
  set.seed(42)
  y <- rnorm(100)
  y_pred <- y + rnorm(100, sd = 0.5)
  r2 <- r_square(y, y_pred)
  expect_true(r2 > 0 && r2 < 1)
})

test_that("normalization max_min works", {
  x <- c(1, 2, 3, 4, 5)
  result <- normalization(x, method = "max_min")
  expect_equal(min(result), 0)
  expect_equal(max(result), 1)
})

test_that("normalization z_score works", {
  x <- c(1, 2, 3, 4, 5)
  result <- normalization(x, method = "z_score")
  expect_equal(mean(result), 0, tolerance = 1e-10)
  expect_equal(sd(result), 1, tolerance = 1e-10)
})

test_that("normalization handles NA with na_rm = TRUE", {
  x <- c(1, NA, 3, 4, 5)
  result <- normalization(x, method = "max_min")
  expect_equal(result[2], 0)
  expect_false(is.na(result[2]))
})

test_that("normalization handles NA with na_rm = FALSE", {
  x <- c(1, NA, 3, 4, 5)
  result <- normalization(x, method = "max_min", na_rm = FALSE)
  expect_true(is.na(result[2]))
})

test_that("normalization NA handling does not pollute calculations", {
  x <- c(10, NA, 30, 40, 50)
  result <- normalization(x, method = "max_min", na_rm = FALSE)
  expect_equal(result[1], 0, tolerance = 1e-10)
  expect_equal(result[5], 1, tolerance = 1e-10)
  expect_true(is.na(result[2]))
})

test_that("normalization softmax works", {
  x <- c(1, 2, 3)
  result <- normalization(x, method = "softmax")
  expect_equal(sum(result), 1, tolerance = 1e-10)
  expect_true(all(result > 0))
})

test_that("normalization mad works", {
  x <- c(1, 2, 3, 4, 5)
  result <- normalization(x, method = "mad")
  expect_equal(
    result,
    (x - median(x)) / mad(x),
    tolerance = 1e-10
  )
})

test_that("normalization unit_vector works", {
  x <- c(3, 4)
  result <- normalization(x, method = "unit_vector")
  expect_equal(sqrt(sum(result^2)), 1, tolerance = 1e-10)
})

test_that("normalization robust_scale works", {
  x <- c(1, 2, 3, 4, 5)
  result <- normalization(x, method = "robust_scale")
  q1 <- quantile(x, 0.25)
  q3 <- quantile(x, 0.75)
  iqr <- q3 - q1
  med <- median(x)
  expected <- as.numeric((x - med) / iqr)
  expect_equal(result, expected, tolerance = 1e-10)
})

test_that("normalization errors on invalid method", {
  expect_error(normalization(1:5, method = "invalid"))
})

test_that("normalization handles degenerate inputs without NaN", {
  expect_equal(normalization(c(1, 1, 1), method = "max_min"), c(0, 0, 0))
  expect_equal(normalization(c(0, 0, 0), method = "sum"), c(0, 0, 0))
  expect_equal(normalization(c(1, 1, 1), method = "z_score"), c(0, 0, 0))
  expect_equal(
    normalization(c(1, 1, NA), method = "softmax", na_rm = FALSE),
    c(0.5, 0.5, NA)
  )
})

test_that("matrix_process works with string methods", {
  m <- simulate_sparse_matrix(10, 10)
  expect_no_error(matrix_process(m, method = "raw"))
  expect_no_error(matrix_process(m, method = "log1p"))
})

test_that("matrix_process works with custom function", {
  m <- matrix(1:20, nrow = 4)
  result <- matrix_process(m, method = function(x) x * 2)
  expect_equal(result, m * 2)
})

test_that("matrix_process errors on invalid string method", {
  m <- matrix(1:20, nrow = 4)
  expect_error(matrix_process(m, method = "invalid"))
})

test_that("matrix_process errors when dimensions change", {
  m <- matrix(1:20, nrow = 4)
  expect_error(
    matrix_process(m, method = function(x) x[1:2, ]),
    "dimensions"
  )
})

#' @title Value selection operator
#'
#' @description
#' This operator returns the left side if it's not `NULL`,
#' otherwise it returns the right side.
#'
#' @md
#' @param a The left side value to check.
#' @param b The right side value to use if `a` is `NULL`.
#'
#' @return `a` if it is not `NULL`, otherwise `b`.
#'
#' @export
#'
#' @examples
#' NULL %ss% 10
#' 5 %ss% 10
`%ss%` <- function(a, b) {
  if (is.null(a)) {
    b
  } else {
    a
  }
}

#' @title Get a function from a namespace
#'
#' @md
#' @param pkg The name of the package.
#' @param fun The name of the function.
#'
#' @return Function.
#'
#' @export
get_namespace_fun <- function(pkg, fun) {
  ns <- getNamespace(pkg)
  if (exists(fun, envir = ns, inherits = FALSE)) {
    get(fun, envir = ns, inherits = FALSE)
  } else {
    log_message(
      "Function {.val {fun}} not found in {.pkg {pkg}} namespace",
      message_type = "warning"
    )
    return(NULL)
  }
}

#' @title Invoke a function with a list of arguments
#'
#' @md
#' @param .fn A function, or function name as a string.
#' @param .args A list of arguments.
#' @param ... Other arguments passed to the function.
#' @param .env Environment in which to evaluate the call.
#' This will be most useful if `.fn` is a string,
#' or the function has side-effects.
#'
#' @export
#'
#' @examples
#' f <- function(x, y) {
#'   x + y
#' }
#' invoke_fun(f, list(x = 1, y = 2))
#' invoke_fun("f", list(x = 1, y = 2))
#' invoke_fun("f", x = 1, y = 2)
invoke_fun <- function(
    .fn,
    .args = list(),
    ...,
    .env = rlang::caller_env()) {
  args <- c(.args, list(...))
  if (!length(args)) {
    if (rlang::is_scalar_character(.fn)) {
      .fn <- rlang::env_get(.env, .fn, inherit = TRUE)
    }
    call <- rlang::call2(.fn, !!!args)
    return(rlang::eval_bare(call, .env))
  }
  fn_nm <- ".fn"
  buried_nms <- paste0("", seq_along(args))
  buried_args <- rlang::set_names(args, buried_nms)
  .env <- rlang::env(.env, !!!buried_args)
  args <- rlang::set_names(buried_nms, names(args))
  args <- rlang::syms(args)
  if (rlang::is_function(.fn)) {
    `:=` <- rlang::`:=`
    rlang::env_bind(.env, `:=`(!!fn_nm, .fn))
    .fn <- fn_nm
  }
  call <- rlang::call2(.fn, !!!args)
  rlang::eval_bare(call, .env)
}

#' @title Capitalize the first letter of each word
#'
#' @param x A vector of character strings to be capitalized.
#' @param force_tolower Whether to force the remaining letters to be lowercase.
#' @export
#'
#' @examples
#' x <- c(
#'   "hello world",
#'   "hello World"
#' )
#' capitalize(x)
capitalize <- function(x, force_tolower = FALSE) {
  if (is.null(x)) {
    return(NULL)
  }
  if (inherits(x, "factor")) {
    x <- as.character(x)
  }
  if (!inherits(x, "character")) {
    log_message(
      "x must be the type of character.",
      message_type = "error"
    )
  }
  if (isTRUE(force_tolower)) {
    x <- paste(
      toupper(substr(x, 1, 1)),
      tolower(substr(x, 2, nchar(x))),
      sep = ""
    )
  } else {
    first_word <- vapply(strsplit(x, "\\s|-"), function(s) s[1], character(1))
    index <- which(first_word == tolower(first_word))
    x[index] <- paste(
      toupper(substr(x[index], 1, 1)),
      substr(x[index], 2, nchar(x[index])),
      sep = ""
    )
  }
  return(x)
}

#' @title Unnest a list-column
#'
#' @description
#' Implement similar functions to the `tidyr::unnest` function.
#'
#' @md
#' @param data A data frame.
#' @param cols Columns to unnest.
#' @param keep_empty By default, you get one row of output for each element of the list your unchopping/unnesting.
#' This means that if there's a size-0 element (like `NULL` or an empty data frame),
#' that entire row will be dropped from the output.
#' If you want to preserve all rows,
#' use `keep_empty = TRUE` to replace size-0 elements with a single row of missing values.
#' @export
#' @examples
#' data <- data.frame(
#'   id = 1:3,
#'   x = c("a", "b", "c"),
#'   stringsAsFactors = FALSE
#' )
#' data$data <- list(
#'   c(1, 2),
#'   c(3, 4, 5),
#'   c(6)
#' )
#' unnest_fun(data, cols = "data")
#'
#' data2 <- data.frame(
#'   id = 1:3,
#'   x = c("a", "b", "c"),
#'   stringsAsFactors = FALSE
#' )
#' data2$data <- list(
#'   c(1, 2),
#'   numeric(0),
#'   c(6)
#' )
#' unnest_fun(data2, cols = "data")
#' unnest_fun(data2, cols = "data", keep_empty = TRUE)
unnest_fun <- function(
    data,
    cols,
    keep_empty = FALSE) {
  if (nrow(data) == 0 || length(cols) == 0) {
    return(data)
  }
  for (col in cols) {
    col_expand <- unlist(data[[col]])
    expand_times <- vapply(data[[col]], length, integer(1))
    if (isTRUE(keep_empty)) {
      data[[col]][expand_times == 0] <- NA
      col_expand <- unlist(data[[col]])
      expand_times[expand_times == 0] <- 1
    }
    data <- data[rep(seq_len(nrow(data)), times = expand_times), ]
    data[, col] <- col_expand
  }
  rownames(data) <- NULL
  return(data)
}

#' @title Remove and normalize spaces
#'
#' @md
#' @param x A vector of character strings.
#' @param trim_start Whether to remove leading spaces before the first word.
#' Default is `TRUE`.
#' @param trim_end Whether to remove trailing spaces after the last word.
#' Default is `FALSE`.
#' @param collapse_multiple Whether to collapse multiple consecutive spaces between words into a single space.
#' Default is `TRUE`.
#' @param preserve_newlines Whether to preserve newline characters when collapsing spaces.
#' Default is `TRUE`.
#'
#' @return A character vector with spaces normalized according to the specified parameters.
#' @export
#'
#' @examples
#' x <- c(
#'   " hello  world ",
#'   "  test   case  ",
#'   "no space",
#'   "   multiple   spaces   "
#' )
#' remove_space(x)
#' remove_space(x, trim_start = FALSE)
#' remove_space(x, trim_end = TRUE)
#' remove_space(x, collapse_multiple = FALSE)
#' remove_space(
#'   x,
#'   trim_start = FALSE,
#'   trim_end = FALSE,
#'   collapse_multiple = FALSE
#' )
#'
#' # with newlines
#' multiline <- c(
#'   "hello\n\n  world  ",
#'   "  first  \n  second  "
#' )
#' remove_space(multiline)
#' remove_space(multiline, preserve_newlines = FALSE)
remove_space <- function(
    x,
    trim_start = TRUE,
    trim_end = FALSE,
    collapse_multiple = TRUE,
    preserve_newlines = TRUE) {
  if (is.null(x)) {
    return(NULL)
  }
  if (inherits(x, "factor")) {
    x <- as.character(x)
  }
  if (!inherits(x, "character")) {
    log_message(
      "{.arg x} must be the type of character",
      message_type = "error"
    )
  }

  if (!is.logical(trim_start) || length(trim_start) != 1) {
    log_message(
      "{.arg trim_start} must be a single logical value",
      message_type = "error"
    )
  }
  if (!is.logical(trim_end) || length(trim_end) != 1) {
    log_message(
      "{.arg trim_end} must be a single logical value",
      message_type = "error"
    )
  }
  if (!is.logical(collapse_multiple) || length(collapse_multiple) != 1) {
    log_message(
      "{.arg collapse_multiple} must be a single logical value",
      message_type = "error"
    )
  }
  if (!is.logical(preserve_newlines) || length(preserve_newlines) != 1) {
    log_message(
      "{.arg preserve_newlines} must be a single logical value",
      message_type = "error"
    )
  }

  if (trim_start) {
    x <- gsub("^\\s+", "", x)
  }
  if (trim_end) {
    x <- gsub("\\s+$", "", x)
  }

  if (collapse_multiple) {
    if (preserve_newlines) {
      x <- gsub("[ \t\r\f\v]+", " ", x)
      x <- gsub("\\s*\n\\s*", "\n", x)
    } else {
      x <- gsub("\\s+", " ", x)
    }
  }

  return(x)
}

#' @title Try to evaluate an expression a set number of times before failing
#'
#' @description
#' The function is used as a fail-safe if code sometimes works and sometimes doesn't,
#' usually because it depends on a resource that may be temporarily unavailable.
#' It tries to evaluate the expression `max_tries` times.
#' If all the attempts fail, it throws an error;
#' if not, the evaluated expression is returned.
#'
#' @md
#' @param expr The expression to be evaluated.
#' @param max_tries The maximum number of attempts to evaluate the expression before giving up.
#' Default is `5`.
#' @param error_message Additional custom error message to be displayed when an error occurs.
#' @param retry_message Message displayed when a new try to evaluate the expression would be attempted.
#'
#' @return
#' The evaluated expression if successful,
#' otherwise it throws an error if all attempts are unsuccessful.
#' @export
#'
#' @examples
#' f <- function() {
#'   value <- runif(1, min = 0, max = 1)
#'   if (value > 0.5) {
#'     log_message("value is larger than 0.5")
#'     return(value)
#'   } else {
#'     log_message(
#'       "value is smaller than 0.5",
#'       message_type = "error"
#'     )
#'   }
#' }
#' f_evaluated <- try_get(expr = f())
#' print(f_evaluated)
try_get <- function(
    expr,
    max_tries = 5,
    error_message = "",
    retry_message = "Retrying...") {
  out <- simpleError("start")
  ntry <- 0
  while (inherits(out, "error")) {
    ntry <- ntry + 1
    out <- tryCatch(
      expr = eval.parent(substitute(expr)),
      error = function(error) {
        log_message(error, message_type = "warning")
        log_message(error_message, message_type = "warning")
        Sys.sleep(1)
        error
      }
    )
    if (inherits(out, "error") && ntry >= max_tries) {
      log_message(
        out,
        message_type = "error"
      )
    } else {
      if (!inherits(out, "error")) {
        break
      } else {
        log_message(retry_message, message_type = "warning")
      }
    }
  }
  return(out)
}

#' @title Download file from the Internet
#'
#' @md
#' @inheritParams utils::download.file
#' @param methods Methods to be used for downloading files.
#' Can be `"auto"`, `"wget"`, `"libcurl"`, `"curl"`, `"wininet"`, `"internal"`.
#' Default is `"auto"`, which means to try different download methods.
#' @param max_tries Number of tries for each download method.
#' Default is `2`.
#' @param ... Other arguments passed to [utils::download.file].
#'
#' @export
download <- function(
    url,
    destfile,
    methods = c(
      "auto", "wget", "libcurl", "curl", "wininet", "internal"
    ),
    quiet = FALSE,
    ...,
    max_tries = 2) {
  if (missing(url) || missing(destfile)) {
    log_message(
      "{.arg url} and {.arg destfile} must be both provided",
      message_type = "error"
    )
  }
  ntry <- 0
  status <- NULL
  while (is.null(status)) {
    for (method in methods) {
      status <- tryCatch(
        expr = {
          suppressWarnings(
            utils::download.file(
              url = url,
              destfile = destfile,
              method = method,
              quiet = quiet,
              ...
            )
          )
          status <- 1
        }, error = function(error) {
          log_message(error)
          log_message(
            "Failed to download using {.pkg {method}}, from {.url {url}}",
            message_type = "warning"
          )
          Sys.sleep(1)
          NULL
        }
      )
      if (!is.null(status)) {
        break
      }
    }
    ntry <- ntry + 1
    if (is.null(status) && ntry >= max_tries) {
      log_message(
        "Download failed after {.val {max_tries}} tries from {.url {url}}",
        message_type = "error"
      )
    }
  }
  return(invisible(NULL))
}

#' @title Wilkinson's P-value
#'
#' @param p A vector of P-values.
#' @param r The number of studies to include in the P-value calculation.
#' @param alpha The significance level.
#' @param log.p Whether to return the log of the P-value.
#' @export
#'
#' @examples
#' p <- c(0.01, 0.02, 0.03, 0.04, 0.05)
#' wilkinsonp(p)
#' wilkinsonp(p, r = 2)
#' wilkinsonp(p, alpha = 0.01)
#' wilkinsonp(p, log.p = TRUE)
wilkinsonp <- function(p, r = 1, alpha = 0.05, log.p = FALSE) {
  alpha <- ifelse(alpha > 1, alpha / 100, alpha)
  stopifnot(alpha > 0, alpha < 1)
  alpha <- ifelse(alpha > 0.5, 1 - alpha, alpha)
  keep <- (p >= 0) & (p <= 1)
  invalid <- sum(1L * keep) < 2
  if (invalid) {
    log_message(
      "Must have at least two valid P-values",
      message_type = "warning"
    )
    res <- list(
      p = NA_real_,
      pr = NA_real_,
      r = r,
      critp = NA_real_,
      alpha = alpha,
      validp = p[keep]
    )
  } else {
    pi <- p[keep]
    k <- length(pi)
    if (k != length(p)) {
      log_message(
        "Some studies omitted",
        message_type = "warning"
      )
    }
    if ((r < 1) || (r > k)) {
      r <- 1
      log_message(
        "Illegal r set to 1",
        message_type = "warning"
      )
    }
    pi <- sort(pi)
    pr <- pi[r]
    res <- list(
      p = stats::pbeta(pr, r, k + 1 - r, log.p = log.p),
      pr = pr,
      r = r,
      critp = stats::qbeta(alpha, r, k + 1 - r),
      alpha = alpha,
      validp = pi
    )
  }
  res
}

#' @title Maximum P-value
#'
#' @inheritParams wilkinsonp
#' @export
#'
#' @examples
#' p <- c(0.01, 0.02, 0.03, 0.04, 0.05)
#' maximump(p)
#' maximump(p, alpha = 0.01)
#' maximump(p, log.p = TRUE)
maximump <- function(p, alpha = 0.05, log.p = FALSE) {
  keep <- (p >= 0) & (p <= 1)
  validp <- p[keep]
  k <- length(validp)
  res <- wilkinsonp(p, r = k, alpha, log.p)
  res
}

#' @title Minimum P-value
#'
#' @inheritParams wilkinsonp
#' @export
#'
#' @examples
#' p <- c(0.01, 0.02, 0.03, 0.04, 0.05)
#' minimump(p)
#' minimump(p, alpha = 0.01)
#' minimump(p, log.p = TRUE)
minimump <- function(p, alpha = 0.05, log.p = FALSE) {
  res <- wilkinsonp(p, r = 1, alpha, log.p)
  res
}

#' @title Mean P-value
#'
#' @inheritParams wilkinsonp
#' @export
#'
#' @examples
#' p <- c(0.01, 0.02, 0.03, 0.04, 0.05)
#' meanp(p)
meanp <- function(p) {
  keep <- (p >= 0) & (p <= 1)
  invalid <- sum(1L * keep) < 4
  validp <- p[keep]
  if (invalid) {
    log_message(
      "Must have at least four valid P-values",
      message_type = "warning"
    )
    res <- list(
      z = NA_real_,
      p = NA_real_,
      validp = validp
    )
  } else {
    pi <- mean(validp)
    k <- length(validp)
    z <- (0.5 - pi) * sqrt(12 * k)
    if (k != length(p)) {
      log_message(
        "Some studies omitted",
        message_type = "warning"
      )
    }
    res <- list(
      z = z,
      p = stats::pnorm(z, lower.tail = FALSE),
      validp = validp
    )
  }
  res
}

#' @title Sum P-value
#'
#' @inheritParams wilkinsonp
#' @export
#'
#' @examples
#' p <- c(0.01, 0.02, 0.03, 0.04, 0.05)
#' sump(p)
sump <- function(p) {
  keep <- (p >= 0) & (p <= 1)
  invalid <- sum(1L * keep) < 2
  if (invalid) {
    log_message(
      "Must have at least two valid P-values",
      message_type = "warning"
    )
    res <- list(
      p = NA_real_,
      conservativep = NA_real_,
      validp = p[keep]
    )
  } else {
    validp <- p[keep]
    sigmap <- sum(validp)
    k <- length(validp)
    conservativep <- exp(k * log(sigmap) - lgamma(k + 1))
    nterm <- floor(sigmap) + 1
    denom <- lfactorial(k)
    i_vec <- 1:nterm
    terms <- lchoose(k, i_vec - 1) +
      k * log(sigmap - i_vec + 1) -
      denom
    pm_vec <- 2 * ((i_vec %% 2) == 1) - 1
    psum <- sum(pm_vec * exp(terms))
    if (k != length(p)) {
      log_message(
        "Some studies omitted",
        message_type = "warning"
      )
    }
    if (sigmap > 20) {
      log_message(
        "Likely to be unreliable, check with another method",
        message_type = "warning"
      )
    }
    res <- list(
      p = psum,
      conservativep = conservativep,
      validp = validp
    )
  }
  res
}

#' @title Vote P-value
#'
#' @inheritParams wilkinsonp
#' @export
#'
#' @examples
#' p <- c(0.01, 0.02, 0.03, 0.04, 0.05)
#' votep(p)
#' votep(p, alpha = 0.01)
votep <- function(p, alpha = 0.5) {
  alpha <- ifelse(alpha > 1, alpha / 100, alpha)
  stopifnot(alpha > 0, alpha < 1)
  keep <- (p >= 0) & (p <= 1)
  alp <- vector("numeric", 2)
  if (alpha <= 0.5) {
    alp[1] <- alpha
    alp[2] <- 1 - alpha
  } else {
    alp[2] <- alpha
    alp[1] <- 1 - alpha
  }
  invalid <- sum(1L * keep) < 2
  if (invalid) {
    log_message(
      "Must have at least two valid P-values",
      message_type = "warning"
    )
    res <- list(
      p = NA_real_,
      pos = NA_integer_,
      neg = NA_integer_,
      alpha = alpha,
      validp = p[keep]
    )
  } else {
    validp <- p[keep]
    k <- length(validp)
    pos <- sum(1L * (validp < alp[1]))
    neg <- sum(1L * (validp > alp[2]))
    if (k != length(p)) {
      log_message(
        "Some studies omitted",
        message_type = "warning"
      )
    }
    if ((pos + neg) <= 0) {
      log_message(
        "All P-values are within specified limits of alpha",
        message_type = "warning"
      )
      p <- 1
    } else {
      p <- stats::binom.test(
        pos, pos + neg, 0.5,
        alternative = "greater"
      )$p.value
    }
    res <- list(
      p = p,
      pos = pos,
      neg = neg,
      alpha = alpha,
      validp = validp
    )
  }
  res
}

#' @title Maximum depth of a list
#'
#' @param x A list.
#' @param depth The depth of the list.
#' @export
#'
#' @examples
#' x <- list(
#'   a = list(b = list(c = 1)),
#'   d = list(e = list(f = 2))
#' )
#' max_depth(x)
max_depth <- function(x, depth = 0) {
  if (is.list(x)) {
    return(max(unlist(lapply(x, max_depth, depth + 1))))
  } else {
    return(depth)
  }
}

#' @title Detect outliers using MAD (Median Absolute Deviation)
#'
#' @md
#' @param x Numeric vector.
#' @param nmads Number of MADs from the median to define the boundaries for outliers.
#' Default is `2.5`.
#' @param constant Constant factor to convert the MAD to a standard deviation.
#' Default is `1.4826`, which is consistent with the MAD of a normal distribution.
#' @param type Type of outliers to detect.
#' Available options are `"both"`, `"lower"`, or `"higher"`.
#' If `type` is `"both"`, it detects both lower and higher outliers.
#' If `type` is `"lower"`, it detects only lower outliers.
#' If `type` is `"higher"`, it detects only higher outliers.
#'
#' @return Numeric vector of indices indicating the positions of outliers in `x`.
#'
#' @export
#'
#' @examples
#' x <- c(1, 2, 3, 4, 5, 100)
#' is_outlier(x) # returns 6
#'
#' x <- c(3, 4, 5, NA, 6, 7)
#' is_outlier(x, nmads = 1.5, type = "lower") # returns 4
#'
#' x <- c(10, 20, NA, 15, 35)
#' is_outlier(x, nmads = 2, type = "higher") # returns 3, 5
is_outlier <- function(
    x,
    nmads = 2.5,
    constant = 1.4826,
    type = c("both", "lower", "higher")) {
  type <- match.arg(type, c("both", "lower", "higher"))
  mad <- stats::mad(x, constant = constant, na.rm = TRUE)
  upper <- stats::median(x, na.rm = TRUE) + nmads * mad
  lower <- stats::median(x, na.rm = TRUE) - nmads * mad
  if (type == "both") {
    out <- which(x > upper | x < lower)
  }
  if (type == "lower") {
    out <- which(x < lower)
  }
  if (type == "higher") {
    out <- which(x > upper)
  }
  out <- c(which(is.na(x)), out)
  return(out)
}

#' @title Check CI environment
#'
#' @return
#' A logical value.
#'
#' @export
check_ci_env <- function() {
  if (interactive()) {
    return(TRUE)
  }

  github_actions <- Sys.getenv("GITHUB_ACTIONS", unset = "")
  github_workflow <- Sys.getenv("GITHUB_WORKFLOW", unset = "")

  if (nzchar(github_actions) && github_actions == "true") {
    if (nzchar(github_workflow) && tolower(github_workflow) == "pkgdown") {
      return(TRUE)
    }
  }

  return(FALSE)
}

#' @title Check if the system is running on Apple Silicon
#'
#' @return
#' A logical value.
#'
#' @export
is_apple_silicon <- function() {
  is_osx() &&
    identical(tolower(Sys.info()[["machine"]]), "arm64")
}

#' @title Check if the operating system is Linux
#'
#' @return
#' A logical value.
#'
#' @export
is_linux <- function() {
  identical(tolower(Sys.info()[["sysname"]]), "linux")
}

#' @title Check if the operating system is macOS
#'
#' @return
#' A logical value.
#'
#' @export
is_osx <- function() {
  identical(tolower(Sys.info()[["sysname"]]), "darwin")
}

#' @title Check if the operating system is Windows
#'
#' @return
#' A logical value.
#'
#' @export
is_windows <- function() {
  identical(.Platform$OS.type, "windows")
}

test_that("%ss% returns left side when not NULL", {
  expect_equal(5 %ss% 10, 5)
  expect_equal("a" %ss% "b", "a")
  expect_equal(0 %ss% 10, 0)
  expect_equal(FALSE %ss% TRUE, FALSE)
})

test_that("%ss% returns right side when left is NULL", {
  expect_equal(NULL %ss% 10, 10)
  expect_equal(NULL %ss% "b", "b")
  expect_equal(NULL %ss% NULL, NULL)
})

test_that("get_namespace_fun returns function from namespace", {
  fn <- get_namespace_fun("base", "mean")
  expect_true(is.function(fn))
  expect_equal(fn(1:10), mean(1:10))
})

test_that("get_namespace_fun returns NULL for missing function", {
  expect_message(
    result <- get_namespace_fun("base", "nonexistent_function_xyz"),
    "not found"
  )
  expect_null(result)
})

test_that("invoke_fun works with function objects", {
  f <- function(x, y) x + y
  expect_equal(invoke_fun(f, list(x = 1, y = 2)), 3)
})

test_that("invoke_fun works with function name strings", {
  f <- function(x, y) x * y
  expect_equal(invoke_fun("f", list(x = 3, y = 4)), 12)
})

test_that("invoke_fun works with ... args", {
  f <- function(x, y) x - y
  expect_equal(invoke_fun("f", x = 10, y = 3), 7)
})

test_that("invoke_fun works with no args", {
  f <- function() 42
  expect_equal(invoke_fun(f), 42)
})

test_that("capitalize works on basic strings", {
  expect_equal(capitalize("hello world"), "Hello world")
  expect_equal(capitalize("Hello World"), "Hello World")
})

test_that("capitalize handles NULL and factor", {
  expect_null(capitalize(NULL))
  expect_equal(capitalize(factor("hello")), "Hello")
})

test_that("capitalize with force_tolower", {
  expect_equal(capitalize("hELLO", force_tolower = TRUE), "Hello")
})

test_that("capitalize errors on non-character", {
  expect_error(capitalize(123))
})

test_that("unnest_fun works on basic case", {
  data <- data.frame(id = 1:2, x = c("a", "b"), stringsAsFactors = FALSE)
  data$vals <- list(c(1, 2), c(3, 4, 5))
  result <- unnest_fun(data, cols = "vals")
  expect_equal(nrow(result), 5)
  expect_equal(result$vals, c(1, 2, 3, 4, 5))
})

test_that("unnest_fun handles empty data", {
  data <- data.frame(id = integer(0), x = character(0), stringsAsFactors = FALSE)
  result <- unnest_fun(data, cols = character(0))
  expect_equal(nrow(result), 0)
})

test_that("unnest_fun with keep_empty", {
  data <- data.frame(id = 1:2, x = c("a", "b"), stringsAsFactors = FALSE)
  data$vals <- list(c(1, 2), numeric(0))
  result <- unnest_fun(data, cols = "vals", keep_empty = TRUE)
  expect_equal(nrow(result), 3)
  expect_true(is.na(result$vals[3]))
})

test_that("remove_space works correctly", {
  expect_equal(remove_space(" hello  world "), "hello world ")
  expect_equal(
    remove_space(" hello  world ", trim_end = TRUE),
    "hello world"
  )
  expect_equal(
    remove_space(" hello  world ", trim_start = FALSE),
    " hello world "
  )
  expect_equal(
    remove_space(" hello  world ", collapse_multiple = FALSE),
    "hello  world "
  )
})

test_that("remove_space handles NULL and factor", {
  expect_null(remove_space(NULL))
  expect_equal(remove_space(factor(" a  b ")), "a b ")
})

test_that("remove_space errors on non-character", {
  expect_error(remove_space(123))
})

test_that("remove_space preserves newlines", {
  expect_equal(
    remove_space("hello\n  world"),
    "hello\nworld"
  )
  expect_equal(
    remove_space("hello\n  world", preserve_newlines = FALSE),
    "hello world"
  )
})

test_that("max_depth works correctly", {
  expect_equal(max_depth(list(a = 1)), 1)
  expect_equal(max_depth(list(a = list(b = 1))), 2)
  expect_equal(max_depth(list(a = list(b = list(c = 1)))), 3)
  expect_equal(max_depth(1), 0)
})

test_that("is_outlier detects outliers", {
  x <- c(1, 2, 3, 4, 5, 100)
  out <- is_outlier(x)
  expect_true(6 %in% out)
})

test_that("is_outlier handles type parameter", {
  x <- c(-100, 1, 2, 3, 4, 5, 100)
  lower_out <- is_outlier(x, type = "lower")
  higher_out <- is_outlier(x, type = "higher")
  expect_true(1 %in% lower_out)
  expect_true(7 %in% higher_out)
})

test_that("is_outlier includes NA indices", {
  x <- c(1, NA, 3, 4, 5)
  out <- is_outlier(x)
  expect_true(2 %in% out)
})

test_that("check_ci_env returns logical", {
  result <- check_ci_env()
  expect_true(is.logical(result))
})

test_that("wilkinsonp returns correct structure", {
  p <- c(0.01, 0.02, 0.03, 0.04, 0.05)
  result <- wilkinsonp(p)
  expect_true(is.list(result))
  expect_true("p" %in% names(result))
  expect_true("pr" %in% names(result))
  expect_true("r" %in% names(result))
})

test_that("wilkinsonp warns with too few valid p-values", {
  expect_message(wilkinsonp(c(-1, 2)), "at least two")
})

test_that("maximump and minimump work", {
  p <- c(0.01, 0.02, 0.03, 0.04, 0.05)
  max_res <- maximump(p)
  min_res <- minimump(p)
  expect_true(is.list(max_res))
  expect_true(is.list(min_res))
  expect_true(!is.na(max_res$p))
  expect_true(!is.na(min_res$p))
})

test_that("meanp returns correct structure", {
  p <- c(0.01, 0.02, 0.03, 0.04, 0.05)
  result <- meanp(p)
  expect_true(is.list(result))
  expect_true("z" %in% names(result))
  expect_true("p" %in% names(result))
})

test_that("meanp warns with too few valid p-values", {
  expect_message(meanp(c(0.01, 0.02, 0.03)), "at least four")
})

test_that("sump returns correct structure", {
  p <- c(0.01, 0.02, 0.03, 0.04, 0.05)
  result <- sump(p)
  expect_true(is.list(result))
  expect_true("p" %in% names(result))
  expect_true("conservativep" %in% names(result))
})

test_that("votep returns correct structure", {
  p <- c(0.01, 0.02, 0.03, 0.04, 0.05)
  result <- votep(p)
  expect_true(is.list(result))
  expect_true("p" %in% names(result))
  expect_true("pos" %in% names(result))
  expect_true("neg" %in% names(result))
})

test_that("try_get succeeds on working expression", {
  result <- suppressMessages(try_get(expr = 42, max_tries = 2))
  expect_equal(result, 42)
})

test_that("download errors on missing arguments", {
  expect_error(download())
})

#' @title The figlet function
#'
#' @description
#' Create ASCII art text using figlet.
#'
#' @md
#' @param text Text to make bigger.
#' @param font Name of font, path to font, or `figlet_font` object.
#' @param width Width to use when justifying and breaking lines.
#' @param justify Text justification to use in rendering ("left", "centre", "right").
#' @param absolute Logical, indicating if alignment is absolute.
#' @param strip Logical, indicating if whitespace should be removed.
#'
#' @return
#' An object of class `figlet_text` which is a character vector with a handy print method.
#'
#' @references
#' \url{https://github.com/richfitz/rfiglet},
#' \url{https://github.com/jbkunst/figletr},
#' \url{https://www.figlet.org/}
#'
#' @export
#'
#' @examples
#' figlet("thisutils")
figlet <- function(
    text,
    font = "Slant",
    width = getOption("width", 80),
    justify = "left",
    absolute = FALSE,
    strip = TRUE) {
  font <- figlet_font(font)
  str <- .figlet_render(
    text,
    font,
    width,
    justify,
    absolute,
    strip
  )
  class(str) <- "figlet_text"
  attr(str, "font") <- font$name
  attr(str, "text") <- text

  return(str)
}

.figlet_render <- function(
    text,
    font,
    width = getOption("width", 80),
    justify = "left",
    absolute = FALSE,
    strip = TRUE) {
  if (any(grepl("\n", text, fixed = TRUE))) {
    text <- unlist(strsplit(text, "\n", fixed = TRUE))
  }
  if (length(text) == 1L) {
    dat <- .figlet_render_horizontal(text, font)
  } else {
    dat <- lapply(text, .figlet_render_horizontal, font)
  }
  lines <- .figlet_render_layout(
    dat, font$options$hard_blank, width, justify,
    absolute
  )
  mat <- .figlet_render_vertical(lines, font)
  .matrix_to_text(mat, strip = strip)
}

.figlet_render_horizontal <- function(text, font) {
  state <- list(
    curr_char_width = 0L,
    prev_char_width = 0L
  )

  height <- font$options$height
  buffer <- matrix(character(), height, 0)

  char_index <- .asc(text)
  for (char in seq_along(char_index)) {
    cur_char <- font$chars[[char_index[[char]]]]
    if (is.null(cur_char)) {
      log_message(
        "The font {.val {font$name}} does not contain the characters {.val {substr(text, char, char)}}",
        message_type = "error"
      )
    }

    state$curr_char_width <- cur_char$width
    max_smush <- .smush_amount(
      buffer,
      cur_char$data,
      font$options,
      state
    )

    if (font$options$right_to_left) {
      add_left <- cur_char$data
      add_right <- buffer
    } else {
      add_left <- buffer
      add_right <- cur_char$data
    }

    n_left <- ncol(add_left)
    n_right <- ncol(add_right)

    for (i in seq_len(max_smush)) {
      idx <- n_left - max_smush + i
      if (idx >= 1L && idx <= n_left) {
        left <- add_left[, idx]
      } else {
        left <- rep("", height)
      }
      right <- add_right[, i]

      smushed <- .vcapply(
        seq_len(height), function(row) {
          .smush_chars(
            left[row],
            right[row],
            font$options,
            state
          )
        }
      )

      idx <- n_left - max_smush + i
      if (idx >= 1L && idx <= n_left) {
        add_left[, idx] <- smushed
      }
    }
    if (n_right > 0) {
      add_right <- add_right[, (max_smush + 1):n_right, drop = FALSE]
    }

    buffer <- cbind(add_left, add_right, deparse.level = 0)
    state$prev_char_width <- state$curr_char_width
  }

  buffer[buffer == font$options$hard_blank] <- " "
  buffer
}

.figlet_render_vertical <- function(lines, font) {
  if (length(lines) == 1L) {
    return(lines[[1]])
  }

  buffer <- lines[[1]]
  width <- ncol(buffer)
  for (line in 2:length(lines)) {
    cur_line <- lines[[line]]
    max_smush <- .vsmush_amount(
      buffer,
      cur_line,
      font$options
    )
    add_above <- buffer
    add_below <- cur_line
    n_above <- nrow(add_above)
    n_below <- nrow(add_below)

    for (i in seq_len(max_smush)) {
      idx <- n_above - max_smush + i
      above <- add_above[idx, ]
      below <- add_below[i, ]

      smushed <- .vcapply(
        seq_len(width), function(col) {
          .vsmush_chars(above[col], below[col], font$options)
        }
      )

      idx <- n_above - max_smush + i
      if (idx >= 1L && idx <= n_above) {
        add_above[idx, ] <- smushed
      }
    }
    if (n_below > 0) {
      add_below <- add_below[(max_smush + 1):n_below, , drop = FALSE]
    }
    buffer <- rbind(add_above, add_below, deparse.level = 0)
  }

  buffer
}

.figlet_render_layout <- function(
    buffer,
    hard_blank,
    width,
    justify,
    absolute) {
  if (is.list(buffer)) {
    template <- list(
      template = character(),
      start = integer(),
      end = integer()
    )
    for (line in buffer) {
      tmp <- .figlet_render_layout_template(
        line, hard_blank,
        length(template$start),
        sum(nchar(template$template))
      )
      template <- Map(c, template, tmp)
    }
    buffer <- do.call(cbind, buffer)
  } else {
    template <- .figlet_render_layout_template(buffer, hard_blank)
  }

  template$template <- gsub("(?<= ) ", "-", template$template, perl = TRUE)
  tmp <- trimws(gsub("-", " ", strwrap(template$template, width = width)))
  lpad <- nchar(sub("[^ ].*", "", format(tmp, justify = justify)))
  rpad <- max(nchar(tmp) + lpad) - nchar(tmp) - lpad
  if (absolute && max(nchar(tmp)) < width) {
    if (justify == "right") {
      lpad <- lpad + (width - max(nchar(tmp)))
    } else if (justify == "centre") {
      lpad <- lpad + ceiling((width - max(nchar(tmp))) / 2)
    }
  }

  map <- c(1:9, letters, LETTERS)
  words <- lapply(strsplit(trimws(tmp), " +"), function(x) {
    range(match(substr(x, 1, 1), map))
  })

  line <- function(i) {
    j <- template$start[words[[i]][1]]:template$end[words[[i]][2]]
    text <- buffer[, j, drop = FALSE]
    if (lpad[i] > 0) {
      text <- cbind(matrix(" ", nrow(text), lpad[i]), text)
    }
    if (rpad[i] > 0) {
      text <- cbind(text, matrix(" ", nrow(text), rpad[i]))
    }
    text
  }

  lapply(seq_along(words), line)
}

.figlet_render_layout_template <- function(
    text, hard_blank,
    offset_char = 0,
    offset_text = 0) {
  pos <- apply(array(text %in% c(" ", hard_blank), dim(text)), 2, all)
  map <- c(1:9, letters, LETTERS)
  n <- cumsum(pos) + 1L + offset_char
  template <- map[n]
  template[pos] <- " "
  start <- c(1, which(pos) + 1L) + offset_text
  end <- c(which(pos) - 1L, ncol(text)) + offset_text
  list(
    template = paste(template, collapse = ""),
    start = start,
    end = end
  )
}

#' Get a figlet font
#'
#' @param font Path or name of the font to load
#'
#' @return A `figlet_font` object for use with [figlet]
#' @export
figlet_font <- function(font) {
  if (inherits(font, "figlet_font")) {
    return(font)
  }

  if (font == "Slant") {
    font_path <- system.file(
      "fonts/Slant.flf",
      package = "thisutils"
    )
    return(.figlet_font_read(font_path))
  }

  if (file.exists(font)) {
    return(.figlet_font_read(font))
  }

  log_message(
    "Font {.val {font}} not found",
    message_type = "error"
  )
}

.figlet_font_read <- function(filename) {
  name <- tools::file_path_sans_ext(basename(filename))
  if (!file.exists(filename)) {
    log_message(
      "{.file {filename}} does not exist",
      message_type = "error"
    )
  }
  data <- readLines(filename, warn = FALSE)
  options <- .figlet_font_options(data, filename, name)

  is_comment <- seq_len(options$comment_lines) + 1L
  chars <- .figlet_font_characters(
    data[-c(1L, is_comment)], options,
    filename, name
  )
  ret <- list(
    name = name,
    comments = data[is_comment],
    chars = chars,
    options = options
  )
  class(ret) <- "figlet_font"
  ret
}

.figlet_font_options <- function(data, filename, name) {
  if (length(data) == 0) {
    log_message(
      "{.file {filename}} is empty",
      message_type = "error"
    )
  }
  re_magic_number <- "^[tf]lf2."
  header <- data[[1]]
  if (!grepl(re_magic_number, header, perl = TRUE)) {
    log_message(
      "{.file {filename}} is not a valid font",
      message_type = "error"
    )
  }
  header <- strsplit(sub(re_magic_number, "", header), " ")[[1]]
  if (length(header) < 6) {
    log_message(
      "{.file {filename}} has a malformed header",
      message_type = "error"
    )
  }

  nms <- c(
    "hard_blank", "height", "base_line", "max_length",
    "old_layout", "comment_lines", "print_direction", "full_layout"
  )

  msg <- max(0L, length(nms) - length(header))
  if (msg > 0L) {
    header <- c(header, rep(list(NA_integer_), msg))
  }

  options <- as.list(header[seq_along(nms)])
  options[-1] <- lapply(options[-1], as.integer)
  names(options) <- nms

  if (is.na(options$full_layout)) {
    if (options$old_layout == 0L) {
      options$full_layout <- 64L
    } else if (options$old_layout < 0L) {
      options$full_layout <- 0L
    } else {
      options$full_layout <- as.integer(
        bitwOr(
          bitwAnd(options$old_layout, 31L), 128L
        )
      )
    }
  }

  options$smush_mode <- options$full_layout
  options$right_to_left <- identical(options$print_direction, 1L)
  options
}

.figlet_font_characters <- function(data, options, filename, name) {
  code_standard <- 32:126
  code_extra <- c(196, 214, 220, 228, 246, 252, 223)
  code_req <- c(code_standard, code_extra)

  get_character <- function(i, d) {
    .figlet_font_character(d[, i], options)
  }

  i_req <- seq_len(length(code_req) * options$height)
  dat_req <- matrix(data[i_req], options$height)
  chars <- vector("list", max(code_req))
  chars[code_req] <- lapply(seq_along(code_req), get_character, dat_req)
  chars
}

.figlet_font_character <- function(x, options) {
  if (any(is.na(iconv(x)))) {
    return(NULL)
  }

  re_end_marker <- ".*?(.)\\s*$"
  char <- sub(re_end_marker, "\\1", x[[1]])
  if (identical(char, "^")) {
    char <- "\\^"
  } else if (identical(char, "\\")) {
    char <- "\\\\"
  }
  re_end <- sprintf("[%s]{1,2}\\s*$", char)
  txt <- sub(re_end, "", x, perl = TRUE)
  char_width <- max(nchar(gsub(options$hard_blank, "", txt, fixed = TRUE)))

  m <- strsplit(txt, NULL)
  n <- lengths(m)
  if (any(i <- n < max(n))) {
    m[i] <- lapply(m[i], function(x) c(x, rep(" ", max(n) - length(x))))
  }
  list(
    width = as.integer(char_width),
    data = matrix(unlist(m), length(txt), byrow = TRUE)
  )
}

.asc <- function(x) {
  strtoi(charToRaw(x), 16L)
}

.vcapply <- function(X, FUN, ...) {
  vapply(X, FUN, character(1), ...)
}

.matrix_to_text <- function(m, strip) {
  ret <- apply(m, 1, paste, collapse = "")
  if (strip) {
    ret <- sub(" +$", "", ret)
    ret <- ret[nzchar(ret)]
  }
  ret <- gsub("\\\\", ".", ret)
  ret
}

.smush_chars <- function(left, right, options, state) {
  if (.is_space(left)) {
    return(right)
  } else if (.is_space(right)) {
    return(left)
  }

  if (state$prev_char_width < 2 || state$curr_char_width < 2) {
    return(NULL)
  }

  smush_mode <- options$smush_mode
  if (bitwAnd(smush_mode, 128) == 0) {
    return(NULL)
  }

  hard_blank <- options$hard_blank
  if (bitwAnd(smush_mode, 63) == 0) {
    if (left == hard_blank) {
      return(right)
    }
    if (right == hard_blank) {
      return(left)
    }
  }

  if (left == hard_blank && right == hard_blank) {
    if (bitwAnd(smush_mode, 32) > 0) {
      return(left)
    } else {
      return(NULL)
    }
  }

  if (bitwAnd(smush_mode, 1)) {
    if (left == right) {
      return(left)
    }
  }

  NULL
}

.smush_amount <- function(buffer, cur_char, options, state) {
  if (bitwAnd(options$smush_mode, bitwOr(128, 64)) == 0) {
    return(0L)
  }

  max_smush <- state$curr_char_width

  for (row in seq_len(options$height)) {
    line_left <- buffer[row, ]
    line_right <- cur_char[row, ]
    if (options$right_to_left) {
      ll <- line_left
      line_left <- line_right
      line_right <- ll
    }

    linebd <- .length_rstrip(line_left)
    if (linebd < 1L) {
      linebd <- 1L
    }

    if (linebd <= length(line_left)) {
      ch1 <- line_left[linebd]
    } else {
      linebd <- 1L
      ch1 <- ""
    }

    charbd <- length(line_right) - .length_lstrip(line_right) + 1L
    if (charbd <= length(line_right)) {
      ch2 <- line_right[charbd]
    } else {
      charbd <- length(line_right) + 1L
      ch2 <- ""
    }

    amt <- (charbd - 1L) + length(line_left) - 1L - (linebd - 1L)

    if (ch1 == "" || ch1 == " ") {
      amt <- amt + 1L
    } else if (ch2 != "" && !is.null(.smush_chars(ch1, ch2, options, state))) {
      amt <- amt + 1L
    }

    if (amt < max_smush) {
      max_smush <- amt
    }
  }

  max_smush
}

.vsmush_amount <- function(buffer, cur_line, options) {
  max_smush <- options$height
  for (col in seq_len(min(ncol(buffer), ncol(cur_line)))) {
    line_above <- buffer[, col]
    line_below <- cur_line[, col]

    linebd <- .length_rstrip(line_above)
    if (linebd < 1L) {
      linebd <- 1L
    }
    ch1 <- line_above[linebd]

    charbd <- length(line_below) - .length_lstrip(line_below) + 1L
    if (charbd <= length(line_below)) {
      ch2 <- line_below[charbd]
    } else {
      charbd <- length(line_below) + 1L
      ch2 <- ""
    }

    amt <- (charbd - 1L) + length(line_above) - 1L - (linebd - 1L)

    if (ch1 == "" || ch1 == " ") {
      amt <- amt + 1L
    } else if (ch2 != "" && !is.null(.vsmush_chars(ch1, ch2, options))) {
      amt <- amt + 1L
    }

    if (amt < max_smush) {
      max_smush <- amt
    }
  }

  max_smush
}

.vsmush_chars <- function(above, below, options) {
  if (.is_space(above)) {
    return(below)
  } else if (.is_space(below)) {
    return(above)
  }

  smush_mode <- options$smush_mode
  if (bitwAnd(smush_mode, 16384) == 0) {
    return(NULL)
  }

  if (bitwAnd(smush_mode, 256)) {
    if (above == below) {
      return(above)
    }
  }

  NULL
}

.is_space <- function(x) {
  grepl("\\s", x, perl = TRUE)
}

.length_lstrip <- function(x) {
  if (length(x) == 0) {
    return(0L)
  }
  i <- x != " "
  if (i[[1]]) {
    return(length(i))
  }
  if (!any(i)) {
    return(0L)
  }
  length(i) - which(i)[[1]] + 1L
}

.length_rstrip <- function(x) {
  if (length(x) == 0) {
    return(0L)
  }
  i <- x != " "
  n <- length(x)
  if (i[[n]]) {
    return(n)
  }
  if (!any(i)) {
    return(0L)
  }
  max(which(i))
}

#' @export
print.figlet_text <- function(x, ...) {
  cat(paste0(x, "\n", collapse = ""))
  invisible(x)
}

#' @export
print.figlet_font <- function(x, preview = TRUE, ...) {
  cat(sprintf("<figlet_font object: %s>\n", x$name))
  if (preview) {
    print(figlet(x$name, x))
  }
  invisible(x)
}

#' @export
as.character.figlet_text <- function(x, ...) {
  paste0(x, collapse = "\n")
}

test_that("figlet produces figlet_text output", {
  result <- figlet("test")
  expect_true(inherits(result, "figlet_text"))
  expect_true(is.character(result))
  expect_true(length(result) > 0)
})

test_that("figlet_font loads default Slant font", {
  font <- figlet_font("Slant")
  expect_true(inherits(font, "figlet_font"))
  expect_equal(font$name, "Slant")
})

test_that("figlet_font returns same object if already figlet_font", {
  font <- figlet_font("Slant")
  font2 <- figlet_font(font)
  expect_identical(font, font2)
})

test_that("figlet_font errors on nonexistent font", {
  expect_error(figlet_font("NonExistentFont12345"))
})

test_that("print.figlet_text works", {
  result <- figlet("hi")
  expect_output(print(result))
})

test_that("as.character.figlet_text works", {
  result <- figlet("hi")
  s <- as.character(result)
  expect_true(is.character(s))
  expect_equal(length(s), 1)
})

test_that("figlet handles multiline input", {
  result <- figlet("a\nb")
  expect_true(inherits(result, "figlet_text"))
})

#' @title Parallelize a function
#'
#' @md
#' @inheritParams log_message
#' @param x A vector or list to apply over.
#' @param fun The function to be applied to each element.
#' @param cores The number of cores to use for parallelization with [foreach::foreach].
#' Default is `1`.
#' @param export_fun The functions to export the function to workers.
#' @param clean_result Whether to remove failed results from output.
#' If `FALSE`, failed results are kept as error objects.
#' Default is `FALSE`.
#' @param throw_error Whether to print detailed error information for failed results.
#' Default is `TRUE`.
#' @param progress_bar_width Width of the verbose progress bar in characters.
#' Default is `10L`.
#'
#' @return
#' A list of computed results.
#' If `clean_result = FALSE`, failed results are included as error objects.
#' If `clean_result = TRUE`, only successful results are returned.
#'
#' @export
#'
#' @examples
#' parallelize_fun(1:3, function(x) {
#'   Sys.sleep(0.2)
#'   x^2
#' })
#'
#' parallelize_fun(list(1, 2, 3), function(x) {
#'   Sys.sleep(0.2)
#'   x^2
#' }, cores = 2)
#'
#' # Examples with error handling
#' parallelize_fun(1:5, function(x) {
#'   if (x == 3) stop("Error on element 3")
#'   x^2
#' }, clean_result = FALSE)
#'
#' parallelize_fun(1:5, function(x) {
#'   if (x == 3) stop("Error on element 3")
#'   x^2
#' }, clean_result = TRUE)
#'
#' # Control error printing
#' parallelize_fun(1:5, function(x) {
#'   if (x == 2) stop("Error on element 3")
#'   if (x == 4) stop("Error on element 4")
#'   x^2
#' })
#'
#' parallelize_fun(1:5, function(x) {
#'   if (x == 3) stop("Error on element 3")
#'   x^2
#' }, throw_error = FALSE)
parallelize_fun <- function(
  x,
  fun,
  cores = 1,
  export_fun = NULL,
  clean_result = FALSE,
  throw_error = TRUE,
  progress_bar_width = 10L,
  timestamp_format = paste0(
    "[",
    format(Sys.time(), "%Y-%m-%d %H:%M:%S"),
    "] "
  ),
  verbose = TRUE
) {
  total <- length(x)
  has_names <- !is.null(names(x)) && any(names(x) != "")
  is_vector <- is.vector(x) && !is.list(x)
  show_values <- !has_names && is_vector

  if (verbose) {
    progress_env <- environment()
    old_cli_opts <- options(
      cli.progress_show_after = 0,
      cli.progress_clear = FALSE
    )
    on.exit(options(old_cli_opts), add = TRUE)

    pb <- cli::cli_progress_bar(
      format = paste0(
        "{cli::make_ansi_style('orange')(cli::pb_spin)} {timestamp_format}",
        "Running for {.pkg {cli::pb_status}}[{.pkg {cli::pb_current}}/{.pkg {cli::pb_total}}] ",
        "{(parallel_progress_bar(cli::pb_current, cli::pb_total, progress_bar_width))} ",
        "{cli::pb_percent} | ETA: {.pkg {cli::pb_eta}}"
      ),
      format_done = paste0(
        "{cli::col_green(cli::symbol$tick)} {timestamp_format}",
        "Completed {.pkg {cli::pb_total}} tasks ",
        "in {.pkg {cli::pb_elapsed}}"
      ),
      total = total,
      clear = FALSE,
      .envir = progress_env
    )
  }

  safe_call <- function(fun, ...) {
    msg_con <- file(nullfile(), open = "w")
    sink(msg_con, type = "message")
    on.exit({
      sink(type = "message")
      close(msg_con)
    })
    suppressWarnings(fun(...))
  }

  if (cores == 1) {
    log_message(
      "Using {.pkg 1} core",
      timestamp_format = timestamp_format,
      verbose = verbose
    )

    if (verbose) {
      output_list <- vector("list", total)

      for (i in seq_along(x)) {
        output_list[i] <- list(tryCatch(
          safe_call(fun, x[[i]]),
          error = function(e) {
            structure(
              list(
                error = e$message,
                index = i,
                input = x[[i]]
              ),
              class = "parallelize_error"
            )
          }
        ))

        if (has_names) {
          cli::cli_progress_update(id = pb, status = names(x)[i])
        } else if (show_values) {
          cli::cli_progress_update(id = pb, status = as.character(x[[i]]))
        } else {
          cli::cli_progress_update(id = pb)
        }
      }

      cli::cli_progress_done(id = pb)
    } else {
      output_list <- base::lapply(
        X = x,
        FUN = function(xi) {
          tryCatch(
            safe_call(fun, xi),
            error = function(e) {
              structure(
                list(
                  error = e$message,
                  input = xi
                ),
                class = "parallelize_error"
              )
            }
          )
        }
      )
    }
  }

  if (cores > 1) {
    cores <- cores_detect(cores, total)
    log_message(
      "Using {.pkg {cores}} cores",
      timestamp_format = timestamp_format,
      verbose = verbose
    )

    if (verbose) {
      output_list <- parallel_collect_results(
        x = x,
        fun = fun,
        total = total,
        cores = cores,
        export_fun = export_fun,
        safe_call = safe_call,
        progress_id = pb,
        progress_env = progress_env,
        has_names = has_names,
        show_values = show_values
      )
      cli::cli_progress_done(id = pb)
    } else {
      doParallel::registerDoParallel(cores = cores)
      on.exit(doParallel::stopImplicitCluster(), add = TRUE)

      i <- NULL
      "%dopar%" <- foreach::"%dopar%"
      output_list <- foreach::foreach(
        i = seq_along(x),
        .export = export_fun,
        .options.multicore = parallel_foreach_options(cores),
        .options.snow = list(
          preschedule = FALSE
        )
      ) %dopar%
        {
          tryCatch(
            safe_call(fun, x[[i]]),
            error = function(e) {
              structure(
                list(
                  error = e$message,
                  index = i,
                  input = x[[i]]
                ),
                class = "parallelize_error"
              )
            }
          )
        }
    }
  }

  log_message(
    "Building results",
    timestamp_format = timestamp_format,
    verbose = verbose
  )

  error_indices <- vapply(
    output_list,
    function(x) inherits(x, "parallelize_error"),
    logical(1)
  )
  if (any(error_indices)) {
    log_message(
      "Found {.pkg {sum(error_indices)}} failed result{?s}",
      timestamp_format = timestamp_format,
      message_type = "warning",
      verbose = verbose
    )

    if (throw_error && verbose) {
      error_objects <- output_list[error_indices]
      error_inputs <- x[error_indices]

      error_msgs <- vapply(
        error_objects,
        function(e) e$error,
        character(1)
      )
      error_groups <- split(
        seq_along(error_msgs),
        error_msgs
      )

      group_lines <- vapply(
        names(error_groups),
        function(msg) {
          idx <- error_groups[[msg]]
          inputs <- error_inputs[idx]
          n <- length(idx)
          max_show <- 3
          shown <- vapply(
            utils::head(inputs, max_show),
            function(v) {
              parse_inline_expressions("{.val {v}}")
            },
            character(1)
          )
          task_str <- paste(shown, collapse = ", ")
          if (n > max_show) {
            task_str <- paste0(
              task_str,
              sprintf(" and %d more", n - max_show)
            )
          }
          parse_inline_expressions(
            paste0("{.emph ", msg, "} (", n, "): ", task_str)
          )
        },
        character(1)
      )

      error_message <- paste0(
        "Error details:\n",
        paste(group_lines, collapse = "\n")
      )
      log_message(
        error_message,
        symbol = cli::symbol$cross,
        text_color = "red",
        verbose = verbose
      )
    }

    if (clean_result) {
      output_list <- output_list[!error_indices]
      x <- x[!error_indices]
      log_message(
        "Removed {.pkg {sum(error_indices)}} failed result{?s}",
        timestamp_format = timestamp_format,
        verbose = verbose
      )
    }
  }

  if (has_names) {
    names(output_list) <- names(x)
  } else if (is_vector) {
    names(output_list) <- as.character(x)
  }

  return(output_list)
}

parallel_collect_results <- function(
  x,
  fun,
  total,
  cores,
  export_fun,
  safe_call,
  progress_id,
  progress_env,
  has_names,
  show_values
) {
  if (.Platform$OS.type != "windows") {
    return(parallel_collect_results_mc(
      x = x,
      fun = fun,
      total = total,
      cores = cores,
      safe_call = safe_call,
      progress_id = progress_id,
      progress_env = progress_env,
      has_names = has_names,
      show_values = show_values
    ))
  }

  parallel_collect_results_cluster(
    x = x,
    fun = fun,
    total = total,
    cores = cores,
    export_fun = export_fun,
    safe_call = safe_call,
    progress_id = progress_id,
    progress_env = progress_env,
    has_names = has_names,
    show_values = show_values
  )
}

parallel_collect_results_mc <- function(
  x,
  fun,
  total,
  cores,
  safe_call,
  progress_id,
  progress_env,
  has_names,
  show_values
) {
  parallel_worker_task <- function(i) {
    tryCatch(
      safe_call(fun, x[[i]]),
      error = function(e) {
        structure(
          list(
            error = e$message,
            index = i,
            input = x[[i]]
          ),
          class = "parallelize_error"
        )
      }
    )
  }

  output_list <- vector("list", total)
  jobs <- list()
  job_indices <- list()
  next_index <- 1L
  completed <- 0L

  launch_job <- function(task_index) {
    job <- parallel::mcparallel(
      parallel_worker_task(task_index),
      silent = TRUE
    )
    job_pid <- as.character(job$pid)
    jobs[[job_pid]] <<- job
    job_indices[[job_pid]] <<- task_index
  }

  while (next_index <= min(cores, total)) {
    launch_job(next_index)
    next_index <- next_index + 1L
  }

  while (completed < total) {
    result <- parallel::mccollect(jobs, wait = FALSE)
    if (is.null(result)) {
      Sys.sleep(0.01)
      next
    }

    for (pid in names(result)) {
      task_index <- job_indices[[pid]]
      output_list[task_index] <- list(result[[pid]])
      jobs[[pid]] <- NULL
      job_indices[[pid]] <- NULL
      completed <- completed + 1L

      cli::cli_progress_update(
        id = progress_id,
        inc = 1L,
        status = progress_status(x, task_index, has_names, show_values),
        .envir = progress_env
      )

      if (next_index <= total) {
        launch_job(next_index)
        next_index <- next_index + 1L
      }
    }
  }

  output_list
}

parallel_collect_results_cluster <- function(
  x,
  fun,
  total,
  cores,
  export_fun,
  safe_call,
  progress_id,
  progress_env,
  has_names,
  show_values
) {
  cl <- make_parallel_cluster(cores)
  on.exit(parallel::stopCluster(cl), add = TRUE)

  parallel_worker_task <- function(i) {
    tryCatch(
      safe_call(fun, x[[i]]),
      error = function(e) {
        structure(
          list(
            error = e$message,
            index = i,
            input = x[[i]]
          ),
          class = "parallelize_error"
        )
      }
    )
  }

  parallel::clusterExport(
    cl = cl,
    varlist = "parallel_worker_task",
    envir = environment()
  )

  if (!is.null(export_fun) && length(export_fun) > 0) {
    export_env <- environment(fun)
    if (is.null(export_env)) {
      export_env <- parent.frame()
    }
    parallel::clusterExport(
      cl = cl,
      varlist = export_fun,
      envir = export_env
    )
  }

  output_list <- vector("list", total)
  next_index <- 1L
  active_workers <- min(length(cl), total)

  sendCall <- get_namespace_fun("parallel", "sendCall")
  for (worker_idx in seq_len(active_workers)) {
    sendCall(
      con = cl[[worker_idx]],
      fun = parallel_worker_task,
      args = list(next_index),
      tag = next_index
    )
    next_index <- next_index + 1L
  }

  completed <- 0L
  while (completed < total) {
    result <- get_namespace_fun("parallel", "recvOneResult")(cl)
    task_index <- as.integer(result$tag)
    output_list[task_index] <- list(result$value)
    completed <- completed + 1L

    cli::cli_progress_update(
      id = progress_id,
      inc = 1L,
      status = progress_status(x, task_index, has_names, show_values),
      .envir = progress_env
    )

    if (next_index <= total) {
      sendCall(
        con = cl[[result$node]],
        fun = parallel_worker_task,
        args = list(next_index),
        tag = next_index
      )
      next_index <- next_index + 1L
    }
  }

  output_list
}

parallel_foreach_options <- function(cores) {
  list(
    preschedule = FALSE,
    cores = cores
  )
}

progress_status <- function(x, index, has_names, show_values) {
  if (has_names) {
    return(names(x)[index])
  }

  if (show_values) {
    return(as.character(x[[index]]))
  }

  NULL
}

make_parallel_cluster <- function(cores) {
  if (.Platform$OS.type == "windows") {
    return(parallel::makePSOCKcluster(cores))
  }

  parallel::makeForkCluster(cores)
}

parallel_progress_bar <- function(
  current,
  total,
  width = 10L
) {
  width <- suppressWarnings(as.integer(width[[1]]))
  if (is.na(width) || width < 1L) {
    width <- 10L
  }

  current <- suppressWarnings(as.numeric(current))
  total <- suppressWarnings(as.numeric(total))

  if (!is.finite(total) || total <= 0) {
    return("")
  }

  if (!is.finite(current)) {
    current <- 0
  }

  current <- max(0, min(current, total))
  filled <- floor(width * current / total)
  if (current >= total) {
    filled <- width
  }

  empty <- max(width - filled, 0L)

  paste0(
    cli::col_green(strrep("\u25A0", filled)),
    strrep(" ", empty)
  )
}

cores_detect <- function(
  cores = 1,
  num_session = NULL
) {
  if (is.null(num_session)) {
    return(1)
  }
  detected_cores <- suppressWarnings(
    parallel::detectCores(logical = FALSE)
  )
  if (!is.finite(detected_cores) || detected_cores < 2) {
    detected_cores <- 2L
  }

  max_cores <- max(1L, as.integer(detected_cores) - 1L)
  requested_cores <- suppressWarnings(as.integer(cores[[1]]))
  if (is.na(requested_cores) || requested_cores < 1L) {
    requested_cores <- 1L
  }

  num_session <- suppressWarnings(as.integer(num_session[[1]]))
  if (is.na(num_session) || num_session < 1L) {
    num_session <- 1L
  }

  min(max_cores, requested_cores, num_session)
}

test_that("parallelize_fun works with single core", {
  result <- suppressMessages(
    parallelize_fun(1:3, function(x) x^2, verbose = FALSE)
  )
  expect_equal(length(result), 3)
  expect_equal(result[[1]], 1)
  expect_equal(result[[2]], 4)
  expect_equal(result[[3]], 9)
})

test_that("parallelize_fun handles errors with clean_result = FALSE", {
  result <- suppressMessages(
    parallelize_fun(1:3, function(x) {
      if (x == 2) stop("fail")
      x
    }, verbose = FALSE, throw_error = FALSE)
  )
  expect_equal(length(result), 3)
  expect_true(inherits(result[[2]], "parallelize_error"))
})

test_that("parallelize_fun handles errors with clean_result = TRUE", {
  result <- suppressMessages(
    parallelize_fun(1:3, function(x) {
      if (x == 2) stop("fail")
      x
    }, clean_result = TRUE, verbose = FALSE, throw_error = FALSE)
  )
  expect_equal(length(result), 2)
})

test_that("parallelize_fun preserves names for named vectors", {
  x <- c(a = 1, b = 2, c = 3)
  result <- suppressMessages(
    parallelize_fun(x, function(v) v^2, verbose = FALSE)
  )
  expect_equal(names(result), c("a", "b", "c"))
})

test_that("parallelize_fun names output with values for unnamed vectors", {
  result <- suppressMessages(
    parallelize_fun(1:3, function(x) x^2, verbose = FALSE)
  )
  expect_equal(names(result), c("1", "2", "3"))
})

test_that("parallelize_fun does not set names for list inputs", {
  result <- suppressMessages(
    parallelize_fun(list(1, 2, 3), function(x) x^2, verbose = FALSE)
  )
  expect_null(names(result))
})

test_that("parallelize_fun restores cli options on error", {
  old_show <- getOption("cli.progress_show_after")
  old_clear <- getOption("cli.progress_clear")
  tryCatch(
    parallelize_fun(1:3, function(x) {
      stop("intentional error")
    }, verbose = TRUE, throw_error = FALSE),
    error = function(e) NULL,
    message = function(m) NULL
  )
  expect_equal(getOption("cli.progress_show_after"), old_show)
  expect_equal(getOption("cli.progress_clear"), old_clear)
})

test_that("parallelize_fun with verbose progress bar", {
  expect_message(
    parallelize_fun(1:3, function(x) x^2, verbose = TRUE)
  )
})

test_that("parallel progress bar uses supplied width", {
  bar <- parallel_progress_bar(6, 10, 12L)
  expect_equal(nchar(cli::ansi_strip(bar)), 12)
})

test_that("parallel progress bar falls back to default width", {
  bar <- parallel_progress_bar(6, 10)
  expect_equal(nchar(cli::ansi_strip(bar)), 10)
})

test_that("parallelize_fun accepts progress_bar_width argument", {
  expect_no_error(
    suppressMessages(
      parallelize_fun(1:2, function(x) x^2, verbose = FALSE, progress_bar_width = 9L)
    )
  )
})

test_that("parallelize_fun preserves input order for multi-core execution", {
  result <- suppressMessages(
    parallelize_fun(1:6, function(x) x, cores = 2, verbose = TRUE)
  )

  expect_equal(unname(unlist(result)), 1:6)
  expect_equal(names(result), as.character(1:6))
})

test_that("parallelize_fun preserves input order for multi-core execution without verbose", {
  result <- suppressMessages(
    parallelize_fun(1:6, function(x) x * 2, cores = 2, verbose = FALSE)
  )

  expect_equal(unname(unlist(result)), (1:6) * 2)
  expect_equal(names(result), as.character(1:6))
})

test_that("parallelize_fun handles uneven multi-core workloads in verbose mode", {
  delays <- c(0.15, 0.01, 0.12, 0.02, 0.08, 0.01)
  result <- suppressMessages(
    parallelize_fun(1:6, function(x) {
      Sys.sleep(delays[[x]])
      x
    }, cores = 2, verbose = TRUE)
  )

  expect_equal(unname(unlist(result)), 1:6)
  expect_equal(names(result), as.character(1:6))
})

test_that("parallelize_fun exports requested dependencies in multi-core mode", {
  offset <- 5
  add_offset <- function(x) x + offset

  result <- suppressMessages(
    parallelize_fun(
      1:4,
      add_offset,
      cores = 2,
      verbose = FALSE,
      export_fun = "offset"
    )
  )

  expect_equal(unname(unlist(result)), 6:9)
  expect_equal(names(result), as.character(1:4))
})

test_that("cores_detect falls back to at least one core", {
  expect_gte(cores_detect(cores = 2, num_session = 4), 1)
})

test_that("parallelize_fun preserves NULL results on a single core", {
  result <- suppressMessages(
    parallelize_fun(
      1:3,
      function(x) if (x == 2) NULL else x,
      verbose = FALSE
    )
  )

  expect_length(result, 3)
  expect_equal(result[[1]], 1)
  expect_null(result[[2]])
  expect_equal(result[[3]], 3)
})

# -*- coding: utf-8 -*-

#' @title Collection of Utility Functions for Data Analysis and Computing
#'
#' @useDynLib thisutils, .registration = TRUE
#'
#' @description
#' Provides utility functions for data analysis and computing. Includes functions for logging, parallel processing, and other computational tasks to streamline workflows.
#'
#' @author Meng Xu (Maintainer), \email{mengxu98@qq.com}
#'
#' @source \url{https://mengxu98.github.io/thisutils/}
#'
#' @md
#' @docType package
#' @name thisutils-package
"_PACKAGE"

#' @title The logo of thisutils
#'
#' @description
#' The thisutils logo, using ASCII or Unicode characters
#' Use [cli::ansi_strip] to get rid of the colors.
#'
#' @md
#' @param unicode Unicode symbols on UTF-8 platforms.
#' Default is [cli::is_utf8_output].
#'
#' @return
#' A character vector with class `thisutils_logo`.
#'
#' @references
#' \url{https://github.com/tidyverse/tidyverse/blob/main/R/logo.R}
#'
#' @export
#' @examples
#' thisutils_logo()
thisutils_logo <- function(unicode = cli::is_utf8_output()) {
  logo <- c(
    "          0          1        2             3     4
            __  __    _              __  _  __
           / /_/ /_  (_)_____ __  __/ /_(_)/ /_____
          / __/ __ ./ // ___// / / / __/ // // ___/
         / /_/ / / / /(__  )/ /_/ / /_/ // /(__  )
         .__/_/ /_/_//____/ .__,_/.__/_//_//____/
      5               6      7        8          9"
  )

  hexa <- c("*", ".", "o", "*", ".", "o", "*", ".", "o", "*")
  if (unicode) {
    hexa <- c("*" = "\u2b22", "o" = "\u2b21", "." = ".")[hexa]
  }

  cols <- c(
    "red", "yellow", "green", "magenta", "cyan",
    "yellow", "green", "white", "magenta", "cyan"
  )

  col_hexa <- mapply(
    function(x, y) cli::make_ansi_style(y)(x),
    hexa, cols,
    SIMPLIFY = FALSE
  )

  for (i in 0:9) {
    pat <- paste0("\\b", i, "\\b")
    logo <- sub(pat, col_hexa[[i + 1]], logo)
  }

  structure(
    cli::col_blue(logo),
    class = "thisutils_logo"
  )
}

#' @title Print logo
#'
#' @param x Input information.
#' @param ... Other parameters.
#'
#' @return Print the ASCII logo
#'
#' @method print thisutils_logo
#'
#' @export
print.thisutils_logo <- function(x, ...) {
  cat(x, ..., sep = "\n")
  invisible(x)
}

.onAttach <- function(libname, pkgname) {
  if (!interactive()) {
    return(invisible())
  }
  verbose <- thisutils::get_verbose()
  if (isTRUE(verbose)) {
    version <- utils::packageVersion(pkgname)
    date <- utils::packageDate(pkgname)
    url <- utils::packageDescription(
      pkgname, fields = "URL"
    )

    msg <- paste0(
      cli::col_grey(strrep("-", 60)),
      "\n",
      cli::col_blue("Version: ", version, " (", date, " update)"),
      "\n",
      cli::col_blue("Website: ", cli::style_italic(url)),
      "\n\n",
      cli::col_grey("This message can be suppressed by:"),
      "\n",
      cli::col_grey("  suppressPackageStartupMessages(library(thisutils))"),
      "\n",
      cli::col_grey("  or options(log_message.verbose = FALSE)"),
      "\n",
      cli::col_grey(strrep("-", 60))
    )

    packageStartupMessage(thisutils_logo())
    packageStartupMessage(msg)
  }
}

#' @title Check and install R packages
#'
#' @md
#' @inheritParams log_message
#' @param packages Package to be installed.
#' Package source can be *CRAN*, *Bioconductor* or *Github*.
#' By default, the package name is extracted according to the `packages` parameter.
#' @param lib The location of the library directories where to install the packages.
#' @param dependencies Whether to install dependencies of the packages.
#' Default is `TRUE`.
#' @param force Whether to force the installation of packages.
#' Default is `FALSE`.
#'
#' @return Package installation status.
#'
#' @export
check_r <- function(
    packages,
    lib = .libPaths()[1],
    dependencies = TRUE,
    force = FALSE,
    verbose = TRUE) {
  status_list <- list()
  for (pkg in packages) {
    version <- NULL
    if (grepl("/", pkg)) {
      pkg_name <- strsplit(pkg, split = "/|@|==", perl = TRUE)[[1]][[2]]
    } else {
      pkg_info <- strsplit(pkg, split = "@|==", perl = TRUE)[[1]]
      pkg_name <- pkg_info[[1]]
      if (length(pkg_info) > 1) {
        version <- pkg_info[[2]]
      }
    }
    check_pkg <- check_pkg_status(
      pkg_name,
      version = version,
      lib = lib
    )

    force_update <- FALSE
    if (check_pkg && !is.null(version)) {
      current_version <- utils::packageVersion(pkg_name)
      force_update <- current_version < package_version(version)
    }
    force_update <- force_update || isTRUE(force)

    if (!check_pkg || force_update) {
      log_message(
        "Installing: {.pkg {pkg_name}}...",
        message_type = "running",
        verbose = verbose
      )
      status_list[[pkg]] <- FALSE
      tryCatch(
        expr = {
          if (!dir.exists(lib)) {
            dir.create(lib, recursive = TRUE, showWarnings = FALSE)
          }
          if (isTRUE(verbose)) {
            pak::pkg_install(
              pkg,
              lib = lib,
              ask = FALSE,
              dependencies = dependencies
            )
          } else {
            invisible(
              suppressMessages(
                pak::pkg_install(
                  pkg,
                  lib = lib,
                  ask = FALSE,
                  dependencies = dependencies
                )
              )
            )
          }
        },
        error = function(e) {
          status_list[[pkg]] <- FALSE
          log_message(
            "Failed to install: {.pkg {pkg_name}}. Error: {.val {e$message}}",
            message_type = "warning",
            verbose = verbose
          )
        }
      )
      status_list[[pkg]] <- check_pkg_status(
        pkg_name,
        version = version,
        lib = lib
      )
    } else {
      status_list[[pkg]] <- TRUE
    }
  }

  success <- sapply(status_list, isTRUE)
  failed <- names(status_list)[!success]

  if (length(failed) > 0) {
    log_message(
      "Failed to install: {.pkg {failed}}. Please install manually",
      message_type = "warning",
      verbose = verbose
    )
  } else {
    log_message(
      "{.pkg {packages}} installed successfully",
      message_type = "success",
      verbose = verbose
    )
  }

  return(invisible(status_list))
}

#' @title Check and remove R packages
#'
#' @md
#' @inheritParams log_message
#' @param packages Package to be removed.
#' @param lib The location of the library directories where to remove the packages.
#'
#' @export
remove_r <- function(
    packages,
    lib = .libPaths()[1],
    verbose = TRUE) {
  status_list <- list()
  for (pkg in packages) {
    pkg_installed <- check_pkg_status(pkg, lib = lib)

    if (pkg_installed) {
      log_message(
        "Removing: {.pkg {pkg}}...",
        verbose = verbose
      )
      status_list[[pkg]] <- FALSE
      tryCatch(
        expr = {
          pak::pkg_remove(pkg, lib = lib)
        },
        error = function(e) {
          log_message(
            "Warning during removal: {.pkg {pkg}}. Error: {.val {e$message}}",
            message_type = "warning",
            verbose = verbose
          )
        }
      )
      status_list[[pkg]] <- !check_pkg_status(pkg, lib = lib)
    } else {
      log_message(
        "{.pkg {pkg}} is not installed, skipping removal",
        message_type = "warning",
        verbose = verbose
      )
      status_list[[pkg]] <- TRUE
    }
  }

  success <- sapply(status_list, isTRUE)
  failed <- names(status_list)[!success]

  if (length(failed) > 0) {
    log_message(
      "Failed to remove: {.pkg {failed}}. Please remove manually",
      message_type = "warning",
      verbose = verbose
    )
  } else {
    log_message(
      "{.pkg {packages}} removed successfully",
      message_type = "success",
      verbose = verbose
    )
  }

  return(invisible(status_list))
}

#' @title Check if a package is installed with the specified version
#'
#' @md
#' @inheritParams check_r
#' @param pkg Package name.
#' @param version Package version to check. If `NULL`, only checks if the package is installed.
#'
#' @return `TRUE` if the package is installed with the specified version, `FALSE` otherwise.
#'
#' @export
check_pkg_status <- function(
    pkg,
    version = NULL,
    lib = .libPaths()[1]) {
  installed_pkgs_info <- utils::installed.packages(lib.loc = lib)
  installed_pkgs <- installed_pkgs_info[, "Package"]
  installed_pkgs_version <- installed_pkgs_info[, "Version"]
  pkg_exists <- pkg %in% installed_pkgs
  if (is.null(version)) {
    version_match <- TRUE
  } else {
    version_match <- installed_pkgs_version[installed_pkgs == pkg] == version
  }

  if (isFALSE(pkg_exists) || isFALSE(version_match)) {
    return(FALSE)
  }

  TRUE
}

#' @title Collapse sparse matrix rows by group
#'
#' @param matrix A sparse matrix.
#' @param group A vector defining the output row groups.
#'
#' @return A sparse matrix with rows collapsed by `group`.
#'
#' @export
#' @examples
#' mat <- Matrix::Matrix(
#'   matrix(c(1, 0, 2, 0, 3, 4), nrow = 3, byrow = TRUE),
#'   sparse = TRUE
#' )
#' collapse_sparse_rows(mat, c("g1", "g1", "g2"))
collapse_sparse_rows <- function(matrix, group) {
  if (length(group) != nrow(matrix)) {
    cli::cli_abort(
      "{.arg group} length must match the number of rows of {.arg matrix}"
    )
  }

  group <- as.character(group)
  keep <- !is.na(group) & nzchar(group)
  matrix <- matrix[keep, , drop = FALSE]
  group <- group[keep]

  levels_use <- unique(group)
  matrix_summary <- Matrix::summary(matrix)
  i_new <- match(group[matrix_summary$i], levels_use)

  Matrix::sparseMatrix(
    i = i_new,
    j = matrix_summary$j,
    x = matrix_summary$x,
    dims = c(length(levels_use), ncol(matrix)),
    dimnames = list(levels_use, colnames(matrix))
  )
}

test_that("collapse_sparse_rows aggregates rows by group", {
  mat <- Matrix::Matrix(
    matrix(
      c(1, 0, 2, 0, 3, 4),
      nrow = 3,
      byrow = TRUE,
      dimnames = list(c("a", "b", "c"), c("x", "y"))
    ),
    sparse = TRUE
  )

  out <- collapse_sparse_rows(mat, c("g1", "g1", "g2"))

  expect_s4_class(out, "dgCMatrix")
  expect_identical(rownames(out), c("g1", "g2"))
  expect_equal(
    unname(as.matrix(out)),
    matrix(c(3, 0, 3, 4), nrow = 2, byrow = TRUE)
  )
})

test_that("collapse_sparse_rows drops missing and empty groups", {
  mat <- Matrix::Matrix(
    matrix(
      c(1, 0, 2, 0, 3, 4, 5, 6),
      nrow = 4,
      byrow = TRUE,
      dimnames = list(c("a", "b", "c", "d"), c("x", "y"))
    ),
    sparse = TRUE
  )

  out <- collapse_sparse_rows(mat, c("g1", NA, "", "g2"))

  expect_identical(rownames(out), c("g1", "g2"))
  expect_equal(
    unname(as.matrix(out)),
    matrix(c(1, 0, 5, 6), nrow = 2, byrow = TRUE)
  )
})

test_that("collapse_sparse_rows errors when group length mismatches matrix rows", {
  mat <- Matrix::Matrix(matrix(1:4, nrow = 2), sparse = TRUE)

  expect_error(
    collapse_sparse_rows(mat, "g1"),
    "length must match"
  )
})

#' @title Compute Local Inverse Simpson's Index (LISI)
#'
#' @description
#' Compute per-cell Local Inverse Simpson's Index (LISI) scores for one or more categorical variables.
#' This is a clean-room reimplementation of the `immunogenomics/LISI`.
#'
#' @md
#' @param X A matrix-like object with cells in rows and embedding/features in columns.
#' @param meta_data A data frame with one row per cell.
#' @param label_colnames Character vector of column names in `meta_data` to evaluate.
#' @param perplexity Effective neighborhood size. Defaults to `30`.
#' @param nn_eps Approximation factor passed to [RANN::nn2] when `RANN` is available and `use_rann = TRUE`.
#' Defaults to `0`.
#' @param use_rann Whether to prefer [RANN::nn2] over `FNN::get.knn` when `nn_method = "auto"` decides not to use the package's built-in exact C++ backend.
#' Defaults to `TRUE`.
#' @param nn_method Nearest-neighbor backend. Defaults to `"auto"`,
#' which prefers the package's exact C++ search only for very small problems.
#' Larger inputs fall back to `RANN`, then `FNN`, and only use the exact backend as a last resort.
#' @param tol Tolerance used in the binary search for the target perplexity.
#' Defaults to `1e-5`.
#' @param max_iter Maximum number of binary-search iterations.
#' Defaults to `50`.
#'
#' @return A data frame with one row per cell and one column per label.
#' @export
#'
#' @references
#' Korsunsky I, Millard N, Fan J, et al. Fast, sensitive and accurate
#' integration of single-cell data with Harmony. *Nature Methods* (2019).
#' \url{https://www.nature.com/articles/s41592-019-0619-0}
#'
#' LISI reference implementation:
#' \url{https://github.com/immunogenomics/LISI}
#'
#' @examples
#' set.seed(1)
#' X <- rbind(
#'   matrix(stats::rnorm(100, mean = -2), ncol = 5),
#'   matrix(stats::rnorm(100, mean = 2), ncol = 5)
#' )
#' meta_data <- data.frame(
#'   batch = rep(c("A", "B"), each = 20),
#'   group = sample(c("g1", "g2"), 40, replace = TRUE)
#' )
#'
#' res <- compute_lisi(
#'   X, meta_data,
#'   c("batch", "group"),
#'   perplexity = 10
#' )
#' head(res)
#' boxplot(res)
compute_lisi <- function(
  X,
  meta_data,
  label_colnames,
  perplexity = 30,
  nn_eps = 0,
  use_rann = TRUE,
  nn_method = c("auto", "rann", "fnn", "exact"),
  tol = 1e-5,
  max_iter = 50
) {
  X <- validate_lisi_input_matrix(X)
  meta_data <- as.data.frame(meta_data, stringsAsFactors = FALSE)

  if (nrow(X) != nrow(meta_data)) {
    log_message(
      "{.arg X} and {.arg meta_data} must have the same number of rows",
      message_type = "error"
    )
  }
  if (length(label_colnames) == 0) {
    log_message(
      "{.arg label_colnames} must contain at least one column name",
      message_type = "error"
    )
  }
  if (!all(label_colnames %in% colnames(meta_data))) {
    missing_cols <- setdiff(label_colnames, colnames(meta_data))
    log_message(
      "The following label columns are missing: {.val {missing_cols}}",
      message_type = "error"
    )
  }
  if (
    !is.numeric(perplexity) ||
      length(perplexity) != 1 ||
      is.na(perplexity) ||
      perplexity <= 0
  ) {
    log_message(
      "{.arg perplexity} must be a single positive number",
      message_type = "error"
    )
  }
  if (nrow(X) < 2) {
    log_message(
      "At least two cells are required to compute {.pkg LISI}",
      message_type = "error"
    )
  }

  nn_method <- match.arg(nn_method)
  n_neighbors <- min(
    nrow(X) - 1L,
    max(1L, as.integer(ceiling(perplexity * 3)) - 1L)
  )
  knn <- lisi_knn(
    X = X,
    n_neighbors = n_neighbors,
    nn_eps = nn_eps,
    use_rann = use_rann,
    nn_method = nn_method
  )

  out <- lapply(label_colnames, function(label_colname) {
    labels <- meta_data[[label_colname]]

    if (any(is.na(labels))) {
      log_message(
        "Cannot compute {.pkg LISI} for {.val {label_colname}} because it contains missing values",
        message_type = "warning"
      )
      return(rep(NA_real_, nrow(X)))
    }

    label_ids <- as.integer(factor(labels))
    simpson <- compute_simpson_index(
      D = t(knn$nn.dists),
      knn_idx = t(knn$nn.idx),
      batch_labels = label_ids,
      perplexity = perplexity,
      tol = tol,
      max_iter = max_iter
    )

    1 / simpson
  })

  out <- as.data.frame(out, optional = TRUE, stringsAsFactors = FALSE)
  colnames(out) <- label_colnames
  rownames(out) <- rownames(meta_data) %ss% rownames(X)
  out
}

#' @title Compute Simpson index from a KNN graph
#'
#' @description
#' Given nearest-neighbor distances, nearest-neighbor indices,
#' and a categorical label for each cell,
#' compute the local Simpson index of each cell after matching the target perplexity.
#'
#' @md
#' @param D Numeric matrix of nearest-neighbor distances with neighbors in rows
#'   and cells in columns.
#' @param knn_idx Integer matrix of nearest-neighbor indices with the same shape as `D`.
#' One-based indices are expected; zero-based indices are also accepted and converted automatically.
#' @param batch_labels Integer-like label vector of length equal to the number of cells.
#' @param perplexity Effective neighborhood size.
#' Defaults to `15`.
#' @param tol Tolerance used in the binary search for the target perplexity.
#' Defaults to `1e-5`.
#' @param max_iter Maximum number of binary-search iterations.
#' Defaults to `50`.
#'
#' @return A numeric vector containing the local Simpson index for each cell.
#' @export
#'
#' @examples
#' D <- matrix(
#'   c(0.1, 0.2, 0.2, 0.1,
#'     0.3, 0.4, 0.4, 0.3),
#'   nrow = 2,
#'   byrow = TRUE
#' )
#' knn_idx <- matrix(
#'   c(2, 1, 4, 3,
#'     3, 4, 2, 1),
#'   nrow = 2,
#'   byrow = TRUE
#' )
#' batch_labels <- c(1, 1, 2, 2)
#' compute_simpson_index(D, knn_idx, batch_labels, perplexity = 2)
compute_simpson_index <- function(
  D,
  knn_idx,
  batch_labels,
  perplexity = 15,
  tol = 1e-5,
  max_iter = 50
) {
  D <- as.matrix(D)
  knn_idx <- as.matrix(knn_idx)
  batch_labels <- as.integer(as.vector(batch_labels))

  if (!is.numeric(D)) {
    log_message(
      "D must be a numeric matrix",
      message_type = "error"
    )
  }
  if (!is.numeric(knn_idx)) {
    log_message(
      "knn_idx must be an integer/numeric matrix",
      message_type = "error"
    )
  }
  if (!identical(dim(D), dim(knn_idx))) {
    log_message(
      "D and knn_idx must have the same dimensions",
      message_type = "error"
    )
  }
  if (ncol(D) != length(batch_labels)) {
    log_message(
      "length(batch_labels) must equal ncol(D)",
      message_type = "error"
    )
  }
  if (
    !is.numeric(perplexity) ||
      length(perplexity) != 1 ||
      is.na(perplexity) ||
      perplexity <= 0
  ) {
    log_message(
      "perplexity must be a single positive number",
      message_type = "error"
    )
  }

  storage.mode(knn_idx) <- "integer"

  compute_simpson_index_cpp(
    D = D,
    knn_idx = knn_idx,
    batch_labels = batch_labels,
    perplexity = perplexity,
    tol = tol,
    max_iter = as.integer(max_iter)
  )
}

validate_lisi_input_matrix <- function(X) {
  if (inherits(X, "sparseMatrix")) {
    X <- Matrix::as.matrix(X)
  }

  X <- as.matrix(X)

  if (!is.numeric(X)) {
    log_message(
      "X must be numeric",
      message_type = "error"
    )
  }

  X
}

lisi_knn <- function(
  X,
  n_neighbors,
  nn_eps = 0,
  use_rann = TRUE,
  nn_method = "auto"
) {
  include_self <- FALSE
  request_k <- as.integer(n_neighbors)

  if (identical(nn_method, "auto")) {
    nn_method <- choose_lisi_nn_method(
      n = nrow(X),
      p = ncol(X),
      n_neighbors = request_k,
      use_rann = use_rann
    )
  }

  log_message(
    "Using {.val {nn_method}} nearest-neighbor backend for {.pkg compute_lisi}",
    message_type = "running"
  )

  knn <- switch(
    EXPR = nn_method,
    "rann" = {
      include_self <- TRUE
      request_k <- request_k + 1L
      RANN::nn2(data = X, query = X, k = request_k, eps = nn_eps)
    },
    "fnn" = {
      res <- FNN::get.knn(X, k = request_k)
      list(nn.idx = res$nn.index, nn.dists = res$nn.dist)
    },
    "exact" = {
      include_self <- TRUE
      request_k <- request_k + 1L
      lisi_exact_knn_cpp(X, as.integer(request_k))
    },
    log_message(
      "Unsupported nn_method: {.val {nn_method}}",
      message_type = "error"
    )
  )

  if (include_self) {
    storage.mode(knn$nn.idx) <- "integer"
    knn <- drop_self_from_knn_cpp(knn$nn.idx, as.matrix(knn$nn.dists))
  }

  knn
}

choose_lisi_nn_method <- function(
  n,
  p,
  n_neighbors,
  use_rann = TRUE
) {
  exact_score <- as.double(n) * as.double(n) * as.double(p)
  small_exact_problem <- (
    n <= 250L &&
      p <= 20L &&
      n_neighbors <= 64L &&
      exact_score <= 1.5e6
  )

  if (small_exact_problem) {
    return("exact")
  }

  if (isTRUE(use_rann) && requireNamespace("RANN", quietly = TRUE)) {
    return("rann")
  }

  if (requireNamespace("FNN", quietly = TRUE)) {
    return("fnn")
  }

  if (n <= 1000L && exact_score <= 5e7) {
    return("exact")
  }

  "exact"
}

test_that("compute_lisi produces correct output shape", {
  set.seed(1)
  X <- rbind(
    matrix(rnorm(100, mean = -2), ncol = 5),
    matrix(rnorm(100, mean = 2), ncol = 5)
  )
  meta_data <- data.frame(
    batch = rep(c("A", "B"), each = 20)
  )
  result <- suppressMessages(
    compute_lisi(X, meta_data, "batch", perplexity = 10)
  )
  expect_true(is.data.frame(result))
  expect_equal(nrow(result), 40)
  expect_equal(ncol(result), 1)
  expect_equal(colnames(result), "batch")
})

test_that("compute_lisi works with multiple labels", {
  set.seed(1)
  X <- rbind(
    matrix(rnorm(100, mean = -2), ncol = 5),
    matrix(rnorm(100, mean = 2), ncol = 5)
  )
  meta_data <- data.frame(
    batch = rep(c("A", "B"), each = 20),
    group = sample(c("g1", "g2"), 40, replace = TRUE)
  )
  result <- suppressMessages(
    compute_lisi(X, meta_data, c("batch", "group"), perplexity = 10)
  )
  expect_equal(ncol(result), 2)
  expect_equal(colnames(result), c("batch", "group"))
})

test_that("compute_lisi errors on mismatched rows", {
  X <- matrix(rnorm(20), ncol = 2)
  meta <- data.frame(batch = c("A", "B", "A"))
  expect_error(
    suppressMessages(compute_lisi(X, meta, "batch")),
    "same number of rows"
  )
})

test_that("compute_lisi errors on missing label column", {
  X <- matrix(rnorm(20), ncol = 2)
  meta <- data.frame(batch = rep("A", 10))
  expect_error(
    suppressMessages(compute_lisi(X, meta, "nonexistent")),
    "missing"
  )
})

test_that("compute_lisi errors on invalid perplexity", {
  X <- matrix(rnorm(20), ncol = 2)
  meta <- data.frame(batch = rep("A", 10))
  expect_error(
    suppressMessages(compute_lisi(X, meta, "batch", perplexity = -1)),
    "positive"
  )
})

test_that("compute_simpson_index works", {
  D <- matrix(
    c(0.1, 0.2, 0.2, 0.1,
      0.3, 0.4, 0.4, 0.3),
    nrow = 2, byrow = TRUE
  )
  knn_idx <- matrix(
    c(2, 1, 4, 3,
      3, 4, 2, 1),
    nrow = 2, byrow = TRUE
  )
  batch_labels <- c(1, 1, 2, 2)
  result <- compute_simpson_index(D, knn_idx, batch_labels, perplexity = 2)
  expect_true(is.numeric(result))
  expect_equal(length(result), 4)
  expect_true(all(result > 0 & result <= 1, na.rm = TRUE))
})

test_that("compute_simpson_index errors on dimension mismatch", {
  D <- matrix(1:6, nrow = 2)
  knn_idx <- matrix(1:4, nrow = 2)
  expect_error(
    compute_simpson_index(D, knn_idx, c(1, 1, 1)),
    "same dimensions"
  )
})

compute_lisi_original_like_test <- function(X, meta_data, label_colnames, perplexity = 30, nn_eps = 0) {
  N <- nrow(meta_data)
  dknn <- RANN::nn2(X, k = perplexity * 3, eps = nn_eps)

  out <- lapply(label_colnames, function(label_colname) {
    labels <- data.frame(meta_data)[, label_colname, drop = TRUE]
    if (any(is.na(labels))) {
      return(rep(NA_real_, N))
    }

    nn_idx <- dknn$nn.idx[, 2:ncol(dknn$nn.idx), drop = FALSE]
    nn_dists <- dknn$nn.dists[, 2:ncol(dknn$nn.dists), drop = FALSE]
    labels <- as.integer(factor(labels)) - 1L
    n_batches <- length(unique(labels))

    simpson <- compute_simpson_index_original_like_test(
      t(nn_dists),
      t(nn_idx) - 1L,
      labels,
      n_batches,
      perplexity
    )

    1 / simpson
  })

  out <- as.data.frame(out, optional = TRUE, stringsAsFactors = FALSE)
  colnames(out) <- label_colnames
  out
}

compute_simpson_index_original_like_test <- function(D, knn_idx, batch_labels, n_batches, perplexity = 15, tol = 1e-5) {
  D <- as.matrix(D)
  knn_idx <- as.matrix(knn_idx)
  n <- ncol(D)
  simpson <- numeric(n)
  logU <- log(perplexity)

  hbeta <- function(D, beta, idx) {
    P <- exp(-D[, idx] * beta)
    sumP <- sum(P)
    if (sumP == 0) {
      list(H = 0, P = D[, idx] * 0)
    } else {
      H <- log(sumP) + beta * sum(D[, idx] * P) / sumP
      list(H = H, P = P / sumP)
    }
  }

  for (cell in seq_len(n)) {
    beta <- 1
    betamin <- -Inf
    betamax <- Inf
    hb <- hbeta(D, beta, cell)
    H <- hb$H
    P <- hb$P
    Hdiff <- H - logU
    tries <- 0

    while (abs(Hdiff) > tol && tries < 50) {
      if (Hdiff > 0) {
        betamin <- beta
        beta <- if (!is.finite(betamax)) beta * 2 else (beta + betamax) / 2
      } else {
        betamax <- beta
        beta <- if (!is.finite(betamin)) beta / 2 else (beta + betamin) / 2
      }

      hb <- hbeta(D, beta, cell)
      H <- hb$H
      P <- hb$P
      Hdiff <- H - logU
      tries <- tries + 1
    }

    if (H == 0) {
      simpson[cell] <- -1
      next
    }

    for (b in 0:(n_batches - 1L)) {
      q <- which(batch_labels[knn_idx[, cell] + 1L] == b)
      if (length(q) > 0) {
        sumP <- sum(P[q])
        simpson[cell] <- simpson[cell] + sumP * sumP
      }
    }
  }

  simpson
}

test_that("choose_lisi_nn_method prefers exact only for very small problems", {
  expect_equal(
    choose_lisi_nn_method(n = 200, p = 10, n_neighbors = 30, use_rann = TRUE),
    "exact"
  )
  expect_equal(
    choose_lisi_nn_method(n = 300, p = 10, n_neighbors = 30, use_rann = TRUE),
    "rann"
  )
  expect_equal(
    choose_lisi_nn_method(n = 1000, p = 30, n_neighbors = 90, use_rann = TRUE),
    "rann"
  )
  expect_equal(
    choose_lisi_nn_method(n = 1000, p = 30, n_neighbors = 90, use_rann = FALSE),
    "fnn"
  )
})

test_that("compute_lisi exact, rann and fnn agree on small data", {
  set.seed(11)
  X <- rbind(
    matrix(rnorm(1000, mean = -1), ncol = 10),
    matrix(rnorm(1000, mean = 1), ncol = 10)
  )
  meta <- data.frame(
    batch = rep(c("A", "B"), each = nrow(X) / 2),
    group = sample(c("g1", "g2", "g3"), nrow(X), replace = TRUE)
  )

  exact_res <- suppressMessages(compute_lisi(X, meta, c("batch", "group"), perplexity = 10, nn_method = "exact"))
  rann_res <- suppressMessages(compute_lisi(X, meta, c("batch", "group"), perplexity = 10, nn_method = "rann"))
  fnn_res <- suppressMessages(compute_lisi(X, meta, c("batch", "group"), perplexity = 10, nn_method = "fnn"))

  expect_equal(rann_res, exact_res, tolerance = 1e-10)
  expect_equal(fnn_res, exact_res, tolerance = 1e-10)
})

test_that("compute_lisi rann matches original LISI implementation path", {
  set.seed(44)
  X <- rbind(
    matrix(rnorm(2000, mean = -1), ncol = 10),
    matrix(rnorm(2000, mean = 1), ncol = 10)
  )
  meta <- data.frame(
    batch = rep(c("A", "B"), each = nrow(X) / 2),
    group = sample(c("g1", "g2", "g3"), nrow(X), replace = TRUE)
  )

  ours <- suppressMessages(compute_lisi(X, meta, c("batch", "group"), perplexity = 10, nn_method = "rann"))
  orig <- compute_lisi_original_like_test(X, meta, c("batch", "group"), perplexity = 10, nn_eps = 0)

  expect_equal(colnames(ours), colnames(orig))
  expect_equal(unname(as.matrix(ours)), unname(as.matrix(orig)), tolerance = 1e-12)
})

test_that("compute_lisi auto matches explicit backend strategy on different sizes", {
  set.seed(22)
  X_small <- rbind(
    matrix(rnorm(1000, mean = -1), ncol = 10),
    matrix(rnorm(1000, mean = 1), ncol = 10)
  )
  meta_small <- data.frame(
    batch = rep(c("A", "B"), each = nrow(X_small) / 2)
  )

  auto_small <- suppressMessages(compute_lisi(X_small, meta_small, "batch", perplexity = 10, nn_method = "auto"))
  exact_small <- suppressMessages(compute_lisi(X_small, meta_small, "batch", perplexity = 10, nn_method = "exact"))
  expect_equal(auto_small, exact_small, tolerance = 1e-10)

  X_medium <- rbind(
    matrix(rnorm(6000, mean = -1), ncol = 20),
    matrix(rnorm(6000, mean = 1), ncol = 20)
  )
  meta_medium <- data.frame(
    batch = rep(c("A", "B"), each = nrow(X_medium) / 2)
  )

  auto_medium <- suppressMessages(compute_lisi(X_medium, meta_medium, "batch", perplexity = 15, nn_method = "auto"))
  rann_medium <- suppressMessages(compute_lisi(X_medium, meta_medium, "batch", perplexity = 15, nn_method = "rann"))
  expect_equal(auto_medium, rann_medium, tolerance = 1e-10)
})

#' @title Sparse correlation function
#'
#' @description
#' Safe correlation function which returns a sparse matrix.
#'
#' @md
#' @param x Sparse matrix or character vector.
#' @param y Sparse matrix or character vector.
#' @param method Method to use for calculating the correlation coefficient.
#' @param allow_neg Logical. Whether to allow negative values or set them to 0.
#' @param remove_na Logical. Whether to replace NA values with 0.
#' @param remove_inf Logical. Whether to replace infinite values with 1.
#' @param ... Other arguments passed to [stats::cor] function.
#'
#' @return A correlation matrix.
#'
#' @export
#'
#' @examples
#' m1 <- simulate_sparse_matrix(
#'   500, 100
#' )
#' m2 <- simulate_sparse_matrix(
#'   500, 100,
#'   seed = 2025
#' )
#' a <- sparse_cor(m1)
#' b <- sparse_cor(m1, m2)
#' c <- as_matrix(
#'   cor(as_matrix(m1)),
#'   return_sparse = TRUE
#' )
#' d <- as_matrix(
#'   cor(as_matrix(m1), as_matrix(m2)),
#'   return_sparse = TRUE
#' )
#'
#' a[1:5, 1:5]
#' c[1:5, 1:5]
#' all.equal(a, c)
#'
#' b[1:5, 1:5]
#' d[1:5, 1:5]
#' all.equal(b, d)
#'
#' m1[sample(1:500, 10)] <- NA
#' m2[sample(1:500, 10)] <- NA
#'
#' sparse_cor(m1, m2)[1:5, 1:5]
#'
#' system.time(
#'   sparse_cor(m1)
#' )
#' system.time(
#'   cor(as_matrix(m1))
#' )
#'
#' system.time(
#'   sparse_cor(m1, m2)
#' )
#' system.time(
#'   cor(as_matrix(m1), as_matrix(m2))
#' )
sparse_cor <- function(
    x,
    y = NULL,
    method = "pearson",
    allow_neg = TRUE,
    remove_na = TRUE,
    remove_inf = TRUE,
    ...) {
  if (!inherits(x, "sparseMatrix")) {
    x <- as_matrix(x, return_sparse = TRUE)
  }

  if (!is.null(y)) {
    if (!inherits(y, "sparseMatrix")) {
      y <- as_matrix(y, return_sparse = TRUE)
    }
    if (nrow(x) != nrow(y)) {
      log_message(
        "{.arg x} and {.arg y} must have the same number of rows",
        message_type = "error"
      )
    }
  }

  corr_mat <- switch(
    EXPR = method,
    "pearson" = pearson_correlation(x, y)$cor,
    "spearman" = {
      if (is.null(y)) {
        stats::cor(
          as_matrix(x),
          method = "spearman",
          ...
        )
      } else {
        stats::cor(
          as_matrix(x),
          as_matrix(y),
          method = "spearman",
          ...
        )
      }
    },
    "kendall" = {
      if (is.null(y)) {
        stats::cor(
          as_matrix(x),
          method = "kendall",
          ...
        )
      } else {
        stats::cor(
          as_matrix(x),
          as_matrix(y),
          method = "kendall",
          ...
        )
      }
    }
  )

  if (is.null(y)) {
    colnames(corr_mat) <- colnames(x)
  } else {
    colnames(corr_mat) <- colnames(y)
  }
  rownames(corr_mat) <- colnames(x)

  if (remove_na) {
    corr_mat[is.na(corr_mat)] <- 0
  }
  if (remove_inf) {
    corr_mat[is.infinite(corr_mat)] <- 1
  }

  corr_mat <- as_matrix(corr_mat, return_sparse = TRUE)

  if (!allow_neg) {
    corr_mat[corr_mat < 0] <- 0
  }

  return(corr_mat)
}

#' @title Correlation and covariance calculation for sparse matrix
#'
#' @inheritParams sparse_cor
#'
#' @return A list with covariance and correlation matrices.
#'
#' @export
#'
#' @examples
#' m1 <- simulate_sparse_matrix(
#'   100, 100
#' )
#' m2 <- simulate_sparse_matrix(
#'   100, 100,
#'   sparsity = 0.05
#' )
#' a <- pearson_correlation(m1, m2)
#' a$cov[1:5, 1:5]
#' a$cor[1:5, 1:5]
pearson_correlation <- function(x, y = NULL) {
  if (!inherits(x, "sparseMatrix")) {
    log_message(
      "x should be a sparse matrix",
      message_type = "error"
    )
  }
  if (!is.null(y) && !inherits(y, "sparseMatrix")) {
    log_message(
      "y should be a sparse matrix",
      message_type = "error"
    )
  }

  n <- nrow(x)
  mu_x <- Matrix::colMeans(x)
  if (is.null(y)) {
    covmat <- (
      (as.matrix(Matrix::crossprod(x)) - n * Matrix::tcrossprod(mu_x)) / (n - 1)
    )
    sdvec <- sqrt(diag(covmat))
    cormat <- covmat / tcrossprod(sdvec)
  } else {
    if (!inherits(y, "sparseMatrix")) {
      log_message(
        "y should be a sparse matrix",
        message_type = "error"
      )
    }
    if (nrow(x) != nrow(y)) {
      log_message(
        "x and y should have the same number of rows",
        message_type = "error"
      )
    }

    mu_y <- Matrix::colMeans(y)
    covmat <- (
      (as.matrix(Matrix::crossprod(x, y)) - n * Matrix::tcrossprod(mu_x, mu_y)) / (n - 1)
    )
    sdvecX <- sqrt((Matrix::colSums(x^2) - n * mu_x^2) / (n - 1))
    sdvecY <- sqrt((Matrix::colSums(y^2) - n * mu_y^2) / (n - 1))
    cormat <- covmat / Matrix::tcrossprod(sdvecX, sdvecY)
  }

  return(
    list(
      cov = covmat,
      cor = cormat
    )
  )
}

test_that("pearson_correlation works for single matrix", {
  m <- simulate_sparse_matrix(50, 10, sparsity = 0.5)
  result <- pearson_correlation(m)
  expect_true(is.list(result))
  expect_true("cov" %in% names(result))
  expect_true("cor" %in% names(result))
  expect_equal(nrow(result$cor), 10)
  expect_equal(ncol(result$cor), 10)
})

test_that("pearson_correlation works for two matrices", {
  m1 <- simulate_sparse_matrix(50, 10, sparsity = 0.5)
  m2 <- simulate_sparse_matrix(50, 8, sparsity = 0.5, seed = 2)
  result <- pearson_correlation(m1, m2)
  expect_equal(nrow(result$cor), 10)
  expect_equal(ncol(result$cor), 8)
})

test_that("pearson_correlation errors on non-sparse input", {
  expect_error(pearson_correlation(matrix(1:20, 4, 5)))
})

test_that("sparse_cor works with pearson method", {
  m <- simulate_sparse_matrix(50, 10, sparsity = 0.5)
  result <- sparse_cor(m, method = "pearson")
  expect_true(inherits(result, "sparseMatrix"))
  expect_equal(nrow(result), 10)
})

test_that("sparse_cor works with two matrices", {
  m1 <- simulate_sparse_matrix(50, 10, sparsity = 0.5)
  m2 <- simulate_sparse_matrix(50, 8, sparsity = 0.5, seed = 2)
  result <- sparse_cor(m1, m2, method = "pearson")
  expect_equal(nrow(result), 10)
  expect_equal(ncol(result), 8)
})

test_that("sparse_cor errors on row mismatch", {
  m1 <- simulate_sparse_matrix(50, 10, sparsity = 0.5)
  m2 <- simulate_sparse_matrix(30, 10, sparsity = 0.5, seed = 2)
  expect_error(sparse_cor(m1, m2))
})

test_that("sparse_cor removes NA values by default", {
  m <- simulate_sparse_matrix(50, 10, sparsity = 0.5)
  result <- sparse_cor(m, remove_na = TRUE)
  expect_false(any(is.na(as.matrix(result))))
})

test_that("sparse_cor allow_neg = FALSE removes negatives", {
  m <- simulate_sparse_matrix(50, 10, sparsity = 0.5)
  result <- sparse_cor(m, allow_neg = FALSE)
  expect_true(all(result >= 0))
})

#' @title Convert matrix into dense/sparse matrix
#'
#' @md
#' @param x A matrix.
#' @param return_sparse Whether to output a sparse matrix.
#' Default is `FALSE`.
#'
#' @return
#' A dense or sparse matrix.
#'
#' @export
#'
#' @examples
#' m <- simulate_sparse_matrix(
#'   1000, 1000,
#'   decimal = 3
#' )
#'
#' a <- as_matrix(m)
#' a[1:5, 1:5]
#'
#' b <- as_matrix(m, return_sparse = TRUE)
#' b[1:5, 1:5]
as_matrix <- function(
    x,
    return_sparse = FALSE) {
  if (!inherits(x, "sparseMatrix")) {
    if (return_sparse) {
      return(
        Matrix::Matrix(x, sparse = TRUE, dimnames = dimnames(x))
      )
    } else {
      return(
        Matrix::as.matrix(x)
      )
    }
  }

  if (return_sparse) {
    return(x)
  }

  return(
    Matrix::as.matrix(x)
  )
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

compute_simpson_index_cpp <- function(D, knn_idx, batch_labels, perplexity = 15, tol = 1e-5, max_iter = 50L) {
    .Call(`_thisutils_compute_simpson_index_cpp`, D, knn_idx, batch_labels, perplexity, tol, max_iter)
}

lisi_exact_knn_cpp <- function(X, k) {
    .Call(`_thisutils_lisi_exact_knn_cpp`, X, k)
}

drop_self_from_knn_cpp <- function(nn_idx, nn_dists) {
    .Call(`_thisutils_drop_self_from_knn_cpp`, nn_idx, nn_dists)
}

#' @title Switch matrix to table
#'
#' @md
#' @param matrix A matrix.
#' @param row_names Character vector of row names to filter by.
#' @param col_names Character vector of column names to filter by.
#' @param threshold The threshold for filtering values based on absolute values.
#' Defaults to `0`.
#' @param keep_zero Whether to keep zero values in the table. Defaults to `false`.
#'
#' @return A table with three columns: `row`, `col`, and `value`.
#' @export
#' @seealso [table_to_matrix]
#'
#' @examples
#' test_matrix <- simulate_sparse_matrix(10, 10)
#' colnames(test_matrix) <- paste0("c", 1:10)
#' rownames(test_matrix) <- paste0("r", 1:10)
#' table <- matrix_to_table(test_matrix)
#' matrix_new <- table_to_matrix(table)
#' test_matrix <- test_matrix[rownames(matrix_new), colnames(matrix_new)] |>
#'   as_matrix()
#' identical(test_matrix, matrix_new)
#'
#' matrix_to_table(
#'   test_matrix,
#'   threshold = 2
#' )
#'
#' matrix_to_table(
#'   test_matrix,
#'   row_names = c("r1", "r2"),
#'   col_names = c("c1", "c2")
#' )
matrix_to_table <- function(matrix, row_names = NULL, col_names = NULL, threshold = 0.0, keep_zero = TRUE) {
    .Call(`_thisutils_matrix_to_table`, matrix, row_names, col_names, threshold, keep_zero)
}

#' @title Split indices.
#'
#' @description An optimised version of split for the special case of splitting row indices into groups.
#'
#' @param group Integer indices
#' @param n The largest integer (may not appear in index).
#' This is hint: if the largest value of \code{group} is bigger than \code{n},
#' the output will silently expand.
#'
#' @return A list of vectors of indices.
#'
#' @references
#' \url{https://github.com/hadley/plyr/blob/d57f9377eb5d56107ba3136775f2f0f005f33aa3/src/split-numeric.cpp#L20}
#'
#' @export
#' @examples
#' split_indices(sample(10, 100, rep = TRUE))
#' split_indices(sample(10, 100, rep = TRUE), 10)
split_indices <- function(group, n = 0L) {
    .Call(`_thisutils_split_indices`, group, n)
}

#' @title Switch table to matrix
#'
#' @md
#' @param table A table with three columns: `row`, `col`, and `value`.
#' @param row_names Character vector of row names to filter by.
#' @param col_names Character vector of column names to filter by.
#' @param threshold The threshold for filtering values based on absolute values.
#' Defaults to `0`.
#' @param return_sparse Whether to return a sparse matrix. Defaults to `false`.
#'
#' @return A matrix.
#' @export
#' @seealso [matrix_to_table]
#'
#' @examples
#' table <- data.frame(
#'   row = c("r1", "r2", "r3", "r4", "r5", "r6"),
#'   col = c("c4", "c5", "c6", "c1", "c2", "c3"),
#'   value = c(0.6, -0.5, -0.4, 0.3, 0.2, 0.1)
#' )
#' matrix <- table_to_matrix(table)
#' table_new <- matrix_to_table(matrix)
#' identical(table, table_new)
#'
#' table_to_matrix(table, threshold = 0.3)
#'
#' table_to_matrix(
#'   table,
#'   row_names = c("r1", "r2"),
#'   col_names = c("c4", "c5")
#' )
#'
#' sparse_matrix <- simulate_sparse_matrix(10, 10)
#' table_sparse <- matrix_to_table(
#'   sparse_matrix
#' )
#' sparse_matrix_new <- table_to_matrix(
#'   table_sparse,
#'   return_sparse = TRUE
#' )
#' identical(sparse_matrix, sparse_matrix_new)
table_to_matrix <- function(table, row_names = NULL, col_names = NULL, threshold = 0.0, return_sparse = FALSE) {
    .Call(`_thisutils_table_to_matrix`, table, row_names, col_names, threshold, return_sparse)
}

test_that("as_matrix converts sparse to dense", {
  m <- simulate_sparse_matrix(10, 10)
  result <- as_matrix(m)
  expect_true(is.matrix(result))
  expect_false(inherits(result, "sparseMatrix"))
})

test_that("as_matrix converts dense to sparse", {
  m <- matrix(1:20, nrow = 4)
  result <- as_matrix(m, return_sparse = TRUE)
  expect_true(inherits(result, "sparseMatrix") || inherits(result, "Matrix"))
})

test_that("as_matrix returns sparse as-is when return_sparse = TRUE", {
  m <- simulate_sparse_matrix(10, 10)
  result <- as_matrix(m, return_sparse = TRUE)
  expect_identical(m, result)
})

test_that("matrix_to_table works on dense matrix", {
  m <- matrix(c(1, 2, 3, 4), nrow = 2)
  rownames(m) <- c("r1", "r2")
  colnames(m) <- c("c1", "c2")
  result <- matrix_to_table(m)
  expect_true(is.data.frame(result))
  expect_equal(ncol(result), 3)
  expect_true(all(c("row", "col", "value") %in% colnames(result)))
})

test_that("matrix_to_table works on sparse matrix", {
  m <- simulate_sparse_matrix(10, 10)
  colnames(m) <- paste0("c", 1:10)
  rownames(m) <- paste0("r", 1:10)
  result <- matrix_to_table(m, keep_zero = FALSE)
  expect_true(is.data.frame(result))
  expect_true(all(result$value != 0))
})

test_that("matrix_to_table keep_zero TRUE on sparse matches dense output", {
  m_sparse <- simulate_sparse_matrix(5, 4, sparsity = 0.7, seed = 2)
  m_dense <- as.matrix(m_sparse)

  sparse_result <- matrix_to_table(m_sparse, keep_zero = TRUE)
  dense_result <- matrix_to_table(m_dense, keep_zero = TRUE)

  expect_equal(sparse_result, dense_result)
  expect_equal(nrow(sparse_result), nrow(m_sparse) * ncol(m_sparse))
})

test_that("matrix_to_table with threshold", {
  m <- matrix(c(0.1, 0.5, 1.0, 2.0), nrow = 2)
  rownames(m) <- c("r1", "r2")
  colnames(m) <- c("c1", "c2")
  result <- matrix_to_table(m, threshold = 0.5)
  expect_true(all(abs(result$value) >= 0.5))
})

test_that("matrix_to_table with row/col filters", {
  m <- matrix(1:9, nrow = 3)
  rownames(m) <- c("r1", "r2", "r3")
  colnames(m) <- c("c1", "c2", "c3")
  result <- matrix_to_table(m, row_names = c("r1"), col_names = c("c1"))
  expect_equal(nrow(result), 1)
})

test_that("matrix_to_table sparse keep_zero TRUE respects filters consistently", {
  m_sparse <- simulate_sparse_matrix(6, 5, sparsity = 0.8, seed = 3)
  m_dense <- as.matrix(m_sparse)

  sparse_result <- matrix_to_table(
    m_sparse,
    row_names = c("row_2", "row_4"),
    col_names = c("col_1", "col_3"),
    keep_zero = TRUE
  )
  dense_result <- matrix_to_table(
    m_dense,
    row_names = c("row_2", "row_4"),
    col_names = c("col_1", "col_3"),
    keep_zero = TRUE
  )

  expect_equal(sparse_result, dense_result)
  expect_equal(nrow(sparse_result), 4)
})

test_that("table_to_matrix roundtrips with matrix_to_table", {
  m <- matrix(c(1, 2, 3, 4, 5, 6), nrow = 2)
  rownames(m) <- c("r1", "r2")
  colnames(m) <- c("c1", "c2", "c3")
  tbl <- matrix_to_table(m)
  m2 <- table_to_matrix(tbl)
  m_reordered <- m[rownames(m2), colnames(m2)]
  expect_equal(m_reordered, m2, tolerance = 1e-10)
})

test_that("table_to_matrix with threshold", {
  tbl <- data.frame(
    row = c("r1", "r2", "r3"),
    col = c("c1", "c1", "c1"),
    value = c(0.1, 0.5, 1.0)
  )
  result <- table_to_matrix(tbl, threshold = 0.5)
  expect_equal(result["r1", "c1"], 0)
  expect_equal(result["r2", "c1"], 0.5)
})

test_that("table_to_matrix with return_sparse", {
  tbl <- data.frame(
    row = c("r1", "r2"),
    col = c("c1", "c2"),
    value = c(1.0, 2.0)
  )
  result <- table_to_matrix(tbl, return_sparse = TRUE)
  expect_true(inherits(result, "sparseMatrix"))
})

test_that("table_to_matrix handles duplicate coordinates consistently", {
  tbl <- data.frame(
    row = c("r1", "r1"),
    col = c("c1", "c1"),
    value = c(1.0, 2.0)
  )

  dense_result <- table_to_matrix(tbl)
  sparse_result <- table_to_matrix(tbl, return_sparse = TRUE)

  expect_equal(dense_result["r1", "c1"], 3)
  expect_equal(as.matrix(sparse_result)["r1", "c1"], 3)
  expect_equal(dense_result, as.matrix(sparse_result))
})

test_that("table_to_matrix sparse path handles names containing separators", {
  tbl <- data.frame(
    row = c("r|1", "r2", "r|1"),
    col = c("c|1", "c2", "c2"),
    value = c(5, 7, 11)
  )

  dense_result <- table_to_matrix(tbl)
  sparse_result <- table_to_matrix(tbl, return_sparse = TRUE)

  expect_equal(as.matrix(sparse_result), dense_result)
  expect_equal(dense_result["r|1", "c|1"], 5)
  expect_equal(dense_result["r2", "c2"], 7)
  expect_equal(dense_result["r|1", "c2"], 11)
})

test_that("table_to_matrix applies threshold before duplicate aggregation consistently", {
  tbl <- data.frame(
    row = c("r1", "r1", "r1", "r2"),
    col = c("c1", "c1", "c1", "c2"),
    value = c(0.2, 0.4, -0.1, 1.0)
  )

  dense_result <- table_to_matrix(tbl, threshold = 0.3)
  sparse_result <- table_to_matrix(tbl, threshold = 0.3, return_sparse = TRUE)

  expect_equal(as.matrix(sparse_result), dense_result)
  expect_equal(dense_result["r1", "c1"], 0.4)
  expect_equal(dense_result["r2", "c2"], 1.0)
})

test_that("split_indices works", {
  result <- split_indices(c(1, 2, 1, 2, 3))
  expect_true(is.list(result))
  expect_equal(length(result), 3)
  expect_equal(sort(result[[1]]), c(1, 3))
  expect_equal(sort(result[[2]]), c(2, 4))
  expect_equal(result[[3]], 5)
})

test_that("split_indices with n parameter", {
  result <- split_indices(c(1, 2, 1), 5)
  expect_equal(length(result), 5)
})

#!/usr/bin/env bash

set -euo pipefail

SCRIPT_DIR="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)"

source "$SCRIPT_DIR/log_message.sh"

# basic usage
log_message "Hello, " "world!"
log_message "hello, world!"
log_message "Hello, world!" --timestamp false
log_message "Hello, " "world!" --message-type success
log_message "Hello, world!" --message-type warning
log_message "Processing data..." --message-type running
log_message "Hello, " "world!" --cli-model false

# suppress messages
log_message "Hello, world!" --verbose false
LOG_MESSAGE_VERBOSE=false log_message "Hello, world!"
LOG_MESSAGE_VERBOSE=true log_message "Hello, world!" --verbose false

# cli inline markup
log_message "{.arg abc} is an argument"
log_message "{.val list('abc')} is a {.cls {class(list('abc'))}}"
log_message "{.code echo hello} is a code example"
log_message "{.dt List}list('abc')"
log_message "address: {.email example@example.com}"
log_message "{.emph Bash} is a scripting language"
log_message "{.envvar SHELL}"
log_message "{.file log_message.sh} is a file"
log_message "{.fn echo} is a function"
log_message "{.fun echo} is a function"
log_message "Use {.help bash} to get help"
log_message "See the {.href [cli homepage](https://cli.r-lib.org)} for details"
log_message "press {.kbd ENTER}"
log_message "press {.key ENTER}"
log_message "URL: {.url https://cli.r-lib.org}"
log_message "Some {.field field}"
log_message "{.path /bin/bash} is a path"
log_message "{.pkg bash} is a package"
log_message "{.val object} is a variable"
log_message "{.run bash scripts/log_message.sh} is a runnable file"
log_message "{.str object} is a string"
log_message "{.strong abc} is a strong string"
log_message "{.topic echo} is a shell help target"

# set indentation
log_message "Hello, world!" --level 2
log_message "Hello, world!" --symbol "->"
log_message "Hello, world!" --symbol "#####" --level 3

# color formatting
log_message "This is a red message" --text-color "#ff9900"
log_message "This is a message with background" --back-color "#EE4000"
log_message "This is a message with both text and background" --text-color "white" --back-color "cyan"
log_message "This is a message with background" --back-color "#EE4000" --cli-model false
log_message "This is a message with both text and background" --text-color "red" --back-color "cyan" --cli-model false
log_message "Hex color with {.arg cli_model = FALSE}" --text-color "#FF5733" --cli-model false
log_message "Bright red text" --text-color "br_red"
log_message "Bright background" --back-color "br_yellow"
log_message "Combined grey and style" --text-color "grey" --text-style "bold"

# text style formatting
log_message "Bold message" --text-style "bold"
log_message "Italic message" --text-style "italic"
log_message "Underlined message" --text-style "underline"
log_message "Combined styles" --text-style "bold,underline"
log_message "Color and style" --text-color "blue" --text-style "bold,italic"
log_message "Hex color and style" --text-color "#FF5733" --text-style "bold,underline"

# multiline message
log_message $'Line 1\nLine 2\nLine 3' --multiline-indent --text-style "italic"
log_message $'Multi-line\ncolored\nmessage' --text-color "blue" --text-style "italic"
log_message $'Multi-line\ncolored\nmessage' --text-color "blue" --timestamp false

# timestamp styling
log_message $'Multi-line message\nwith timestamp styling' --text-color "red" --text-style "bold" --timestamp-style
log_message $'Multi-line message\nwithout timestamp styling' --text-color "#669999" --text-style "bold,italic"

# combine style and log_message (green text + blue highlighted substring)
part1="$(style_formatting 'I am a green line ' 'green' '' '')"
part2="$(style_formatting 'with a blue substring' 'blue' '' 'bold,underline')"
part3="$(style_formatting ' that becomes green again!' 'green' '' '')"
log_message "${part1}${part2}${part3}"

# print objects directly
df_text=$'  x y  z\n1 1 a  a\n2 2 b b  \n3 3 c  c'
log_message "Content:\n${df_text}"

# interactive prompt
if [[ -t 0 && -t 1 ]]; then
	log_message "Do you want to continue?" --message-type ask
fi

#!/usr/bin/env bash

set -o nounset
set -o pipefail

is_hex_color() {
  [[ "$1" =~ ^#([0-9a-fA-F]{3}|[0-9a-fA-F]{6})$ ]]
}

check_color() {
  local color="$1"
  case "$color" in
    black|red|green|yellow|blue|magenta|cyan|white|grey|silver|none|br_black|br_red|br_green|br_yellow|br_blue|br_magenta|br_cyan|br_white)
      return 0
      ;;
  esac

  if is_hex_color "$color"; then
    return 0
  fi

  return 1
}

hex_to_rgb() {
  local hex="${1#\#}"
  if [[ ${#hex} -eq 3 ]]; then
    hex="${hex:0:1}${hex:0:1}${hex:1:1}${hex:1:1}${hex:2:1}${hex:2:1}"
  fi
  local r=$((16#${hex:0:2}))
  local g=$((16#${hex:2:2}))
  local b=$((16#${hex:4:2}))
  echo "$r;$g;$b"
}

make_color_code() {
  local color="$1"
  local bg="$2"

  if is_hex_color "$color"; then
    local rgb
    rgb=$(hex_to_rgb "$color")
    if [[ "$bg" == "true" ]]; then
      echo "48;2;${rgb}"
    else
      echo "38;2;${rgb}"
    fi
    return 0
  fi

  if [[ "$bg" == "true" ]]; then
    case "$color" in
      black) echo "40" ;;
      red) echo "41" ;;
      green) echo "42" ;;
      yellow) echo "43" ;;
      blue) echo "44" ;;
      magenta) echo "45" ;;
      cyan) echo "46" ;;
      white) echo "47" ;;
      none) echo "49" ;;
      br_black) echo "100" ;;
      br_red) echo "101" ;;
      br_green) echo "102" ;;
      br_yellow) echo "103" ;;
      br_blue) echo "104" ;;
      br_magenta) echo "105" ;;
      br_cyan) echo "106" ;;
      br_white) echo "107" ;;
      *) echo "" ;;
    esac
  else
    case "$color" in
      black) echo "30" ;;
      red) echo "31" ;;
      green) echo "32" ;;
      yellow) echo "33" ;;
      blue) echo "34" ;;
      magenta) echo "35" ;;
      cyan) echo "36" ;;
      white) echo "37" ;;
      grey) echo "90" ;;
      silver) echo "37" ;;
      none) echo "39" ;;
      br_black) echo "90" ;;
      br_red) echo "91" ;;
      br_green) echo "92" ;;
      br_yellow) echo "93" ;;
      br_blue) echo "94" ;;
      br_magenta) echo "95" ;;
      br_cyan) echo "96" ;;
      br_white) echo "97" ;;
      *) echo "" ;;
    esac
  fi
}

style_formatting() {
  local msg="$1"
  local text_color="$2"
  local back_color="$3"
  local text_style="$4"

  local codes=()

  if [[ -n "$text_color" ]]; then
    local code
    code=$(make_color_code "$text_color" false)
    [[ -n "$code" ]] && codes+=("$code")
  fi

  if [[ -n "$back_color" ]]; then
    local code
    code=$(make_color_code "$back_color" true)
    [[ -n "$code" ]] && codes+=("$code")
  fi

  if [[ -n "$text_style" ]]; then
    local old_ifs="$IFS"
    IFS=',' read -r -a styles <<< "$text_style"
    IFS="$old_ifs"
    local style
    for style in "${styles[@]}"; do
      case "$style" in
        bold) codes+=("1") ;;
        italic) codes+=("3") ;;
        underline) codes+=("4") ;;
        strikethrough) codes+=("9") ;;
        dim) codes+=("2") ;;
        inverse) codes+=("7") ;;
      esac
    done
  fi

  if [[ ${#codes[@]} -eq 0 ]]; then
    printf '%s' "$msg"
    return 0
  fi

  local joined
  joined=$(IFS=';'; echo "${codes[*]}")
  printf '\033[%sm%s\033[0m' "$joined" "$msg"
}

capitalize() {
  local x="$1"
  [[ -z "$x" ]] && { printf '%s' "$x"; return 0; }

  local first_word
  first_word=$(echo "$x" | awk -F'[ -]+' '{print $1}')
  local lower
  lower=$(printf '%s' "$first_word" | tr '[:upper:]' '[:lower:]')

  if [[ "$first_word" == "$lower" ]]; then
    local first="${x:0:1}"
    local rest="${x:1}"
    printf '%s' "$(printf '%s' "$first" | tr '[:lower:]' '[:upper:]')${rest}"
  else
    printf '%s' "$x"
  fi
}

build_message() {
  local msg=""
  local part
  for part in "$@"; do
    msg+="$part"
  done
  capitalize "$msg"
}

parse_inline_expressions() {
  local msg="$1"

  eval_inline_expression() {
    local expr="$1"
    expr="$(printf '%s' "$expr" | sed -E 's/^[[:space:]]+|[[:space:]]+$//g')"

    if [[ "$expr" =~ ^[A-Za-z_][A-Za-z0-9_]*$ ]]; then
      local var_value="${!expr-}"
      if [[ -n "${var_value}" ]]; then
        printf '%s' "$var_value"
        return 0
      fi
      printf '%s' "$expr"
      return 0
    fi

    if [[ "$expr" =~ ^[0-9[:space:]+*/%().-]+$ ]]; then
      local calc
      calc=$(awk "BEGIN { v = ($expr); if (v == int(v)) printf \"%d\", v; else printf \"%s\", v }")
      printf '%s' "$calc"
      return 0
    fi

    printf '%s' "$expr"
  }

  while [[ "$msg" =~ \{\.val[[:space:]]+\{([^{}]+)\}\} ]]; do
    local expr="${BASH_REMATCH[1]}"
    local full_match="${BASH_REMATCH[0]}"
    local value
    value=$(eval_inline_expression "$expr")
    value=$(style_formatting "$value" "blue" "" "")
    msg="${msg/$full_match/$value}"
  done

  while [[ "$msg" =~ \{([^{}.][^{}]*)\} ]]; do
    local expr="${BASH_REMATCH[1]}"
    local full_match="${BASH_REMATCH[0]}"
    if [[ "$expr" =~ ^[[:space:]]*$ ]]; then
      break
    fi
    local value
    value=$(eval_inline_expression "$expr")
    msg="${msg/$full_match/$value}"
  done

  msg=$(printf '%s' "$msg" | perl -CSDA -pe '
    s/\{\.([A-Za-z_][A-Za-z0-9_]*)(\s{2,})([^{}]+)\}/"{.$1 ".(" " x (length($2)-1))."$3}"/ge;
    s/\{\.arg\s([^{}]+)\}/`$1`/g;
    s/\{\.code\s([^{}]+)\}/`$1`/g;
    s/\{\.run\s([^{}]+)\}/`$1`/g;
    s/\{\.str\s([^{}]+)\}/"$1"/g;
    s/\{\.val\s([^{}]+)\}/\e[34m"$1"\e[0m/g;
    s/\{\.(?:fn|fun)\s([^{}]+)\}/`$1()`/g;
    s/\{\.help\s([^{}]+)\}/`$1 --help`/g;
    s/\{\.topic\s([^{}]+)\}/`help $1`/g;
    s/\{\.vignette\s([^{}]+)\}/`vignette($1)`/g;
    s/\{\.url\s([^{}]+)\}/\e[3;34m<$1>\e[0m/g;
    s/\{\.email\s([^{}]+)\}/\e[34m\x27$1\x27\e[0m/g;
    s/\{\.path\s([^{}]+)\}/\e[34m$1\e[0m/g;
    s/\{\.file\s([^{}]+)\}/\e[34m$1\e[0m/g;
    s/\{\.pkg\s([^{}]+)\}/\e[34m$1\e[0m/g;
    s/\{\.field\s([^{}]+)\}/$1/g;
    s/\{\.emph\s([^{}]+)\}/$1/g;
    s/\{\.strong\s([^{}]+)\}/\e[1m$1\e[0m/g;
    s/\{\.envvar\s([^{}]+)\}/`$1`/g;
    s/\{\.kbd\s([^{}]+)\}/\e[34m[$1]\e[0m/g;
    s/\{\.key\s([^{}]+)\}/\e[34m[$1]\e[0m/g;
    s/\{\.dt\s([^{}]+)\}/$1: /g;
    s/\{\.cls\s([^{}]+)\}/<$1>/g;
    s/\{\.href\s\[([^\]]+)\]\(([^\)]+)\)\}/$1 (\e[3;34m<$2>\e[0m)/g;
    s/\{\.([A-Za-z_][A-Za-z0-9_]*)\s\{([^{}]+)\}\}/$2/g;
    s/\{\.([A-Za-z_][A-Za-z0-9_]*)\s([^{}]+)\}/$2/g;
  ')

  printf '%s' "$msg"
}

parse_bool() {
  local value="${1:-}"
  local normalized
  normalized=$(printf '%s' "$value" | tr '[:upper:]' '[:lower:]')
  case "$normalized" in
    true|t|1|yes|y) echo "true" ;;
    false|f|0|no|n) echo "false" ;;
    *) echo "invalid" ;;
  esac
}

get_verbose() {
  local local_verbose="$1"
  local global_verbose="${LOG_MESSAGE_VERBOSE:-}"

  if [[ -z "$global_verbose" ]]; then
    if [[ -z "$local_verbose" ]]; then
      echo "true"
      return 0
    fi

    local parsed_local
    parsed_local=$(parse_bool "$local_verbose")
    if [[ "$parsed_local" == "invalid" ]]; then
      printf '%s\n' "WARNING: verbose is not a logical value, set to TRUE" >&2
      echo "true"
      return 0
    fi

    echo "$parsed_local"
    return 0
  fi

  local parsed_global
  parsed_global=$(parse_bool "$global_verbose")
  if [[ "$parsed_global" == "invalid" ]]; then
    printf '%s\n' "WARNING: LOG_MESSAGE_VERBOSE is not a logical value, treated as NULL" >&2
    printf '%s\n' "WARNING: or use LOG_MESSAGE_VERBOSE=true/false" >&2
    echo "false"
    return 0
  fi

  echo "$parsed_global"
}

get_indent_part() {
  local symbol="$1"
  local level="$2"

  if [[ "$symbol" != "  " ]]; then
    local repeated=""
    local i
    for ((i = 0; i < level; i++)); do
      repeated+="$symbol"
    done
    printf '%s ' "$repeated"
  elif (( level > 1 )); then
    local repeated=""
    local i
    for ((i = 1; i < level; i++)); do
      repeated+="  "
    done
    printf '%s' "$repeated"
  else
    printf ''
  fi
}

validate_params() {
  local level="$1"
  local symbol="$2"
  local text_color="$3"
  local back_color="$4"
  local text_style="$5"
  local timestamp_style="$6"

  if ! [[ "$level" =~ ^[0-9]+$ ]] || (( level < 1 )); then
    printf '%s\n' "ERROR: level must be a positive integer" >&2
    return 1
  fi

  if [[ -z "$symbol" ]]; then
    printf '%s\n' "ERROR: symbol must be a single character string" >&2
    return 1
  fi

  if [[ -n "$text_color" ]] && ! check_color "$text_color"; then
    printf '%s\n' "ERROR: text_color must be a valid color name or hexadecimal color code" >&2
    return 1
  fi

  if [[ -n "$back_color" ]] && ! check_color "$back_color"; then
    printf '%s\n' "ERROR: back_color must be a valid color name or hexadecimal color code" >&2
    return 1
  fi

  if [[ -n "$text_color" && -n "$back_color" && "$text_color" == "$back_color" ]]; then
    printf '%s\n' "ERROR: text_color and back_color cannot be the same color" >&2
    return 1
  fi

  if [[ -n "$text_style" ]]; then
    local old_ifs="$IFS"
    IFS=',' read -r -a styles <<< "$text_style"
    IFS="$old_ifs"
    local style
    local valid
    for style in "${styles[@]}"; do
      valid="false"
      local candidate
      for candidate in bold italic underline strikethrough dim inverse; do
        if [[ "$style" == "$candidate" ]]; then
          valid="true"
          break
        fi
      done
      if [[ "$valid" != "true" ]]; then
        printf '%s\n' "ERROR: text_style must be one or more of: bold, italic, underline, strikethrough, dim, inverse" >&2
        return 1
      fi
    done
  fi

  local parsed_ts
  parsed_ts=$(parse_bool "$timestamp_style")
  if [[ "$parsed_ts" == "invalid" ]]; then
    printf '%s\n' "ERROR: timestamp_style must be a single logical value" >&2
    return 1
  fi

  return 0
}

output_cli_message() {
  local message="$1"
  local message_type="$2"

  local symbol=""
  case "$message_type" in
    info) symbol="ℹ" ;;
    success) symbol="✔" ;;
    warning) symbol="!" ;;
    running) symbol="◌" ;;
    ask) symbol="?" ;;
  esac

  local symbol_color=""
  case "$message_type" in
    info) symbol_color="cyan" ;;
    success) symbol_color="green" ;;
    warning) symbol_color="yellow" ;;
    running) symbol_color="#FFA500" ;;
    ask) symbol_color="magenta" ;;
  esac

  if [[ -n "$symbol" && -n "$symbol_color" ]]; then
    symbol=$(style_formatting "$symbol" "$symbol_color" "" "")
  fi

  printf '%s %s\n' "$symbol" "$message" >&2
}

plain_text_output() {
  local text="$1"
  local _cli_model="$2"
  local text_color="$3"
  local back_color="$4"
  local text_style="$5"

  if [[ -n "$text_color" || -n "$back_color" || -n "$text_style" ]]; then
    text=$(style_formatting "$text" "$text_color" "$back_color" "$text_style")
  fi

  printf '%s\n' "$text" >&2
}

format_line_with_style() {
  local line="$1"
  local prefix="$2"
  local text_color="$3"
  local back_color="$4"
  local text_style="$5"
  local timestamp_style="$6"

  if [[ -z "$text_color" && -z "$back_color" && -z "$text_style" ]]; then
    printf '%s%s' "$prefix" "$line"
    return 0
  fi

  if [[ "$timestamp_style" == "true" ]]; then
    style_formatting "${prefix}${line}" "$text_color" "$back_color" "$text_style"
  else
    local styled_line
    styled_line=$(style_formatting "$line" "$text_color" "$back_color" "$text_style")
    printf '%s%s' "$prefix" "$styled_line"
  fi
}

output_message() {
  local msg="$1"
  local message_type="$2"
  local cli_model="$3"
  local text_color="$4"
  local back_color="$5"
  local text_style="$6"
  local timestamp="$7"
  local timestamp_format="$8"
  local level="$9"
  local symbol="${10}"
  local multiline_indent="${11}"
  local timestamp_style="${12}"
  local plain_text="${13}"

  if [[ "$plain_text" == "true" ]]; then
    if [[ "$msg" == *$'\n'* ]]; then
      while IFS= read -r line || [[ -n "$line" ]]; do
        plain_text_output "$line" "$cli_model" "$text_color" "$back_color" "$text_style"
      done <<< "$msg"
    else
      plain_text_output "$msg" "$cli_model" "$text_color" "$back_color" "$text_style"
    fi
    return 0
  fi

  if [[ "$cli_model" == "true" && "$msg" == *$'\n'* ]]; then
    local i=0
    local line
    while IFS= read -r line || [[ -n "$line" ]]; do
      i=$((i + 1))
      local indent_part
      indent_part=$(get_indent_part "$symbol" "$level")
      local prefix

      if (( i == 1 )) || [[ "$multiline_indent" == "true" ]]; then
        local timestamp_part=""
        if [[ "$timestamp" == "true" ]]; then
          timestamp_part="$timestamp_format"
        fi
        prefix="${timestamp_part}${indent_part}"
      else
        local alignment_spaces=""
        if [[ "$timestamp" == "true" ]]; then
          local width=${#timestamp_format}
          alignment_spaces=$(printf '%*s' "$width" '')
        fi
        prefix="${alignment_spaces}${indent_part}"
      fi

      local formatted_line
      formatted_line=$(format_line_with_style "$line" "$prefix" "$text_color" "$back_color" "$text_style" "$timestamp_style")
      output_cli_message "$formatted_line" "$message_type"
    done <<< "$msg"
    return 0
  fi

  if [[ "$cli_model" == "true" ]]; then
    local timestamp_part=""
    if [[ "$timestamp" == "true" ]]; then
      timestamp_part="$timestamp_format"
    fi

    local indent_part
    indent_part=$(get_indent_part "$symbol" "$level")

    local final_msg
    if [[ "$symbol" != "  " ]]; then
      local repeated=""
      local i
      for ((i = 0; i < level; i++)); do
        repeated+="$symbol"
      done
      final_msg="${timestamp_part}${repeated} ${msg}"
    else
      final_msg="${timestamp_part}${indent_part}${msg}"
    fi

    if [[ -n "$text_color" || -n "$back_color" || -n "$text_style" ]]; then
      if [[ "$timestamp_style" == "true" ]]; then
        final_msg=$(style_formatting "$final_msg" "$text_color" "$back_color" "$text_style")
      else
        local styled_msg
        styled_msg=$(style_formatting "$msg" "$text_color" "$back_color" "$text_style")

        if [[ "$symbol" != "  " ]]; then
          local repeated=""
          local i
          for ((i = 0; i < level; i++)); do
            repeated+="$symbol"
          done
          final_msg="${timestamp_part}${repeated} ${styled_msg}"
        else
          final_msg="${timestamp_part}${indent_part}${styled_msg}"
        fi
      fi
    fi

    output_cli_message "$final_msg" "$message_type"
    return 0
  fi

  local formatted_msg="$msg"
  if [[ -n "$text_color" || -n "$back_color" || -n "$text_style" ]]; then
    formatted_msg=$(style_formatting "$formatted_msg" "$text_color" "$back_color" "$text_style")
  fi

  local prefix=""
  case "$message_type" in
    info) prefix="" ;;
    success) prefix="SUCCESS: " ;;
    warning) prefix="WARNING: " ;;
    running) prefix="RUNNING: " ;;
    ask) prefix="? " ;;
  esac

  printf '%s%s\n' "$prefix" "$formatted_msg" >&2
}

ask_yes_no_cancel() {
  while true; do
    read -r -p "[y/n/c]: " answer || { echo "NA"; return 0; }
    local normalized
    normalized=$(printf '%s' "$answer" | tr '[:upper:]' '[:lower:]')
    case "$normalized" in
      y|yes) echo "TRUE"; return 0 ;;
      n|no) echo "FALSE"; return 0 ;;
      c|cancel|"") echo "NA"; return 0 ;;
    esac
  done
}

log_message() {
  local verbose="true"
  local message_type="info"
  local cli_model="true"
  local level=1
  local symbol="  "
  local text_color=""
  local back_color=""
  local text_style=""
  local multiline_indent="false"
  local timestamp="true"
  local timestamp_format=""
  local timestamp_style="false"
  local plain_text="false"

  local message_parts=()

  while (($# > 0)); do
    case "$1" in
      --verbose)
        verbose="$2"
        shift 2
        ;;
      --message-type)
        message_type="$2"
        shift 2
        ;;
      --cli-model)
        cli_model="$2"
        shift 2
        ;;
      --level)
        level="$2"
        shift 2
        ;;
      --symbol)
        symbol="$2"
        shift 2
        ;;
      --text-color)
        text_color="$2"
        shift 2
        ;;
      --back-color)
        back_color="$2"
        shift 2
        ;;
      --text-style)
        text_style="$2"
        shift 2
        ;;
      --multiline-indent)
        multiline_indent="true"
        shift
        ;;
      --timestamp)
        timestamp="$2"
        shift 2
        ;;
      --timestamp-format)
        timestamp_format="$2"
        shift 2
        ;;
      --timestamp-style)
        timestamp_style="true"
        shift
        ;;
      --plain-text)
        plain_text="true"
        shift
        ;;
      --)
        shift
        while (($# > 0)); do
          message_parts+=("$1")
          shift
        done
        ;;
      *)
        message_parts+=("$1")
        shift
        ;;
    esac
  done

  local valid_type="false"
  local t
  for t in info success warning error running ask; do
    if [[ "$message_type" == "$t" ]]; then
      valid_type="true"
      break
    fi
  done

  if [[ "$valid_type" != "true" ]]; then
    printf '%s\n' "ERROR: message_type must be one of: info, success, warning, error, running, ask" >&2
    return 1
  fi

  local msg
  msg=$(build_message "${message_parts[@]}")
  msg=$(parse_inline_expressions "$msg")
  msg="${msg//\\n/$'\n'}"

  if [[ "$message_type" == "error" ]]; then
    printf '%s\n' "ERROR: ${msg}" >&2
    return 1
  fi

  local verbose_value
  verbose_value=$(get_verbose "$verbose")
  if [[ "$verbose_value" != "true" ]]; then
    return 0
  fi

  local cli_model_parsed timestamp_parsed timestamp_style_parsed multiline_indent_parsed
  cli_model_parsed=$(parse_bool "$cli_model")
  timestamp_parsed=$(parse_bool "$timestamp")
  timestamp_style_parsed=$(parse_bool "$timestamp_style")
  multiline_indent_parsed=$(parse_bool "$multiline_indent")

  if [[ "$cli_model_parsed" == "invalid" || "$timestamp_parsed" == "invalid" || "$timestamp_style_parsed" == "invalid" || "$multiline_indent_parsed" == "invalid" ]]; then
    printf '%s\n' "ERROR: cli_model/timestamp/multiline_indent/timestamp_style must be logical values" >&2
    return 1
  fi

  if ! validate_params "$level" "$symbol" "$text_color" "$back_color" "$text_style" "$timestamp_style_parsed"; then
    return 1
  fi

  if [[ -z "$timestamp_format" ]]; then
    timestamp_format="[$(date '+%Y-%m-%d %H:%M:%S')] "
  fi

  output_message \
    "$msg" \
    "$message_type" \
    "$cli_model_parsed" \
    "$text_color" \
    "$back_color" \
    "$text_style" \
    "$timestamp_parsed" \
    "$timestamp_format" \
    "$level" \
    "$symbol" \
    "$multiline_indent_parsed" \
    "$timestamp_style_parsed" \
    "$plain_text"

  if [[ "$message_type" == "ask" ]]; then
    ask_yes_no_cancel
  fi
}

if [[ -n "${BASH_VERSION-}" && -n "${BASH_SOURCE-}" && "${BASH_SOURCE[0]}" == "$0" ]]; then
  log_message "$@"
fi

---
title: "Semantics-aware sparse research workflows"
author: "Xu Meng"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Semantics-aware sparse research workflows}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(thisutils)
library(Matrix)
```

Research packages often move the same data through sparse matrices, tabular
representations, neighborhood summaries, and parallel workers. These
transitions are easy to write but difficult to make predictable: an unstored
sparse entry may mean the numeric value zero or the absence of an edge, dense
intermediates may dominate memory, and implicit worker selection may make the
same analysis behave differently across machines. This vignette demonstrates
the four connected contracts used by `thisutils`: semantic, numerical,
resource, and execution contracts.

## Declare the meaning of sparse zeros

For an ordinary numeric sparse matrix, implicit entries are zeros. The default
top-k helper therefore compares stored values and implicit zeros together.

```{r sparse-semantics}
x <- Matrix(
  c(-3, 0, 2,
    -1, 4, 0),
  nrow = 3,
  sparse = TRUE,
  dimnames = list(paste0("r", 1:3), paste0("c", 1:2))
)

run_sparse_topk(x, k = 2, by = "col")
```

Set `by = "row"` when rows, rather than columns, define the independent
selection groups.

For a graph, however, an unstored entry normally means that no edge exists.
Use the stored-entry variant to express that different contract.

```{r graph-semantics}
run_sparse_topk_stored(x, k = 2, by = "col")
```

Sparse-to-table conversion retains its historical expanded default. Compact
stored-entry output is available explicitly and avoids materializing implicit
zeros.

```{r sparse-table}
compact <- matrix_to_table(x, keep_zero = FALSE)
expanded <- matrix_to_table(x)
c(compact_rows = nrow(compact), expanded_rows = nrow(expanded))
```

## Use stable summaries

The row-variance implementation retains the established fast calculation and
uses a stable two-pass fallback for rows at risk of cancellation. This avoids
the failure that can occur when small variations are added to a large offset
without slowing ordinary inputs materially.

```{r stable-variance}
offset <- matrix(1e12 + c(0, 1, 2, 3, 0, 1, 2, 3), nrow = 2, byrow = TRUE)
fast_row_vars(offset)
apply(offset, 1, stats::var)
```

## Bound dense correlation work

Pearson correlation is evaluated in column blocks. `max_dense_bytes` bounds
the dense working block, while `threshold` controls output sparsity. A finite
`max_output_entries` can be supplied when a workflow also needs a hard output
budget.

```{r bounded-correlation}
set.seed(1)
a <- rsparsematrix(100, 20, density = 0.1)
b <- rsparsematrix(100, 8, density = 0.1)

cors <- sparse_cor(
  a,
  b,
  threshold = 0.15,
  block_size = 4,
  max_dense_bytes = 1024^2,
  max_output_entries = 200
)
cors
```

Spearman and Kendall correlation currently require dense ranking and result
matrices. The same memory argument is checked before conversion, so this
boundary fails explicitly rather than depending only on available machine
memory.

## Align serial and parallel execution

The resource-bounded operation can be applied to several datasets without
changing its result contract. `parallelize_fun()` preserves input order and
names, keeps `NULL` results, and represents task failures consistently in
serial and parallel modes. PSOCK is the cross-platform default; selecting the
backend explicitly makes the example independent of platform-specific worker
selection.

```{r controlled-execution}
datasets <- list(first = a, second = b)
correlate <- function(mat) {
  thisutils::sparse_cor(
    mat,
    threshold = 0.15,
    block_size = 4,
    max_dense_bytes = 1024^2,
    max_output_entries = 400
  )
}

serial <- parallelize_fun(
  datasets,
  correlate,
  cores = 1,
  seed = 2026,
  verbose = FALSE
)
psock <- parallelize_fun(
  datasets,
  correlate,
  cores = 2,
  backend = "psock",
  seed = 2026,
  verbose = TRUE,
  progress = FALSE,
  timestamp_format = ""
)

vapply(
  Map(function(lhs, rhs) isTRUE(all.equal(lhs, rhs)), serial, psock),
  identity,
  logical(1)
)
```

The seed is assigned per input rather than per worker, so stochastic tasks are
also reproducible across worker counts and scheduling order.

```{r reproducible-execution}
draw <- function(cores) {
  parallelize_fun(
    1:4,
    function(i) stats::rnorm(2),
    cores = cores,
    backend = "psock",
    seed = 42,
    verbose = FALSE
  )
}

identical(draw(1), draw(2))
```

Per-task and total timeouts provide lifecycle bounds for multi-core calls, and
nested calls automatically run their inner level serially. Explicit fork
workers remain available on Unix-like systems, but are opt-in because R's fork
signal handler can conflict with child-process managers such as `callr` and
`processx` in long-lived sessions.

## Scope

Together, the four contracts distinguish what sparse data mean, how summaries
are calculated, which allocations are permitted, and how independent tasks are
executed. They do not define a distributed-computing framework: Spearman and
Kendall correlation still require dense ranking, output size must be bounded
separately from working memory, and PSOCK worker startup can outweigh parallel
gains for short tasks.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_pkg_file.R
\name{add_pkg_file}
\alias{add_pkg_file}
\title{Add a package file and print package information}
\usage{
add_pkg_file(
  use_figlet = TRUE,
  figlet_font = "Slant",
  colors = c("red", "yellow", "green", "magenta", "cyan", "yellow", "green", "white",
    "magenta", "cyan"),
  verbose = TRUE
)
}
\arguments{
\item{use_figlet}{Whether to use figlet for ASCII art generation.
Default is \code{TRUE}. Details see \link{figlet}.}

\item{figlet_font}{Character string, figlet font to use.
Default is \code{"Slant"}.}

\item{colors}{Character vector, colors to use for the logo elements.}

\item{verbose}{Whether to print the message.
Default is \code{TRUE}.}
}
\value{
Creates a file named \verb{R/<pkg_name>-package.R}.
}
\description{
Add a package file and print package information
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{try_get}
\alias{try_get}
\title{Try to evaluate an expression a set number of times before failing}
\usage{
try_get(expr, max_tries = 5, error_message = "", retry_message = "Retrying...")
}
\arguments{
\item{expr}{The expression to be evaluated.}

\item{max_tries}{The maximum number of attempts to evaluate the expression before giving up.
Default is \code{5}.}

\item{error_message}{Additional custom error message to be displayed when an error occurs.}

\item{retry_message}{Message displayed when a new try to evaluate the expression would be attempted.}
}
\value{
The evaluated expression if successful,
otherwise it throws an error if all attempts are unsuccessful.
}
\description{
The function is used as a fail-safe if code sometimes works and sometimes doesn't,
usually because it depends on a resource that may be temporarily unavailable.
It tries to evaluate the expression \code{max_tries} times.
If all the attempts fail, it throws an error;
if not, the evaluated expression is returned.
}
\examples{
f <- function() {
  value <- runif(1, min = 0, max = 1)
  if (value > 0.5) {
    log_message("value is larger than 0.5")
    return(value)
  } else {
    log_message(
      "value is smaller than 0.5",
      message_type = "error"
    )
  }
}
f_evaluated <- try_get(expr = f())
print(f_evaluated)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/thisutils-package.R
\docType{package}
\name{thisutils-package}
\alias{thisutils}
\alias{thisutils-package}
\title{Collection of Utility Functions for Data Analysis and Computing}
\source{
\url{https://mengxu98.github.io/thisutils/}
}
\description{
Provides utility functions for data analysis and computing. Includes functions for logging, parallel processing, and other computational tasks to streamline workflows.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://mengxu98.github.io/thisutils/}
  \item Report bugs at \url{https://github.com/mengxu98/thisutils/issues}
}

}
\author{
Meng Xu (Maintainer), \email{mengxu98@qq.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/figlet.R
\name{figlet_font}
\alias{figlet_font}
\title{Get a figlet font}
\usage{
figlet_font(font)
}
\arguments{
\item{font}{Path or name of the font to load}
}
\value{
A `figlet_font` object for use with [figlet]
}
\description{
Get a figlet font
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package_management.R
\name{check_r}
\alias{check_r}
\title{Check and install R packages}
\usage{
check_r(
  packages,
  lib = .libPaths()[1],
  dependencies = TRUE,
  force = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{packages}{Package to be installed.
Package source can be \emph{CRAN}, \emph{Bioconductor} or \emph{Github}.
By default, the package name is extracted according to the \code{packages} parameter.}

\item{lib}{The location of the library directories where to install the packages.}

\item{dependencies}{Whether to install dependencies of the packages.
Default is \code{TRUE}.}

\item{force}{Whether to force the installation of packages.
Default is \code{FALSE}.}

\item{verbose}{Whether to print the message.
Default is \code{TRUE}.}
}
\value{
Package installation status.
}
\description{
Check and install R packages
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{matrix_to_table}
\alias{matrix_to_table}
\title{Switch matrix to table}
\usage{
matrix_to_table(
  matrix,
  row_names = NULL,
  col_names = NULL,
  threshold = 0,
  keep_zero = TRUE
)
}
\arguments{
\item{matrix}{A matrix.}

\item{row_names}{Character vector of row names to filter by.}

\item{col_names}{Character vector of column names to filter by.}

\item{threshold}{The threshold for filtering values based on absolute values.
Defaults to \code{0}.}

\item{keep_zero}{Whether to keep zero values in the table. Defaults to \code{false}.}
}
\value{
A table with three columns: \code{row}, \code{col}, and \code{value}.
}
\description{
Switch matrix to table
}
\examples{
test_matrix <- simulate_sparse_matrix(10, 10)
colnames(test_matrix) <- paste0("c", 1:10)
rownames(test_matrix) <- paste0("r", 1:10)
table <- matrix_to_table(test_matrix)
matrix_new <- table_to_matrix(table)
test_matrix <- test_matrix[rownames(matrix_new), colnames(matrix_new)] |>
  as_matrix()
identical(test_matrix, matrix_new)

matrix_to_table(
  test_matrix,
  threshold = 2
)

matrix_to_table(
  test_matrix,
  row_names = c("r1", "r2"),
  col_names = c("c1", "c2")
)
}
\seealso{
\link{table_to_matrix}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computation.R
\name{simulate_sparse_matrix}
\alias{simulate_sparse_matrix}
\title{Generate a simulated sparse matrix}
\usage{
simulate_sparse_matrix(
  nrow,
  ncol,
  sparsity = 0.95,
  distribution_fun = function(n) stats::rpois(n, lambda = 0.5) + 1,
  decimal = 0,
  seed = 1
)
}
\arguments{
\item{nrow}{Number of rows in the matrix.}

\item{ncol}{Number of columns in the matrix.}

\item{sparsity}{Proportion of zero elements (sparsity level).
Default is \code{0.95}, meaning 95\% of elements are zero (5\% are non-zero).}

\item{distribution_fun}{Function to generate non-zero values.}

\item{decimal}{Controls the number of decimal places in the generated values.
If set to \code{0}, values will be integers.
When \code{decimal} > 0, random decimal parts are uniformly distributed across the full range.
Default is \code{0}.}

\item{seed}{Random seed for reproducibility.}
}
\value{
A sparse matrix of class "dgCMatrix".
}
\description{
This function generates a sparse matrix with a specified number of rows and columns,
a given sparsity level, and a distribution function for the non-zero values.
}
\examples{
simulate_sparse_matrix(1000, 500) |>
  check_sparsity()

simulate_sparse_matrix(10, 10, decimal = 1)
simulate_sparse_matrix(10, 10, decimal = 5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{capitalize}
\alias{capitalize}
\title{Capitalize the first letter of each word}
\usage{
capitalize(x, force_tolower = FALSE)
}
\arguments{
\item{x}{A vector of character strings to be capitalized.}

\item{force_tolower}{Whether to force the remaining letters to be lowercase.}
}
\description{
Capitalize the first letter of each word
}
\examples{
x <- c(
  "hello world",
  "hello World"
)
capitalize(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{is_outlier}
\alias{is_outlier}
\title{Detect outliers using MAD (Median Absolute Deviation)}
\usage{
is_outlier(
  x,
  nmads = 2.5,
  constant = 1.4826,
  type = c("both", "lower", "higher")
)
}
\arguments{
\item{x}{Numeric vector.}

\item{nmads}{Number of MADs from the median to define the boundaries for outliers.
Default is \code{2.5}.}

\item{constant}{Constant factor to convert the MAD to a standard deviation.
Default is \code{1.4826}, which is consistent with the MAD of a normal distribution.}

\item{type}{Type of outliers to detect.
Available options are \code{"both"}, \code{"lower"}, or \code{"higher"}.
If \code{type} is \code{"both"}, it detects both lower and higher outliers.
If \code{type} is \code{"lower"}, it detects only lower outliers.
If \code{type} is \code{"higher"}, it detects only higher outliers.}
}
\value{
Numeric vector of indices indicating the positions of outliers in \code{x}.
}
\description{
Detect outliers using MAD (Median Absolute Deviation)
}
\examples{
x <- c(1, 2, 3, 4, 5, 100)
is_outlier(x) # returns 6

x <- c(3, 4, 5, NA, 6, 7)
is_outlier(x, nmads = 1.5, type = "lower") # returns 4

x <- c(10, 20, NA, 15, 35)
is_outlier(x, nmads = 2, type = "higher") # returns 3, 5
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse_cor.R
\name{pearson_correlation}
\alias{pearson_correlation}
\title{Correlation and covariance calculation for sparse matrix}
\usage{
pearson_correlation(x, y = NULL)
}
\arguments{
\item{x}{Sparse matrix or character vector.}

\item{y}{Sparse matrix or character vector.}
}
\value{
A list with covariance and correlation matrices.
}
\description{
Correlation and covariance calculation for sparse matrix
}
\examples{
m1 <- simulate_sparse_matrix(
  100, 100
)
m2 <- simulate_sparse_matrix(
  100, 100,
  sparsity = 0.05
)
a <- pearson_correlation(m1, m2)
a$cov[1:5, 1:5]
a$cor[1:5, 1:5]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{table_to_matrix}
\alias{table_to_matrix}
\title{Switch table to matrix}
\usage{
table_to_matrix(
  table,
  row_names = NULL,
  col_names = NULL,
  threshold = 0,
  return_sparse = FALSE
)
}
\arguments{
\item{table}{A table with three columns: \code{row}, \code{col}, and \code{value}.}

\item{row_names}{Character vector of row names to filter by.}

\item{col_names}{Character vector of column names to filter by.}

\item{threshold}{The threshold for filtering values based on absolute values.
Defaults to \code{0}.}

\item{return_sparse}{Whether to return a sparse matrix. Defaults to \code{false}.}
}
\value{
A matrix.
}
\description{
Switch table to matrix
}
\examples{
table <- data.frame(
  row = c("r1", "r2", "r3", "r4", "r5", "r6"),
  col = c("c4", "c5", "c6", "c1", "c2", "c3"),
  value = c(0.6, -0.5, -0.4, 0.3, 0.2, 0.1)
)
matrix <- table_to_matrix(table)
table_new <- matrix_to_table(matrix)
identical(table, table_new)

table_to_matrix(table, threshold = 0.3)

table_to_matrix(
  table,
  row_names = c("r1", "r2"),
  col_names = c("c4", "c5")
)

sparse_matrix <- simulate_sparse_matrix(10, 10)
table_sparse <- matrix_to_table(
  sparse_matrix
)
sparse_matrix_new <- table_to_matrix(
  table_sparse,
  return_sparse = TRUE
)
identical(sparse_matrix, sparse_matrix_new)
}
\seealso{
\link{matrix_to_table}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_namespace_fun}
\alias{get_namespace_fun}
\title{Get a function from a namespace}
\usage{
get_namespace_fun(pkg, fun)
}
\arguments{
\item{pkg}{The name of the package.}

\item{fun}{The name of the function.}
}
\value{
Function.
}
\description{
Get a function from a namespace
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package_management.R
\name{check_pkg_status}
\alias{check_pkg_status}
\title{Check if a package is installed with the specified version}
\usage{
check_pkg_status(pkg, version = NULL, lib = .libPaths()[1])
}
\arguments{
\item{pkg}{Package name.}

\item{version}{Package version to check. If \code{NULL}, only checks if the package is installed.}

\item{lib}{The location of the library directories where to install the packages.}
}
\value{
\code{TRUE} if the package is installed with the specified version, \code{FALSE} otherwise.
}
\description{
Check if a package is installed with the specified version
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{remove_space}
\alias{remove_space}
\title{Remove and normalize spaces}
\usage{
remove_space(
  x,
  trim_start = TRUE,
  trim_end = FALSE,
  collapse_multiple = TRUE,
  preserve_newlines = TRUE
)
}
\arguments{
\item{x}{A vector of character strings.}

\item{trim_start}{Whether to remove leading spaces before the first word.
Default is \code{TRUE}.}

\item{trim_end}{Whether to remove trailing spaces after the last word.
Default is \code{FALSE}.}

\item{collapse_multiple}{Whether to collapse multiple consecutive spaces between words into a single space.
Default is \code{TRUE}.}

\item{preserve_newlines}{Whether to preserve newline characters when collapsing spaces.
Default is \code{TRUE}.}
}
\value{
A character vector with spaces normalized according to the specified parameters.
}
\description{
Remove and normalize spaces
}
\examples{
x <- c(
  " hello  world ",
  "  test   case  ",
  "no space",
  "   multiple   spaces   "
)
remove_space(x)
remove_space(x, trim_start = FALSE)
remove_space(x, trim_end = TRUE)
remove_space(x, collapse_multiple = FALSE)
remove_space(
  x,
  trim_start = FALSE,
  trim_end = FALSE,
  collapse_multiple = FALSE
)

# with newlines
multiline <- c(
  "hello\n\n  world  ",
  "  first  \n  second  "
)
remove_space(multiline)
remove_space(multiline, preserve_newlines = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{unnest_fun}
\alias{unnest_fun}
\title{Unnest a list-column}
\usage{
unnest_fun(data, cols, keep_empty = FALSE)
}
\arguments{
\item{data}{A data frame.}

\item{cols}{Columns to unnest.}

\item{keep_empty}{By default, you get one row of output for each element of the list your unchopping/unnesting.
This means that if there's a size-0 element (like \code{NULL} or an empty data frame),
that entire row will be dropped from the output.
If you want to preserve all rows,
use \code{keep_empty = TRUE} to replace size-0 elements with a single row of missing values.}
}
\description{
Implement similar functions to the \code{tidyr::unnest} function.
}
\examples{
data <- data.frame(
  id = 1:3,
  x = c("a", "b", "c"),
  stringsAsFactors = FALSE
)
data$data <- list(
  c(1, 2),
  c(3, 4, 5),
  c(6)
)
unnest_fun(data, cols = "data")

data2 <- data.frame(
  id = 1:3,
  x = c("a", "b", "c"),
  stringsAsFactors = FALSE
)
data2$data <- list(
  c(1, 2),
  numeric(0),
  c(6)
)
unnest_fun(data2, cols = "data")
unnest_fun(data2, cols = "data", keep_empty = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/thisutils-package.R
\name{print.thisutils_logo}
\alias{print.thisutils_logo}
\title{Print logo}
\usage{
\method{print}{thisutils_logo}(x, ...)
}
\arguments{
\item{x}{Input information.}

\item{...}{Other parameters.}
}
\value{
Print the ASCII logo
}
\description{
Print logo
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parallelize_fun.R
\name{parallelize_fun}
\alias{parallelize_fun}
\title{Parallelize a function}
\usage{
parallelize_fun(
  x,
  fun,
  cores = 1,
  export_fun = NULL,
  clean_result = FALSE,
  throw_error = TRUE,
  progress_bar_width = 10L,
  timestamp_format = paste0("[", format(Sys.time(), "\%Y-\%m-\%d \%H:\%M:\%S"), "] "),
  verbose = TRUE
)
}
\arguments{
\item{x}{A vector or list to apply over.}

\item{fun}{The function to be applied to each element.}

\item{cores}{The number of cores to use for parallelization with \link[foreach:foreach]{foreach::foreach}.
Default is \code{1}.}

\item{export_fun}{The functions to export the function to workers.}

\item{clean_result}{Whether to remove failed results from output.
If \code{FALSE}, failed results are kept as error objects.
Default is \code{FALSE}.}

\item{throw_error}{Whether to print detailed error information for failed results.
Default is \code{TRUE}.}

\item{progress_bar_width}{Width of the verbose progress bar in characters.
Default is \code{10L}.}

\item{timestamp_format}{Format string for timestamp display.
Default is \code{"\%Y-\%m-\%d \%H:\%M:\%S"}.}

\item{verbose}{Whether to print the message.
Default is \code{TRUE}.}
}
\value{
A list of computed results.
If \code{clean_result = FALSE}, failed results are included as error objects.
If \code{clean_result = TRUE}, only successful results are returned.
}
\description{
Parallelize a function
}
\examples{
parallelize_fun(1:3, function(x) {
  Sys.sleep(0.2)
  x^2
})

parallelize_fun(list(1, 2, 3), function(x) {
  Sys.sleep(0.2)
  x^2
}, cores = 2)

# Examples with error handling
parallelize_fun(1:5, function(x) {
  if (x == 3) stop("Error on element 3")
  x^2
}, clean_result = FALSE)

parallelize_fun(1:5, function(x) {
  if (x == 3) stop("Error on element 3")
  x^2
}, clean_result = TRUE)

# Control error printing
parallelize_fun(1:5, function(x) {
  if (x == 2) stop("Error on element 3")
  if (x == 4) stop("Error on element 4")
  x^2
})

parallelize_fun(1:5, function(x) {
  if (x == 3) stop("Error on element 3")
  x^2
}, throw_error = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{is_linux}
\alias{is_linux}
\title{Check if the operating system is Linux}
\usage{
is_linux()
}
\value{
A logical value.
}
\description{
Check if the operating system is Linux
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse_matrix_tools.R
\name{collapse_sparse_rows}
\alias{collapse_sparse_rows}
\title{Collapse sparse matrix rows by group}
\usage{
collapse_sparse_rows(matrix, group)
}
\arguments{
\item{matrix}{A sparse matrix.}

\item{group}{A vector defining the output row groups.}
}
\value{
A sparse matrix with rows collapsed by `group`.
}
\description{
Collapse sparse matrix rows by group
}
\examples{
mat <- Matrix::Matrix(
  matrix(c(1, 0, 2, 0, 3, 4), nrow = 3, byrow = TRUE),
  sparse = TRUE
)
collapse_sparse_rows(mat, c("g1", "g1", "g2"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{download}
\alias{download}
\title{Download file from the Internet}
\usage{
download(
  url,
  destfile,
  methods = c("auto", "wget", "libcurl", "curl", "wininet", "internal"),
  quiet = FALSE,
  ...,
  max_tries = 2
)
}
\arguments{
\item{url}{a \code{\link{character}} string (or longer vector
    for the \code{"libcurl"} method) naming the URL of a resource to be
    downloaded.}

\item{destfile}{a character string (or vector, see the \code{url}
    argument) with the file path where the downloaded file is to be
    saved.  Tilde-expansion is performed.}

\item{methods}{Methods to be used for downloading files.
Can be \code{"auto"}, \code{"wget"}, \code{"libcurl"}, \code{"curl"}, \code{"wininet"}, \code{"internal"}.
Default is \code{"auto"}, which means to try different download methods.}

\item{quiet}{If \code{TRUE}, suppress status messages (if any), and
    the progress bar.}

\item{...}{Other arguments passed to \link[utils:download.file]{utils::download.file}.}

\item{max_tries}{Number of tries for each download method.
Default is \code{2}.}
}
\description{
Download file from the Internet
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/log_message.R
\name{log_message}
\alias{log_message}
\title{Print formatted message}
\usage{
log_message(
  ...,
  expr = NULL,
  verbose = NULL,
  message_type = c("info", "success", "warning", "error", "running", "ask"),
  cli_model = TRUE,
  level = 1,
  symbol = "  ",
  text_color = NULL,
  back_color = NULL,
  text_style = NULL,
  multiline_indent = FALSE,
  timestamp = TRUE,
  timestamp_format = paste0("[", format(Sys.time(), "\%Y-\%m-\%d \%H:\%M:\%S"), "] "),
  timestamp_style = FALSE,
  plain_text = FALSE,
  .envir = parent.frame(),
  .frame = .envir
)
}
\arguments{
\item{...}{The message to print.}

\item{expr}{An optional expression to evaluate while capturing its standard output,
messages, and warnings, then re-printing them with \code{log_message()} formatting.
The evaluated result is returned invisibly unless it is visible by default.}

\item{verbose}{Whether to print the message.
Default is \code{TRUE}.}

\item{message_type}{Type of message.
Could be choose one of \code{"info"}, \code{"success"}, \code{"warning"}, \code{"error"}, \code{"running"}, and \code{"ask"}.
When \code{"ask"} is used, the function will prompt the user for a Yes/No/Cancel response using \link[utils:askYesNo]{utils::askYesNo},
and returns \code{TRUE} for Yes, \code{FALSE} for No, and \code{NA} for Cancel.
Default is \code{"info"}.}

\item{cli_model}{Whether to use the \code{cli} package to print the message.
Default is \code{TRUE}.}

\item{level}{The level of the message, which affects the indentation.
Level \code{1} has no indentation, higher levels add more indentation.
Default is \code{1}.}

\item{symbol}{The symbol used for indentation.
When specified, it ignores the level parameter and uses the symbol directly.
Default is \code{"  "} (two spaces).}

\item{text_color}{Color for the message text.
Supports R color names (e.g., \code{"orange"}),
hexadecimal colors (e.g., \code{"#000000"}),
basic colors: \code{"red"}, \code{"green"}, \code{"blue"}, \code{"yellow"}, \code{"magenta"},
\code{"cyan"}, \code{"white"}, \code{"black"}, \code{"grey"}, \code{"silver"}, \code{"none"},
and bright colors: \code{"br_red"}, \code{"br_green"}, \code{"br_blue"},
\code{"br_yellow"}, \code{"br_magenta"}, \code{"br_cyan"}, \code{"br_white"}, \code{"br_black"}.
Default is \code{NULL}.}

\item{back_color}{Background color for the message text.
Details see parameter \code{text_color}.
Default is \code{NULL}.}

\item{text_style}{Text styles to apply.
Can be one or more of:
\code{"bold"}, \code{"italic"}, \code{"underline"}, \code{"strikethrough"}, \code{"dim"}, \code{"inverse"}.
Multiple styles can be combined (e.g., \code{c("bold", "underline")}).
Default is \code{NULL}.}

\item{multiline_indent}{Whether to apply consistent formatting (timestamp and indentation) to each line in multiline messages.
When \code{TRUE}, each line gets the full formatting; when \code{FALSE}, only the first line gets the timestamp.
Default is \code{FALSE}.}

\item{timestamp}{Whether to show the current time in the message.
Default is \code{TRUE}.}

\item{timestamp_format}{Format string for timestamp display.
Default is \code{"\%Y-\%m-\%d \%H:\%M:\%S"}.}

\item{timestamp_style}{Whether to apply the same text styling to the timestamp as the message text.
When \code{TRUE}, timestamp formatting matches the message;
when \code{FALSE}, timestamp keeps its default appearance.
Default is \code{FALSE}.}

\item{plain_text}{Whether to print only the text content.
When \code{TRUE}, level, symbol, timestamp, and message type formatting are suppressed,
but color and multiline settings still apply.}

\item{.envir}{The environment to evaluate calls in.
Default is \link{parent.frame}.}

\item{.frame}{The frame to use for error reporting.
Default is \code{.envir}.}
}
\value{
Formated message, a logical value (\code{TRUE}/\code{FALSE}/\code{NA}) if \code{message_type = "ask"},
or the evaluated result of \code{expr} if \code{expr} is supplied.
}
\description{
Integrate the message printing function with the \href{https://cli.r-lib.org}{cli} package,
and the \link[base:message]{base::message} function.
The message could be suppressed by \link[base:message]{base::suppressMessages}.
}
\examples{
# basic usage
log_message("Hello, ", "world!")

log_message("hello, world!")

log_message("Hello, world!", timestamp = FALSE)

log_message(
  "Hello, ", "world!",
  message_type = "success"
)

log_message(
  "Hello, world!",
  message_type = "warning"
)

log_message(
  "Processing data...",
  message_type = "running"
)

log_message(
  "Hello, ", "world!",
  cli_model = FALSE
)


# suppress messages
suppressMessages(log_message("Hello, world!"))
log_message("Hello, world!", verbose = FALSE)
options(log_message.verbose = FALSE)
log_message("Hello, world!")

# for global verbose option
options(log_message.verbose = TRUE)
log_message("Hello, world!", verbose = FALSE)
options(log_message.verbose = NULL)


# cli inline markup
log_message("{.arg abc} is a argument")

## 'message' can not deal with cli inline markup
message("hello, {.code world}!")

log_message("{.val list('abc')} is a {.cls {class(list('abc'))}}")

log_message("{.code lm(y ~ x)} is a code example")

log_message("{.dt List}list('abc')")

log_message("address: {.email example@example.com}")

log_message("{.emph R} is a programming language")

log_message("{.envvar R_HOME}")

log_message("{.file log_message.R} is a file")

log_message("{.fn lm} is a function")

log_message("{.fun lm} is a function")

log_message("{.help lm} to get help")

log_message("... see {.help [{.fun lm}](stats::lm)} to learn more")

log_message(
  "See the {.href [cli homepage](https://cli.r-lib.org)} for details"
)

log_message("press {.kbd ENTER}")

log_message("press {.key ENTER}")

log_message("URL: {.url https://cli.r-lib.org}")

log_message("Some {.field field}")

log_message("{.path /usr/bin/R} is a path")

log_message("{.pkg cli} is a package")

log_message("{.val object} is a variable")

log_message("{.run Rscript log_message.R} is a runnable file")

log_message("{.str object} is a string")

log_message("{.strong abc} is a strong string")

log_message("{.topic stats::lm} is a topic")

log_message("{.vignette cli} is a vignette")


# set indentation
log_message("Hello, world!", level = 2)

log_message("Hello, world!", symbol = "->")

log_message(
  "Hello, world!",
  symbol = "#####",
  level = 3
)

# color formatting
log_message(
  "This is a red message",
  text_color = "#ff9900"
)

log_message(
  "This is a message with background",
  back_color = "#EE4000"
)

log_message(
  "This is a message with both text and background",
  text_color = "white",
  back_color = "cyan"
)

log_message(
  "This is a message with background",
  back_color = "#EE4000",
  cli_model = FALSE
)

log_message(
  "This is a message with both text and background",
  text_color = "red",
  back_color = "cyan",
  cli_model = FALSE
)

log_message(
  "Hex color with {.arg cli_model = FALSE}",
  text_color = "#FF5733",
  cli_model = FALSE
)

log_message(
  "Bright red text",
  text_color = "br_red"
)

log_message(
  "Bright background",
  back_color = "br_yellow"
)

log_message(
  "Combined grey and style",
  text_color = "grey",
  text_style = "bold"
)

# text style formatting
log_message(
  "Bold message",
  text_style = "bold"
)

log_message(
  "Italic message",
  text_style = "italic"
)

log_message(
  "Underlined message",
  text_style = "underline"
)

log_message(
  "Combined styles",
  text_style = c("bold", "underline")
)

log_message(
  "Color and style",
  text_color = "blue",
  text_style = c("bold", "italic")
)

log_message(
  "Hex color and style",
  text_color = "#FF5733",
  text_style = c("bold", "underline")
)


# multiline message
log_message(
  "Line 1\nLine 2\nLine 3",
  multiline_indent = TRUE,
  text_style = "italic"
)

log_message(
  "Multi-line\ncolored\nmessage",
  text_color = "blue",
  text_style = "italic"
)

log_message(
  "Multi-line\ncolored\nmessage",
  text_color = "blue",
  timestamp = FALSE
)

# timestamp styling
log_message(
  "Multi-line message\nwith timestamp styling",
  text_color = "red",
  text_style = "bold",
  timestamp_style = TRUE
)

log_message(
  "Multi-line message\nwithout timestamp styling",
  text_color = "#669999",
  text_style = c("bold", "italic")
)


# combine cli package and log_message
log_message(
  cli::col_green(
    "I am a green line ",
    cli::col_blue(
      cli::style_underline(
        cli::style_bold("with a blue substring")
      )
    ),
    " that becomes green again!"
  )
)

# cli variables
fun <- function(x = 1) {
  log_message("{.val x}")
  log_message("{.val {x}}")
  log_message("{.val {x + 1}}")
}
fun()


# print objects directly
df <- data.frame(
  x = 1:3,
  y = letters[1:3],
  z = c(" a", "b  ", "c")
)
log_message("Content:\n", df)

# interactive prompt
if (interactive()) {
  log_message(
    "Do you want to continue?",
    message_type = "ask"
  )
}

# capture output from another expression
fun <- function() {
  cat("This is standard output\n")
  message("This is a message")
  return(1 + 1)
}
fun()

log_message(
  expr = fun(),
  message_type = "running"
)
}
\references{
\url{https://cli.r-lib.org/articles/index.html}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computation.R
\name{normalization}
\alias{normalization}
\title{Normalize numeric vector}
\usage{
normalization(x, method = "max_min", na_rm = TRUE, ...)
}
\arguments{
\item{x}{Input numeric vector.}

\item{method}{Method used for normalization.}

\item{na_rm}{Whether to remove \code{NA} values,
and if setting \code{TRUE}, using \code{0} instead.
Default is \code{TRUE}.}

\item{...}{Parameters for other methods.}
}
\value{
Normalized numeric vector.
}
\description{
Normalize numeric vector
}
\examples{
x <- c(runif(2), NA, -runif(2))
x
normalization(x, method = "max_min")
normalization(x, method = "maximum")
normalization(x, method = "sum")
normalization(x, method = "softmax")
normalization(x, method = "z_score")
normalization(x, method = "mad")
normalization(x, method = "unit_vector")
normalization(x, method = "unit_vector", na_rm = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/figlet.R
\name{figlet}
\alias{figlet}
\title{The figlet function}
\usage{
figlet(
  text,
  font = "Slant",
  width = getOption("width", 80),
  justify = "left",
  absolute = FALSE,
  strip = TRUE
)
}
\arguments{
\item{text}{Text to make bigger.}

\item{font}{Name of font, path to font, or \code{figlet_font} object.}

\item{width}{Width to use when justifying and breaking lines.}

\item{justify}{Text justification to use in rendering ("left", "centre", "right").}

\item{absolute}{Logical, indicating if alignment is absolute.}

\item{strip}{Logical, indicating if whitespace should be removed.}
}
\value{
An object of class \code{figlet_text} which is a character vector with a handy print method.
}
\description{
Create ASCII art text using figlet.
}
\examples{
figlet("thisutils")
}
\references{
\url{https://github.com/richfitz/rfiglet},
\url{https://github.com/jbkunst/figletr},
\url{https://www.figlet.org/}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse_cor.R
\name{sparse_cor}
\alias{sparse_cor}
\title{Sparse correlation function}
\usage{
sparse_cor(
  x,
  y = NULL,
  method = "pearson",
  allow_neg = TRUE,
  remove_na = TRUE,
  remove_inf = TRUE,
  ...
)
}
\arguments{
\item{x}{Sparse matrix or character vector.}

\item{y}{Sparse matrix or character vector.}

\item{method}{Method to use for calculating the correlation coefficient.}

\item{allow_neg}{Logical. Whether to allow negative values or set them to 0.}

\item{remove_na}{Logical. Whether to replace NA values with 0.}

\item{remove_inf}{Logical. Whether to replace infinite values with 1.}

\item{...}{Other arguments passed to \link[stats:cor]{stats::cor} function.}
}
\value{
A correlation matrix.
}
\description{
Safe correlation function which returns a sparse matrix.
}
\examples{
m1 <- simulate_sparse_matrix(
  500, 100
)
m2 <- simulate_sparse_matrix(
  500, 100,
  seed = 2025
)
a <- sparse_cor(m1)
b <- sparse_cor(m1, m2)
c <- as_matrix(
  cor(as_matrix(m1)),
  return_sparse = TRUE
)
d <- as_matrix(
  cor(as_matrix(m1), as_matrix(m2)),
  return_sparse = TRUE
)

a[1:5, 1:5]
c[1:5, 1:5]
all.equal(a, c)

b[1:5, 1:5]
d[1:5, 1:5]
all.equal(b, d)

m1[sample(1:500, 10)] <- NA
m2[sample(1:500, 10)] <- NA

sparse_cor(m1, m2)[1:5, 1:5]

system.time(
  sparse_cor(m1)
)
system.time(
  cor(as_matrix(m1))
)

system.time(
  sparse_cor(m1, m2)
)
system.time(
  cor(as_matrix(m1), as_matrix(m2))
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{invoke_fun}
\alias{invoke_fun}
\title{Invoke a function with a list of arguments}
\usage{
invoke_fun(.fn, .args = list(), ..., .env = rlang::caller_env())
}
\arguments{
\item{.fn}{A function, or function name as a string.}

\item{.args}{A list of arguments.}

\item{...}{Other arguments passed to the function.}

\item{.env}{Environment in which to evaluate the call.
This will be most useful if \code{.fn} is a string,
or the function has side-effects.}
}
\description{
Invoke a function with a list of arguments
}
\examples{
f <- function(x, y) {
  x + y
}
invoke_fun(f, list(x = 1, y = 2))
invoke_fun("f", list(x = 1, y = 2))
invoke_fun("f", x = 1, y = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{check_ci_env}
\alias{check_ci_env}
\title{Check CI environment}
\usage{
check_ci_env()
}
\value{
A logical value.
}
\description{
Check CI environment
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computation.R
\name{check_sparsity}
\alias{check_sparsity}
\title{Check sparsity of matrix}
\usage{
check_sparsity(x)
}
\arguments{
\item{x}{A matrix.}
}
\value{
Sparsity of matrix.
}
\description{
Check sparsity of matrix
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/log_message.R
\name{parse_inline_expressions}
\alias{parse_inline_expressions}
\title{Parse inline expressions}
\usage{
parse_inline_expressions(text, env = parent.frame())
}
\arguments{
\item{text}{A character string containing inline expressions to parse.}

\item{env}{Environment in which to evaluate expressions.
Defaults to the calling environment.}
}
\value{
A character string with expressions evaluated but formatting preserved.
}
\description{
Parse `{}` inline expressions and evaluate them in the current environment,
while preserving outer formatting markers like `\{.val ...\}`.
}
\examples{
i <- 1
parse_inline_expressions(
  "{.val {i}}"
)

x <- 5
y <- 10
parse_inline_expressions(
  "{.pkg {x + y}}"
)

name <- "testing"
name <- parse_inline_expressions(
  "{.pkg {name}}"
)
name

log_message(name)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{is_apple_silicon}
\alias{is_apple_silicon}
\title{Check if the system is running on Apple Silicon}
\usage{
is_apple_silicon()
}
\value{
A logical value.
}
\description{
Check if the system is running on Apple Silicon
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{sump}
\alias{sump}
\title{Sum P-value}
\usage{
sump(p)
}
\arguments{
\item{p}{A vector of P-values.}
}
\description{
Sum P-value
}
\examples{
p <- c(0.01, 0.02, 0.03, 0.04, 0.05)
sump(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_matrix.R
\name{as_matrix}
\alias{as_matrix}
\title{Convert matrix into dense/sparse matrix}
\usage{
as_matrix(x, return_sparse = FALSE)
}
\arguments{
\item{x}{A matrix.}

\item{return_sparse}{Whether to output a sparse matrix.
Default is \code{FALSE}.}
}
\value{
A dense or sparse matrix.
}
\description{
Convert matrix into dense/sparse matrix
}
\examples{
m <- simulate_sparse_matrix(
  1000, 1000,
  decimal = 3
)

a <- as_matrix(m)
a[1:5, 1:5]

b <- as_matrix(m, return_sparse = TRUE)
b[1:5, 1:5]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lisi.R
\name{compute_lisi}
\alias{compute_lisi}
\title{Compute Local Inverse Simpson's Index (LISI)}
\usage{
compute_lisi(
  X,
  meta_data,
  label_colnames,
  perplexity = 30,
  nn_eps = 0,
  use_rann = TRUE,
  nn_method = c("auto", "rann", "fnn", "exact"),
  tol = 1e-05,
  max_iter = 50
)
}
\arguments{
\item{X}{A matrix-like object with cells in rows and embedding/features in columns.}

\item{meta_data}{A data frame with one row per cell.}

\item{label_colnames}{Character vector of column names in \code{meta_data} to evaluate.}

\item{perplexity}{Effective neighborhood size. Defaults to \code{30}.}

\item{nn_eps}{Approximation factor passed to \link[RANN:nn2]{RANN::nn2} when \code{RANN} is available and \code{use_rann = TRUE}.
Defaults to \code{0}.}

\item{use_rann}{Whether to prefer \link[RANN:nn2]{RANN::nn2} over \code{FNN::get.knn} when \code{nn_method = "auto"} decides not to use the package's built-in exact C++ backend.
Defaults to \code{TRUE}.}

\item{nn_method}{Nearest-neighbor backend. Defaults to \code{"auto"},
which prefers the package's exact C++ search only for very small problems.
Larger inputs fall back to \code{RANN}, then \code{FNN}, and only use the exact backend as a last resort.}

\item{tol}{Tolerance used in the binary search for the target perplexity.
Defaults to \code{1e-5}.}

\item{max_iter}{Maximum number of binary-search iterations.
Defaults to \code{50}.}
}
\value{
A data frame with one row per cell and one column per label.
}
\description{
Compute per-cell Local Inverse Simpson's Index (LISI) scores for one or more categorical variables.
This is a clean-room reimplementation of the \code{immunogenomics/LISI}.
}
\examples{
set.seed(1)
X <- rbind(
  matrix(stats::rnorm(100, mean = -2), ncol = 5),
  matrix(stats::rnorm(100, mean = 2), ncol = 5)
)
meta_data <- data.frame(
  batch = rep(c("A", "B"), each = 20),
  group = sample(c("g1", "g2"), 40, replace = TRUE)
)

res <- compute_lisi(
  X, meta_data,
  c("batch", "group"),
  perplexity = 10
)
head(res)
boxplot(res)
}
\references{
Korsunsky I, Millard N, Fan J, et al. Fast, sensitive and accurate
integration of single-cell data with Harmony. \emph{Nature Methods} (2019).
\url{https://www.nature.com/articles/s41592-019-0619-0}

LISI reference implementation:
\url{https://github.com/immunogenomics/LISI}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{split_indices}
\alias{split_indices}
\title{Split indices.}
\usage{
split_indices(group, n = 0L)
}
\arguments{
\item{group}{Integer indices}

\item{n}{The largest integer (may not appear in index).
This is hint: if the largest value of \code{group} is bigger than \code{n},
the output will silently expand.}
}
\value{
A list of vectors of indices.
}
\description{
An optimised version of split for the special case of splitting row indices into groups.
}
\examples{
split_indices(sample(10, 100, rep = TRUE))
split_indices(sample(10, 100, rep = TRUE), 10)
}
\references{
\url{https://github.com/hadley/plyr/blob/d57f9377eb5d56107ba3136775f2f0f005f33aa3/src/split-numeric.cpp#L20}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{wilkinsonp}
\alias{wilkinsonp}
\title{Wilkinson's P-value}
\usage{
wilkinsonp(p, r = 1, alpha = 0.05, log.p = FALSE)
}
\arguments{
\item{p}{A vector of P-values.}

\item{r}{The number of studies to include in the P-value calculation.}

\item{alpha}{The significance level.}

\item{log.p}{Whether to return the log of the P-value.}
}
\description{
Wilkinson's P-value
}
\examples{
p <- c(0.01, 0.02, 0.03, 0.04, 0.05)
wilkinsonp(p)
wilkinsonp(p, r = 2)
wilkinsonp(p, alpha = 0.01)
wilkinsonp(p, log.p = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{votep}
\alias{votep}
\title{Vote P-value}
\usage{
votep(p, alpha = 0.5)
}
\arguments{
\item{p}{A vector of P-values.}

\item{alpha}{The significance level.}
}
\description{
Vote P-value
}
\examples{
p <- c(0.01, 0.02, 0.03, 0.04, 0.05)
votep(p)
votep(p, alpha = 0.01)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package_management.R
\name{remove_r}
\alias{remove_r}
\title{Check and remove R packages}
\usage{
remove_r(packages, lib = .libPaths()[1], verbose = TRUE)
}
\arguments{
\item{packages}{Package to be removed.}

\item{lib}{The location of the library directories where to remove the packages.}

\item{verbose}{Whether to print the message.
Default is \code{TRUE}.}
}
\description{
Check and remove R packages
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{meanp}
\alias{meanp}
\title{Mean P-value}
\usage{
meanp(p)
}
\arguments{
\item{p}{A vector of P-values.}
}
\description{
Mean P-value
}
\examples{
p <- c(0.01, 0.02, 0.03, 0.04, 0.05)
meanp(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computation.R
\name{matrix_process}
\alias{matrix_process}
\title{Process matrix}
\usage{
matrix_process(
  matrix,
  method = c("raw", "zscore", "fc", "log2fc", "log1p"),
  ...
)
}
\arguments{
\item{matrix}{A matrix.}

\item{method}{Method to use for processing the matrix.}

\item{...}{Other arguments passed to the method.}
}
\value{
A processed matrix.
}
\description{
Process matrix
}
\examples{
m <- simulate_sparse_matrix(10, 10)
matrix_process(m, method = "raw")
matrix_process(m, method = "zscore")
matrix_process(m, method = "fc")
matrix_process(m, method = "log2fc")
matrix_process(m, method = "log1p")
m <- as_matrix(m)
matrix_process(m, method = function(x) x / rowMeans(x))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lisi.R
\name{compute_simpson_index}
\alias{compute_simpson_index}
\title{Compute Simpson index from a KNN graph}
\usage{
compute_simpson_index(
  D,
  knn_idx,
  batch_labels,
  perplexity = 15,
  tol = 1e-05,
  max_iter = 50
)
}
\arguments{
\item{D}{Numeric matrix of nearest-neighbor distances with neighbors in rows
and cells in columns.}

\item{knn_idx}{Integer matrix of nearest-neighbor indices with the same shape as \code{D}.
One-based indices are expected; zero-based indices are also accepted and converted automatically.}

\item{batch_labels}{Integer-like label vector of length equal to the number of cells.}

\item{perplexity}{Effective neighborhood size.
Defaults to \code{15}.}

\item{tol}{Tolerance used in the binary search for the target perplexity.
Defaults to \code{1e-5}.}

\item{max_iter}{Maximum number of binary-search iterations.
Defaults to \code{50}.}
}
\value{
A numeric vector containing the local Simpson index for each cell.
}
\description{
Given nearest-neighbor distances, nearest-neighbor indices,
and a categorical label for each cell,
compute the local Simpson index of each cell after matching the target perplexity.
}
\examples{
D <- matrix(
  c(0.1, 0.2, 0.2, 0.1,
    0.3, 0.4, 0.4, 0.3),
  nrow = 2,
  byrow = TRUE
)
knn_idx <- matrix(
  c(2, 1, 4, 3,
    3, 4, 2, 1),
  nrow = 2,
  byrow = TRUE
)
batch_labels <- c(1, 1, 2, 2)
compute_simpson_index(D, knn_idx, batch_labels, perplexity = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{\%ss\%}
\alias{\%ss\%}
\title{Value selection operator}
\usage{
a \%ss\% b
}
\arguments{
\item{a}{The left side value to check.}

\item{b}{The right side value to use if \code{a} is \code{NULL}.}
}
\value{
\code{a} if it is not \code{NULL}, otherwise \code{b}.
}
\description{
This operator returns the left side if it's not \code{NULL},
otherwise it returns the right side.
}
\examples{
NULL \%ss\% 10
5 \%ss\% 10
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/log_message.R
\name{get_verbose}
\alias{get_verbose}
\title{Get the verbose option}
\usage{
get_verbose(verbose = NULL)
}
\arguments{
\item{verbose}{The verbose option.
Default is `NULL`, which means to get the verbose option from the global options.}
}
\value{
The verbose option.
}
\description{
Get the verbose option from the global options or the local argument.
}
\examples{
get_verbose()
get_verbose(verbose = FALSE)
get_verbose(verbose = TRUE)

options(log_message.verbose = FALSE)
get_verbose()
get_verbose(verbose = TRUE)

options(log_message.verbose = TRUE)
get_verbose()

options(log_message.verbose = NULL)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/computation.R
\name{r_square}
\alias{r_square}
\title{Coefficient of determination (\eqn{R^2})}
\usage{
r_square(y_true, y_pred)
}
\arguments{
\item{y_true}{A numeric vector with ground truth values.}

\item{y_pred}{A numeric vector with predicted values.}
}
\value{
The \eqn{R^2} value.
}
\description{
Coefficient of determination (\eqn{R^2})
}
\examples{
y <- rnorm(100)
y_pred <- y + rnorm(100, sd = 0.5)
r_square(y, y_pred)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{max_depth}
\alias{max_depth}
\title{Maximum depth of a list}
\usage{
max_depth(x, depth = 0)
}
\arguments{
\item{x}{A list.}

\item{depth}{The depth of the list.}
}
\description{
Maximum depth of a list
}
\examples{
x <- list(
  a = list(b = list(c = 1)),
  d = list(e = list(f = 2))
)
max_depth(x)
}
